unit Dateedit;

interface

uses Buttons,Windows,Messages,CalPop,Classes,Controls,IniFiles,SysUtils,StdCtrls;

type

  PTDateTime=^TDateTime;

  TDateButton=class(TBitBtn)
  protected
    procedure Click; override;
  end;

  TDateEdit=class(TEdit)
  private
    hBitmap:HBitmap;
    FButton:TDateButton;
    FDate:TDateTime;
    FOnDateChange:TNotifyEvent;
    sSep:Char;
    sDateFmt:String[20];
    Token:Integer;
    nSep1,nSep2:Integer;
    procedure SetToken;
    procedure SelectToken;
    procedure SetSeperators;
    procedure SelectErrorToken;
  protected
    procedure WMSize(var Message:TWMSize); message WM_SIZE;
    function GetDate:TDateTime;
    procedure SetDate(dtArg:TDateTime);
    procedure KeyPress(var Key:Char); override;
    procedure MouseUp(Button:TMouseButton; ShiftState:TShiftState; X,Y:Integer); override;
    procedure DblClick; override;
    procedure DoExit; override;
    procedure DoEnter; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure CreateParams(var Params:TCreateParams); override;
    property Date:TDateTime read GetDate write SetDate;
  published
    property OnDateChange:TNotifyEvent read FOnDateChange write FOnDateChange;
  end;

procedure Register;

var frmCalendar:TfrmCalPop;

implementation

{$R DATEEDIT}

const DaysMon:Array[1..12] of Integer=(31,28,31,30,31,30,31,31,30,31,30,31);
{--- TDateButton ---}

procedure TDateButton.Click;
var editParent:TDateEdit;
    parentDate:TDateTime;
begin
  try
    editParent:=TDateEdit(Parent);
    if editParent.Text<>'' then
      parentDate:=StrToDate(editParent.Text);
    frmCalendar:=TfrmCalPop.Create(editParent);
    frmCalendar.ShowModal;
    frmCalendar.Free;
    inherited Click;
  except
    on EConvertError do begin
      editParent.SetFocus;
      editParent.SelectErrorToken;
    end;
  end;
end;

{--- TDateEdit ---}

constructor TDateEdit.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  sSep:='.';
  sDateFmt:='dd.MM.yyyy.';
  Token:=1;
  FDate:=0.0;
  Text:='';
  FButton:=TDateButton.Create(self);
  FButton.Visible:=TRUE;
  FButton.Parent:=self;
  FButton.Glyph.Handle:=LoadBitmap(hInstance,'CALPOPUP');
  ControlStyle:=ControlStyle-[csSetCaption];
  MaxLength:=11;
end;

procedure TDateEdit.CreateParams(var Params:TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style:=Params.Style or WS_CLIPCHILDREN;
end;

destructor TDateEdit.Destroy;
begin
  FButton:=nil;
  inherited Destroy;
end;

procedure TDateEdit.WMSize(var Message:TWMSize);
begin
  FButton.Top:=0;
  FButton.Height:=Height-4;
  FButton.Width:=Height-4;
  FButton.Left:=Width-Height;
  FButton.Refresh;
  if FDate=0.0 then begin
    Text:='';
  end;
//  if FDate=0.0 then Date:=Now;
end;

function TDateEdit.GetDate:TDateTime;
begin
  Result:=FDate;
end;

procedure TDateEdit.SetDate(dtArg:TDateTime);
begin
  if FDate<>dtArg then begin
    FDate:=dtArg;
    Modified:=TRUE;
    if FDate=0 then Text:=''
    else Text:=FormatDateTime(sDateFmt,FDate);
    if Assigned(FOnDateChange) then FOnDateChange(self);
  end;
end;

procedure TDateEdit.DoEnter;
begin
  inherited DoEnter;
  Token:=1;
  SetSeperators;
  SelectToken;
end;

procedure TDateEdit.DoExit;
begin
  inherited DoExit;
  try
    if Text<>'' then
      Date:=StrToDate(Text)
    else
      Date:=0;
  except
    on EConvertError do begin
      SetFocus;
    end;
  end;
end;

procedure TDateEdit.KeyPress(var Key:Char);
begin
  if ((Key<'0') or (Key>'9')) and (Key<>sSep) and (Key<>#8) then
    Key:=#0
  else
    if Key=sSep then begin
      if Token<3 then begin
        Inc(Token);
        SetSeperators;
      end
      else
        Key:=#0;
    end
    else
      inherited KeyPress(Key);
end;

procedure TDateEdit.DblClick;
begin
	inherited DblClick;

  Date:=Now;
end;

{Determine which token the user is on and highlight the entire text of that token.}
procedure TDateEdit.MouseUp(Button:TMouseButton; ShiftState:TShiftState; X,Y:Integer);
begin
  SetToken;
  SelectToken;
  inherited MouseUp(Button,ShiftState,X,Y);
end;

{Set the positions of the seperators in text.}
procedure TDateEdit.SetSeperators;
var i:Integer;
begin
  nSep1:=Pos(sSep,Text);
  for i:=nSep1+1 to Length(Text) do
    if Text[i]=sSep then begin
      nSep2:=i;
      Break;
    end;
end;

{Determine which token the cursor is over;}
procedure TDateEdit.SetToken;
var nPos:Integer;
begin
  nPos:=SelStart;
  SetSeperators;
  if nPos<=nSep1 then Token:=1
  else
    if nPos<=nSep2 then Token:=2
    else Token:=3;
end;

{Select the token the cursor is on.}
procedure TDateEdit.SelectToken;
begin
  case Token of
    1:SendMessage(Handle,EM_SETSEL,0,nSep1-1);
    2:SendMessage(Handle,EM_SETSEL,nSep1,nSep2-1);
    3:SendMessage(Handle,EM_SETSEL,nSep2,Length(Text)-1);
  end;
end;

procedure TDateEdit.SelectErrorToken;
var i,Ov:Integer;
    Da:Array[1..3] of Integer;
begin
  for i:=1 to 3 do begin
    Token:=i;
    SelectToken;
    Da[i]:=StrToInt(SelText);
  end;
  if (Da[3] mod 4=0) and (Da[2]=2) then Ov:=1 else Ov:=0;
  if Da[1]>DaysMon[Da[2]+Ov] then Token:=1
  else
    if Da[2]>12 then Token:=2
    else Token:=3;
  SelectToken;
  MessageBeep($FFFF);
end;

procedure Register;
begin
  RegisterComponents('Samples',[TDateEdit]);
end;

end.