{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: Component for saving form size and location
//******************************************************
//******************************************************

unit FSaver;

interface

uses        
	Classes, Controls, Messages, WinTypes, Forms, Graphics, WinProcs, StdCtrls,
	SysUtils, IniFiles,INIFile,INIMis;

const
	CSecPrefix='Form\';
	CMinW=96;
	CMinH=36;

type
  TOnGetSectionName=procedure(Sender:TObject;var SectionName:string) of object;
  
	TFormSaver=class(TComponent)
	private
    FOnGetSectionName:TOnGetSectionName;
		FPrivateMode:TPrivateMode;
		FINIFileName:TFileName;
		FSectionName:string;

	public
		constructor Create(Owner:TComponent); override;
		destructor Destroy; override;

	protected
		FParent:TForm;

		procedure Loaded; override;

		function GetSectionName(const Name:string):string;
		function GetFileName(const Name:string):string;

	published
		property INIFileName:TFileName read FINIFileName write FINIFileName;
		property SectionName:string read FSectionName write FSectionName;
		property PrivateMode:TPrivateMode read FPrivateMode write FPrivateMode default pmPublic;
    property OnGetSectionName:TOnGetSectionName read FOnGetSectionName write FOnGetSectionName;
	end;

var WriteXY,WriteWH:Boolean;

implementation

constructor TFormSaver.Create(Owner:TComponent);
begin
	inherited Create(Owner);

  FOnGetSectionName:=nil;
	FINIFileName:='';
	FSectionName:='FormData';
  FPrivateMode:=pmPublic;
end;

destructor TFormSaver.Destroy;
var INIFile:TINIFile;
		secName:string;
begin
	if not (csDesigning in ComponentState) then begin
		if FParent is TForm then begin
			INIFile:=TINIFile.Create(GetFileName(INIFileName));

			try
      	try
          secName:=GetSectionName(SectionName);

          if WriteXY then begin
            INIFile.WriteInteger(secName,'Left',FParent.Left);
            INIFile.WriteInteger(secName,'Top',FParent.Top);
          end;

          if not (FParent.BorderStyle in [bsSingle,bsDialog]) then begin
            if WriteWH then begin
              INIFile.WriteInteger(secName,'Width',FParent.Width);
              INIFile.WriteInteger(secName,'Height',FParent.Height);
            end;
          end;

          INIFile.WriteInteger(secName,'State',Integer(FParent.WindowState));

          except
          	on Exception do begin
            end;
        end;

				finally
					INIFile.Free;
			end;
		end;
	end;

	inherited Destroy;
end;

procedure TFormSaver.Loaded; 
var INIFile:TINIFile;
		W,H,X,Y:Integer;
		secName:string;
begin
	inherited Loaded;

	if not (csDesigning in ComponentState) then begin
		if Owner is TForm then begin
			FParent:=Owner as TForm;

			with FParent do begin
				INIFile:=TINIFile.Create(GetFileName(INIFileName));

				try
					W:=Width;
					H:=Height;

					secName:=GetSectionName(SectionName);

					WindowState:=TWindowState(INIFile.ReadInteger(secName,'State',Integer(WindowState)));

					X:=Min(INIFile.ReadInteger(secName,'Left',Left),Screen.Width-1);
					Y:=Min(INIFile.ReadInteger(secName,'Top',Top),Screen.Height-1);

					if not (FParent.BorderStyle in [bsSingle,bsDialog]) then begin
						W:=INIFile.ReadInteger(secName,'Width',Width);
						H:=INIFile.ReadInteger(secName,'Height',Height);
					end;

          if WindowState<>wsMaximized then begin
            if (W<>Width) or (H<>Height) or (X<>Left) or (Y<>Top) then begin
              SetBounds(X,Y,Max(CMinW,W),Max(CMinH,H));
            end;
          end;

					finally
						INIFile.Free;
				end;
			end;
		end else begin
			FParent:=nil;
		end;
	end;
end;

function TFormSaver.GetSectionName(const Name:string):string;
begin
	Result:=CSecPrefix+Name;

  if Assigned(FOnGetSectionName) then FOnGetSectionName(Self,Result);
end;

function TFormSaver.GetFileName(const Name:string):string;
begin
	Result:=GetPrivatePath(PrivateMode,VirtualApplication.ExeDir,VirtualApplication.GroupDir,VirtualApplication.UserDir,Name)
end;

initialization
	WriteXY:=True;
	WriteWH:=True;
end.