{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: Form for editinig TINIDatabase component
//******************************************************
//******************************************************

unit INIEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, INIUtil, ExtCtrls, FSaver, INISec, ComCtrls, Chkgroup,
  Spin, INIEnt,DsgnIntf,INIFile,Menus;

type
  TfrmEditINIDB = class(TForm)
    Panel1: TPanel;
    pnlButtons: TAutoAlignPanel;
    Panel2: TPanel;
    FormSaver: TFormSaver;
    pgControl: TPageControl;
    tsMain: TTabSheet;
    tsGlyphs: TTabSheet;
    tsDestination: TTabSheet;
    Label1: TLabel;
    cmbDestination: TComboBox;
    Label2: TLabel;
    EFileName: TEdit;
    tcGlyphs: TTabControl;
    Label5: TLabel;
    ECaption: TEdit;
    Label6: TLabel;
    Label7: TLabel;
    ENumGlyphs: TSpinEdit;
    Panel3: TPanel;
    imgGlyph: TImage;
    tsDirs: TTabSheet;
    Label8: TLabel;
    EPublicDir: TEdit;
    SBPublicDir: TSpeedButton;
    Label9: TLabel;
    EGroupDir: TEdit;
    SpeedButton2: TSpeedButton;
    Label10: TLabel;
    EUserDir: TEdit;
    SpeedButton3: TSpeedButton;
    cbRunPath: TCheckBox;
    chkPrivateModes: TCheckGroup;
    Label3: TLabel;
    EUserPL: TSpinEdit;
    Label4: TLabel;
    tsNames: TTabSheet;
    Label11: TLabel;
    EComputerName: TEdit;
    Label12: TLabel;
    EGroupName: TEdit;
    Label13: TLabel;
    EUserName: TEdit;
    INIDB: TINIDatabase;
    chbCreatePageOnChange: TCheckBox;
    Label14: TLabel;
    EDistance: TSpinEdit;
    SpeedButton6: TSpeedButton;
    SpeedButton7: TSpeedButton;
    secDBEdit: TINISection;
    secDBEditTabSheet: TINIEntryString;
    OD: TOpenDialog;
    secDBEditTabIndex: TINIEntryInteger;
    popImage: TPopupMenu;
    POPStretch: TMenuItem;
    N1: TMenuItem;
    Load1: TMenuItem;
    Clear1: TMenuItem;
    secDBEditStretch: TINIEntryBoolean;
    ODGlyph: TOpenDialog;
    secDBEditODDir: TINIEntryString;
    BBExplore: TBitBtn;
    BBOk: TBitBtn;
    BBCancel: TBitBtn;
    BBHelp: TBitBtn;
    cbWindowsPath: TCheckBox;
    cbWindowsUser: TCheckBox;
    cmbParent: TComboBox;
    Label15: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure INIDBBeforeWrite(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SBPublicDirClick(Sender: TObject);
    procedure SpeedButton2Click(Sender: TObject);
    procedure SpeedButton3Click(Sender: TObject);
    procedure BBOkClick(Sender: TObject);
    procedure tcGlyphsChange(Sender: TObject);
    procedure POPStretchClick(Sender: TObject);
    procedure popImagePopup(Sender: TObject);
    procedure secDBEditStretchChange(Sender: TINIField);
    procedure Load1Click(Sender: TObject);
    procedure Clear1Click(Sender: TObject);
    procedure SpeedButton6Click(Sender: TObject);
    procedure SpeedButton7Click(Sender: TObject);
    procedure tcGlyphsChanging(Sender: TObject; var AllowChange: Boolean);
    procedure FormDeactivate(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure BBHelpClick(Sender: TObject);
    procedure BBExploreClick(Sender: TObject);
    procedure cbRunPathClick(Sender: TObject);
    procedure cbWindowsPathClick(Sender: TObject);
    procedure cbWindowsUserClick(Sender: TObject);
    procedure pgControlChange(Sender: TObject);
  private
  	FExplored:Boolean;
  	FSaveHLPFile:string;
  	FDesigner:TFormDesigner;
    FDatabase:TINIDatabase;
    FGlyphs:array[0..2] of TGlyphDescriptor;
    
		procedure SetDirectory(EditC:TEdit);
    
		procedure SetComponentValues;
		procedure FillDestinations;

		procedure SetValues;
		procedure SetTabSheet;
    
		procedure SetGlyphData(Index:Integer);
		procedure RefreshGlyph(Index:Integer);

		procedure LoadGlyph(Index:Integer);
		procedure ClearGlyph(Index:Integer);

    procedure UpdateControls;

    procedure FillCombo;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

  	function Go(Designer:TFormDesigner;INIDatabase:TINIDatabase):Boolean;
  end;

var
  frmEditINIDB: TfrmEditINIDB;

implementation

uses INIDsgn,INIDsgnF;

const CHLPFile='INIDb.HLP';

function TfrmEditINIDB.Go(Designer:TFormDesigner;INIDatabase:TINIDatabase):Boolean;
begin
	FDesigner:=Designer;
  FDatabase:=INIDatabase;

  FillCombo;

	Result:=(ShowModal=mrOK) or FExplored;

  if Result then FDesigner.Modified;
end;

{$R *.DFM}

constructor TfrmEditINIDB.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FGlyphs[0]:=TGlyphDescriptor.Create;
  FGlyphs[1]:=TGlyphDescriptor.Create;
  FGlyphs[2]:=TGlyphDescriptor.Create;

  FExplored:=False;
end;

destructor TfrmEditINIDB.Destroy;
begin
  FGlyphs[0].Free;
  FGlyphs[1].Free;
  FGlyphs[2].Free;

  inherited Destroy;
end;

procedure TfrmEditINIDB.FormCreate(Sender: TObject);
begin
  INIDB.Read;

  FillDestinations;
end;

procedure TfrmEditINIDB.FormDestroy(Sender: TObject);
begin
  INIDB.Write;
end;

procedure TfrmEditINIDB.FillDestinations;
var i:Integer;
begin
	for i:=0 to GetStreamClassesCount-1 do begin
  	cmbDestination.Items.Add(GetStreamClassDescription(i));
  end;
end;

procedure TfrmEditINIDB.INIDBBeforeWrite(Sender: TObject);
begin
	if pgControl.ActivePage<>nil then secDBEditTabSheet.AsString:=pgControl.ActivePage.Caption;

  secDBEditTabIndex.AsInteger:=tcGlyphs.TabIndex;
end;

procedure TfrmEditINIDB.FormShow(Sender: TObject);
begin
  SetTabSheet;
  SetValues;

  RefreshGlyph(tcGlyphs.TabIndex);

  UpdateControls;
end;

procedure TfrmEditINIDB.SetValues;
begin
	cmbDestination.ItemIndex:=cmbDestination.Items.IndexOf(FDatabase.Destination);
  EFileName.Text:=FDatabase.FileName;

  if pmPublic in FDatabase.PrivateModes then chkPrivateModes.CheckItem(0);
  if pmGroupPrivate in FDatabase.PrivateModes then chkPrivateModes.CheckItem(1);
  if pmUserPrivate in FDatabase.PrivateModes then chkPrivateModes.CheckItem(2);

  EUserPL.Value:=FDatabase.UserPrivilegLevel;

  with FDatabase.Names do begin
	  EComputerName.Text:=Computer;
    EGroupName.Text:=Group;
	  EUserName.Text:=User;

    cbWindowsUser.Checked:=UseWindowsUserInstead;
  end;

  with FDatabase.Dirs do begin
	  EPublicDir.Text:=Public;
    EGroupDir.Text:=Group;
	  EUserDir.Text:=User;

    cbRunPath.Checked:=UseRunDirInstead;
    cbWindowsPath.Checked:=UseWindowsDirInstead;
  end;

  with FDatabase.Glyphs do begin
  	FGlyphs[0].GAssign(Save);
  	FGlyphs[1].GAssign(Delete);
  	FGlyphs[2].GAssign(Clear);
  end;

  if FDatabase.Parent<>nil then begin
  	if cmbParent.Items.IndexOf(FDatabase.Parent.Name)<>-1 then cmbParent.ItemIndex:=cmbParent.Items.IndexOf(FDatabase.Parent.Name);
  end else
  	cmbParent.ItemIndex:=0;

  chbCreatePageOnChange.Checked:=FDatabase.CreatePageOnChange;
  EDistance.Value:=FDatabase.Distance;
end;

procedure TfrmEditINIDB.SetComponentValues;
var C:TComponent;
begin
	FDatabase.Destination:=cmbDestination.Text;
  FDatabase.FileName:=EFileName.Text;

  FDatabase.PrivateModes:=[];

  if chkPrivateModes.ItemChecked(0) then FDatabase.PrivateModes:=FDatabase.PrivateModes+[pmPublic];
  if chkPrivateModes.ItemChecked(1) then FDatabase.PrivateModes:=FDatabase.PrivateModes+[pmGroupPrivate];
  if chkPrivateModes.ItemChecked(2) then FDatabase.PrivateModes:=FDatabase.PrivateModes+[pmUserPrivate];

  FDatabase.UserPrivilegLevel:=EUserPL.Value;

  with FDatabase.Names do begin
	  Computer:=EComputerName.Text;
    Group:=EGroupName.Text;
	  User:=EUserName.Text;

    UseWindowsUserInstead:=cbWindowsUser.Checked;
  end;

  with FDatabase.Dirs do begin
	  Public:=EPublicDir.Text;
    Group:=EGroupDir.Text;
	  User:=EUserDir.Text;

    UseRunDirInstead:=cbRunPath.Checked;
    UseWindowsDirInstead:=cbWindowsPath.Checked;
  end;

  with FDatabase.Glyphs do begin
  	Save.GAssign(FGlyphs[0]);
  	Delete.GAssign(FGlyphs[1]);
  	Clear.GAssign(FGlyphs[2]);
  end;

  if (FDesigner<>nil) and (FDesigner.Form<>nil) then begin
  	if cmbParent.Text<>'' then begin
    	C:=FDesigner.Form.FindComponent(cmbParent.Text);

      if C is TWinControl then FDatabase.Parent:=C as TWinControl;
    end else begin
    	FDatabase.Parent:=nil;
    end;
  end;

  FDatabase.CreatePageOnChange:=chbCreatePageOnChange.Checked;
  FDatabase.Distance:=EDistance.Value;
end;

procedure TfrmEditINIDB.SetTabSheet;
var i:Integer;
begin
  tcGlyphs.TabIndex:=secDBEditTabIndex.AsInteger;
  
	for i:=0 to pgControl.ControlCount-1 do begin
  	if pgControl.Pages[i].Caption=secDBEditTabSheet.AsString then begin
    	pgControl.ActivePage:=pgControl.Pages[i];

      Exit;
    end;
  end;
end;

procedure TfrmEditINIDB.SBPublicDirClick(Sender: TObject);
begin
	SetDirectory(EPublicDir);
end;

procedure TfrmEditINIDB.SpeedButton2Click(Sender: TObject);
begin
	SetDirectory(EGroupDir);
end;

procedure TfrmEditINIDB.SpeedButton3Click(Sender: TObject);
begin
	SetDirectory(EUserDir);
end;

procedure TfrmEditINIDB.SetDirectory(EditC:TEdit);
begin
	OD.InitialDir:=EditC.Text;

  if OD.Execute then EditC.Text:=ExtractFilePath(OD.FileName);
end;

procedure TfrmEditINIDB.BBOkClick(Sender: TObject);
begin
	SetGlyphData(tcGlyphs.TabIndex);
  SetComponentValues;

  ModalResult:=mrOK;
end;

procedure TfrmEditINIDB.tcGlyphsChange(Sender: TObject);
begin
	RefreshGlyph(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.RefreshGlyph(Index:Integer);
begin
	ECaption.Text:=FGlyphs[Index].Caption;
	imgGlyph.Picture.Assign(FGlyphs[Index].Glyph);
	ENumGlyphs.Value:=FGlyphs[Index].NumGlyphs;
end;

procedure TfrmEditINIDB.SetGlyphData(Index:Integer);
begin
	FGlyphs[Index].Caption:=ECaption.Text;
	FGlyphs[Index].Glyph.Assign(imgGlyph.Picture);
	FGlyphs[Index].NumGlyphs:=ENumGlyphs.Value;
end;

procedure TfrmEditINIDB.POPStretchClick(Sender: TObject);
begin
	secDBEditStretch.AsBoolean:=not secDBEditStretch.AsBoolean;

  POPStretch.Checked:=secDBEditStretch.AsBoolean;
end;

procedure TfrmEditINIDB.popImagePopup(Sender: TObject);
begin
  POPStretch.Checked:=secDBEditStretch.AsBoolean;
end;

procedure TfrmEditINIDB.secDBEditStretchChange(Sender: TINIField);
begin
	imgGlyph.Stretch:=Sender.AsBoolean;
end;

procedure TfrmEditINIDB.Load1Click(Sender: TObject);
begin
	LoadGlyph(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.Clear1Click(Sender: TObject);
begin
	ClearGlyph(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.SpeedButton6Click(Sender: TObject);
begin
	LoadGlyph(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.SpeedButton7Click(Sender: TObject);
begin
	ClearGlyph(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.LoadGlyph(Index:Integer);
begin
	ODGlyph.InitialDir:=secDBEditODDir.AsString;

	if ODGlyph.Execute then begin
  	if FileExists(ODGlyph.FileName) then begin
    	FGlyphs[Index].Glyph.LoadFromFile(ODGlyph.FileName);

		  RefreshGlyph(Index);

      secDBEditODDir.AsString:=ExtractFilePath(ODGlyph.FileName);
    end;
  end;
end;

procedure TfrmEditINIDB.ClearGlyph(Index:Integer);
begin
{$ifndef VER90}
	FGlyphs[Index].Glyph.Handle:=0;
{$endif}
{$ifdef VER90}
	with FGlyphs[Index].Glyph do begin
  	Width:=0;
    Height:=0;
  end;
{$endif}

  RefreshGlyph(Index);
end;

procedure TfrmEditINIDB.tcGlyphsChanging(Sender: TObject;
  var AllowChange: Boolean);
begin
	SetGlyphData(tcGlyphs.TabIndex);
end;

procedure TfrmEditINIDB.FormDeactivate(Sender: TObject);
begin
	Application.HelpFile:=FSaveHLPFile;
end;

procedure TfrmEditINIDB.FormHide(Sender: TObject);
begin
	Application.HelpFile:=FSaveHLPFile;
end;

procedure TfrmEditINIDB.FormActivate(Sender: TObject);
begin
	FSaveHLPFile:=Application.HelpFile;
  
	Application.HelpFile:=CHLPFile;

  if (frmINIDatabase<>nil) and (frmINIDatabase.Visible) then begin
  	frmINIDatabase.Free;

    frmINIDatabase:=nil;
  end;

  BBExplore.Enabled:=not ((frmINIDatabase<>nil) and (frmINIDatabase.Visible));
end;

procedure TfrmEditINIDB.BBHelpClick(Sender: TObject);
begin
	if pgControl.ActivePage<>nil then
		Application.HelpJump(pgControl.ActivePage.Caption)
  else
		Application.HelpJump('Edit database');
end;

procedure TfrmEditINIDB.BBExploreClick(Sender: TObject);
begin
	ExploreDatabase(FDesigner,FDatabase,True);

  FExplored:=True;
end;


procedure TfrmEditINIDB.cbRunPathClick(Sender: TObject);
begin
	if cbRunPath.Checked then cbWindowsPath.Checked:=False;

  with FDatabase.Dirs do begin
    UseRunDirInstead:=cbRunPath.Checked;

	  EPublicDir.Text:=Public;
    EGroupDir.Text:=Group;
	  EUserDir.Text:=User;
  end;
  
  UpdateControls;
end;

procedure TfrmEditINIDB.cbWindowsPathClick(Sender: TObject);
begin
	if cbWindowsPath.Checked then cbRunPath.Checked:=False;

  with FDatabase.Dirs do begin
    UseWindowsDirInstead:=cbWindowsPath.Checked;

	  EPublicDir.Text:=Public;
    EGroupDir.Text:=Group;
	  EUserDir.Text:=User;
  end;

  UpdateControls;
end;


procedure TfrmEditINIDB.cbWindowsUserClick(Sender: TObject);
begin
  with FDatabase.Names do begin
    UseWindowsUserInstead:=cbWindowsUser.Checked;

	  EUserName.Text:=User;
  end;

  UpdateControls;
end;

procedure TfrmEditINIDB.UpdateControls;
begin
	EUserName.Enabled:=not cbWindowsUser.Checked;
	EPublicDir.Enabled:=not cbWindowsPath.Checked and not cbRunPath.Checked;
  SBPublicDir.Enabled:=EPublicDir.Enabled;
end;

procedure TfrmEditINIDB.FillCombo;
var C:TComponent;
		i:Integer;
begin
	if (FDesigner<>nil) and (FDesigner.Form<>nil) then begin
		C:=FDesigner.Form;

    cmbParent.Items.Add('');

  	for i:=0 to C.ComponentCount-1 do begin
    	if C.Components[i] is TWinControl then cmbParent.Items.Add(C.Components[i].Name);
	  end;
  end;
end;

procedure TfrmEditINIDB.pgControlChange(Sender: TObject);
begin
	UpdateControls;
end;

end.
