{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: "INI" entries
//******************************************************
//******************************************************

unit INIEnt;

interface

uses
	SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
	Forms, Dialogs, DsgnIntf,Menus,ExtCtrls,Spin,Buttons,StdCtrls,
  ComCtrls,INIMis,INISec;

type
	TSelectType=(stComboBox,stRadioGroup,stListBox);

	TINIEntrySeparator=class(TINIEntry)
  private
  	FDouble:Boolean;
    
	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;

	published
	 	property CaptionStyle;

    property Double:Boolean read FDouble write FDouble stored True default False;
	end;

	TINIEntryNextColumn=class(TINIEntry)
	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
	end;

	TINIEntryButton=class(TINIEntry)
	private
		FGlyph:TBitmap;
		FOnClick:TNotifyEvent;
		FW,FH,FNumGlyphs:Integer;
    FShowCaption:Boolean;

		procedure SetGlyph(gBMP:TBitmap);

	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure RefreshControl; override;
		procedure Reset; override;

	published
  	property CaptionStyle;

    property ShowCaption:Boolean read FShowCaption write FShowCaption default False;
		property NumGlyphs:Integer read FNumGlyphs write FNumGlyphs default 1;
		property Glyph:TBitmap read FGlyph write SetGlyph;
		property Width:Integer read FW write FW default 32;
		property Height:Integer read FH write FH default 32;

		property OnClick:TNotifyEvent read FOnClick write FOnClick;
	end;

	TINIEntryPaintBox=class(TINIEntry)
	private
		FOnClick,FOnPaint:TNotifyEvent;
		FW,FH:Integer;

    function GetPaintBox:TPaintBox;
    function GetCanvas:TCanvas;

    procedure SetW(W:Integer);
    procedure SetH(H:Integer);

	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure RefreshControl; override;
		procedure Reset; override;

    property PaintBox:TPaintBox read GetPaintBox;
    property Canvas:TCanvas read GetCanvas;

	published
	 	property CaptionStyle;

		property OnClick:TNotifyEvent read FOnClick write FOnClick;
		property OnPaint:TNotifyEvent read FOnPaint write FOnPaint;
		property Width:Integer read FW write SetW default 32;
		property Height:Integer read FH write SetH default 32;
	end;

	TINIEntryInteger=class(TINIEntryStreamed)
	private
		FIMinValue,FIMaxValue,FIIncrement:LongInt;

    procedure SetIMinValue(IV:LongInt);
    procedure SetIMaxValue(IV:LongInt);
    procedure SetIIncrement(IV:LongInt);

	protected
  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; override;

		function DoGetAsFloat:Double; override;
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;
		function DoGetAsBoolean:Boolean; override;

		procedure DoSetAsFloat(NValue:Double); override;
		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;
		procedure DoSetAsBoolean(NValue:Boolean); override;

		procedure SetValue(NValue:LongInt); virtual;

		function GetCaption:string; override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean; override;
	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

		property Value:LongInt read GetAsInteger write SetAsInteger;

	published
		property IMinValue:LongInt read FIMinValue write SetIMinValue default 0;
		property IMaxValue:LongInt read FIMaxValue write SetIMaxValue default 0;
		property IIncrement:LongInt read FIIncrement write SetIIncrement default 1;

		property DefaultValue:LongInt read GetDefaultAsInteger write SetDefaultAsInteger default 0;
    property ConnectedControl;
	end;

	TINIEntryColor=class(TINIEntryInteger)
	protected
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;

		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;

		procedure SetDefaultValue(NColor:TColor);
		function GetDefaultValue:TColor;

		procedure SetAsColor(NColor:TColor);
		function GetAsColor:TColor;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean; override;
	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;

		property AsColor:TColor read GetAsColor write SetAsColor;

	published
		property DefaultValue:TColor read GetDefaultValue write SetDefaultValue default clWhite;
	end;

	TINIEntryFloat=class(TINIEntryStreamed)
  private
  	FDisplayFormat:string;
    FExitEvent:TNotifyEvent;

    procedure OnEditExit(Sender:TObject);

	protected
  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; override;

		function DoGetAsFloat:Double; override;
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;
		function DoGetAsBoolean:Boolean; override;

		procedure DoSetAsFloat(NValue:Double); override;
		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;
		procedure DoSetAsBoolean(NValue:Boolean); override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean;  override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

		property Value:Double read GetAsFloat write SetAsFloat;

	published
		property DefaultValue:Double read GetDefaultAsFloat write SetDefaultAsFloat;

		property EditMask;
    property ConnectedControl;
    property DisplayFormat:string read FDisplayFormat write FDisplayFormat;
	end;

	TINIEntryString=class(TINIEntryStreamed)
	private
		FLength:Integer;

		procedure SetLength(len:Integer);

	protected
  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; override;

		function DoGetAsFloat:Double; override;
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;
		function DoGetAsBoolean:Boolean; override;

		procedure DoSetAsFloat(NValue:Double); override;
		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;
		procedure DoSetAsBoolean(NValue:Boolean); override;

    procedure DoWriteEntry; override;
    procedure DoReadEntry; override;
    
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean;  override;
    
	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

		property Value:string read GetAsString write SetAsString;

	published
		property DefaultValue:string read GetDefaultAsString write SetDefaultAsString;
		property Length:Integer read FLength write SetLength default 0;

		property EditMask;
    property ConnectedControl;
	end;

	TINIEntryBoolean=class(TINIEntryStreamed)
	protected
  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; override;

		function DoGetAsFloat:Double; override;
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;
		function DoGetAsBoolean:Boolean; override;

		procedure DoSetAsFloat(NValue:Double); override;
		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;
		procedure DoSetAsBoolean(NValue:Boolean); override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean;  override;
    
	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

		property Value:Boolean read GetAsBoolean write SetAsBoolean;
    
	published
		property DefaultValue:Boolean read GetDefaultAsBoolean write SetDefaultAsBoolean default False;
    property ConnectedControl;
	end;

	TINIEntryMenuItem=class(TINIEntryBoolean)
	private
		FMenuItem:TMenuItem;

	protected
		procedure Change; override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;

	published
		property MenuItem:TMenuItem read FMenuItem write FMenuItem;
	end;

	TINIEntryShortCut=class(TINIEntryInteger)
	private
		FMenuItem:TMenuItem;
		FEntryMenuItem:TINIEntryMenuItem;
		FBorrowed:Boolean;

	protected
		procedure Change; override;

		function GetAsShortCut:string;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;

		property AsShortCut:string read GetAsShortCut;

	published
		property DefaultValue:LongInt read GetDefaultAsInteger write SetDefaultAsInteger default 0;
		property MenuItem:TMenuItem read FMenuItem write FMenuItem;
		property EntryMenuItem:TINIEntryMenuItem read FEntryMenuItem write FEntryMenuItem;
		property Borrowed:Boolean read FBorrowed write FBorrowed default False;
	end;

	TINIEntryItems=class(TINIEntryStreamed)
	private
		FItems,FDescriptions:TStringList;
		FConnection:TINICustomMultiSection;

		procedure InternalConnectToMultiSection(MultiSection:TINICustomMultiSection;RefreshList:Boolean);

		function GetCount:Integer;

		function GetItem(Index:Integer):string;

    procedure ItemsChanged(Sender:TObject); virtual;
    procedure DescriptionsChanged(Sender:TObject); virtual;

	protected
    function IsConnected:Boolean;
    function IsDisconnected:Boolean;

    function StoreItems:Boolean; virtual;

		procedure Init; override;

		procedure SetItems(sList:TStringList);
		procedure SetDescriptions(sList:TStringList);

		procedure SyncronizeControls; virtual; abstract;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure ReConnect;
		procedure ConnectToMultiSection(MultiSection:TINICustomMultiSection);
		procedure AssignItems(slItems,slDescs:TStrings);

		procedure InsertItem(const Item,Description:string;Index:Integer);
		procedure AddItem(const Item,Description:string);

		procedure Clear; virtual;

		property Item[Index:Integer]:string read GetItem;
		property Items:TStringList read FItems write SetItems stored StoreItems;
		property Descriptions:TStringList read FDescriptions write SetDescriptions stored StoreItems;
		property Count:Integer read GetCount;
		property Connection:TINICustomMultiSection read FConnection write FConnection;
	end;

	TINIEntryCustomOneOfStrings=class(TINIEntryItems)
	private
  	FDisplayHeight:Integer;
		FItemIndex:LongInt;
		FEditControl:TSelectType;

		function GetDescription:string;

		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;

		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;

    procedure SetDisplayHeight(DH:Integer);

		function DoGetStoreIt:Boolean;  override;

	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		procedure SyncronizeControls; override;

    procedure Loaded; override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;
		procedure Validate; override;

		property Description:string read GetDescription;
		property Value:string read GetAsString write SetAsString;
		property ItemIndex:LongInt read GetAsInteger write SetAsInteger;
    property DisplayHeight:Integer read FDisplayHeight write SetDisplayHeight default 0;

		property EditControl:TSelectType read FEditControl write FEditControl default stComboBox;

  published
		property DefaultValue:string read GetDefaultAsString write SetDefaultAsString;
    property ConnectedControl;
	end;

	TINIEntryOneOfStrings=class(TINIEntryCustomOneOfStrings)
	published
  	property Items;
  	property Descriptions;
    property Connection;
		property EditControl;
    property DisplayHeight;
	end;

	TINIEntryManyOfStrings=class(TINIEntryItems)
	private
  	FDisplayHeight:Integer;
    FBorderStyle:TBorderStyle;

    procedure SetDisplayHeight(DH:Integer);

		function DoGetAsString:string; override;
		procedure DoSetAsString(const NValue:string); override;

		function DoGetStoreIt:Boolean; override;

    procedure SetChecked(Index:Integer;bValue:Boolean);
    function GetChecked(Index:Integer):Boolean;

    procedure DescriptionsChanged(Sender:TObject); override;

	protected
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		procedure SyncronizeControls; override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure RefreshControl; override;
		procedure Reset; override;

    property Checked[Index:Integer]:Boolean read GetChecked write SetChecked;
		property Value:string read GetAsString write SetAsString;

	published
  	property Descriptions;
    property Connection;

    property BorderStyle:TBorderStyle read FBorderStyle write FBorderStyle default bsSingle;
    property DisplayHeight:Integer read FDisplayHeight write SetDisplayHeight default 0;
		property DefaultValue:string read GetDefaultAsString write SetDefaultAsString;
    property ConnectedControl;
	end;

	TINIEntryBinary=class(TINIEntryStreamed)
  private
  	FLines:Integer;
    FScrollBars:TScrollStyle;
    FWordWrap:Boolean;
  	FData:Pointer;
    FDataSize:Integer;

		procedure SetLines(sLines:Integer);

    function AllocBuffer(Size:Integer):Boolean;

    procedure FromTextToBuffer;
    procedure FromBufferToText;

  protected
		function DoGetAsFloat:Double; override;
		function DoGetAsInteger:LongInt; override;
		function DoGetAsString:string; override;
		function DoGetAsBoolean:Boolean; override;

		procedure DoSetAsFloat(NValue:Double); override;
		procedure DoSetAsInteger(NValue:LongInt); override;
		procedure DoSetAsString(const NValue:string); override;
		procedure DoSetAsBoolean(NValue:Boolean); override;

		function GetDefaultWidth:Integer; override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

		function DoGetStoreIt:Boolean; override;

		procedure RefreshControl; override;
	  procedure DoWriteEntry; override;
    procedure DoReadEntry; override;

  public
		constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    procedure SetData(const Buffer;Size:Integer);
    function GetData(var Buffer;MaxSize:Integer):Integer;

  published
  	property Lines:Integer read FLines write SetLines default 2;
    property ScrollBars:TScrollStyle read FScrollBars write FScrollBars default ssNone;
    property WordWrap:Boolean read FWordWrap write FWordWrap stored True default True;
  end;

implementation

uses INICtrl,INIUtil;

{
*******************************************************
	INIEntrySeparator
********************************************************
}

constructor TINIEntrySeparator.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FDouble:=False;
end;

procedure TINIEntrySeparator.RefreshControl;
begin
end;

procedure TINIEntrySeparator.Reset;
begin
end;

function TINIEntrySeparator.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
  	CtrlH:=iIF(FDouble,3,2);

    if CaptionStyle<>csWithout then begin
    	CtrlW:=W;

    	if CaptionStyle<=csLeftBottom then CtrlW:=CtrlW-TextWidth(DisplayLabel);

			Result:=CreateTitledControl(AOwner,TCfgSeparator,WParent,cInfo) as TCfgSeparator;
    end else begin
	    CtrlW:=W;

			Result:=CreateUntitledControl(AOwner,TCfgSeparator,WParent,cInfo) as TCfgSeparator;
    end;
	end;

	(Result as TCfgSeparator).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryNextColumn
********************************************************
}

constructor TINIEntryNextColumn.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);
end;

procedure TINIEntryNextColumn.RefreshControl;
begin
end;

procedure TINIEntryNextColumn.Reset;
begin
end;

function TINIEntryNextColumn.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
  	CtrlH:=0;

    if NextColumn then FirstRow;

		Result:=CreateUntitledControl(AOwner,TCfgSeparator,WParent,cInfo) as TCfgSeparator;

		with Result do begin
			Width:=0;
			Height:=0;
			Visible:=False;
		end;
	end;

	(Result as TCfgSeparator).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryButton
********************************************************
}

procedure TINIEntryButton.SetGlyph(gBMP:TBitmap);
begin
	FGlyph.Assign(gBMP);
end;

function TINIEntryButton.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		CtrlW:=Width;
		CtrlH:=Height;

		Result:=INISec.CreateTitledControl(AOwner,TCfgButton,WParent,cInfo) as TCfgButton;

		with Result as TCfgButton do begin
			Glyph:=Self.Glyph;
			NumGlyphs:=Self.NumGlyphs;
			OnClick:=Self.OnClick;

      if Self.ShowCaption then Caption:=DisplayLabel;
		end;
	end;

	(Result as TCfgButton).INIEntry:=Self;
end;

constructor TINIEntryButton.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FW:=32;
	FH:=32;
  FNumGlyphs:=1;
  FShowCaption:=False;

	FGlyph:=TBitmap.Create;
end;

destructor TINIEntryButton.Destroy;
begin
	FGlyph.Free;

	inherited Destroy;
end;

procedure TINIEntryButton.RefreshControl;
begin
end;

procedure TINIEntryButton.Reset;
begin
end;

{
*******************************************************
	INIEntryPaintBox
********************************************************
}

function TINIEntryPaintBox.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		CtrlW:=Self.FW;
		CtrlH:=Self.FH;

		Result:=INISec.CreateTitledControl(AOwner,TCfgPaintBox,WParent,cInfo) as TCfgPaintBox;

		with Result as TCfgPaintBox do begin
			OnClick:=Self.OnClick;
			OnPaint:=Self.OnPaint;
		end;
	end;

	(Result as TCfgPaintBox).INIEntry:=Self;
end;

constructor TINIEntryPaintBox.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FW:=32;
	FH:=32;
end;

destructor TINIEntryPaintBox.Destroy;
begin
	inherited Destroy;
end;

procedure TINIEntryPaintBox.RefreshControl;
begin
	if Control<>nil then Control.Refresh;
end;

procedure TINIEntryPaintBox.Reset;
begin
end;

function TINIEntryPaintBox.GetPaintBox:TPaintBox;
begin
	Result:=nil;

	if Control<>nil then Result:=Control as TPaintBox;
end;

function TINIEntryPaintBox.GetCanvas:TCanvas;
begin
	Result:=nil;

	if GetPaintBox<>nil then Result:=GetPaintBox.Canvas;
end;

procedure TINIEntryPaintBox.SetW(W:Integer);
begin
	if W>=0 then FW:=W;
end;

procedure TINIEntryPaintBox.SetH(H:Integer);
begin
	if H>=0 then FH:=H;
end;

{
*******************************************************
	INIEntryInteger
********************************************************
}

constructor TINIEntryInteger.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FValue:=0;
	FDefaultValue:=0;
	FIMinValue:=0;
	FIMaxValue:=0;
	FIIncrement:=1;
end;

procedure TINIEntryInteger.SetIMinValue(IV:LongInt);
begin
	FIMinValue:=IV;

  if FDefaultValue<IV then FDefaultValue:=IV; 
end;

procedure TINIEntryInteger.SetIMaxValue(IV:LongInt);
begin
	FIMaxValue:=IV;

  if FDefaultValue>IV then FDefaultValue:=IV;
end;

procedure TINIEntryInteger.SetIIncrement(IV:LongInt);
begin
	FIIncrement:=IV;
end;

procedure TINIEntryInteger.RefreshControl;
begin
	if Control<>nil then (Control as TCfgSpinEdit).INIEntry:=Self;
end;

procedure TINIEntryInteger.Reset;
begin
	AsInteger:=DefaultValue;
end;

function TINIEntryInteger.DoGetAsFloat:Double;
begin
	Result:=FValue;
end;

function TINIEntryInteger.DoGetAsInteger:LongInt;
begin
	Result:=FValue;
end;

function TINIEntryInteger.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryInteger.DoGetAsString:string;
begin
	Result:=IntToStr(FValue);
end;

function TINIEntryInteger.DoGetAsBoolean:Boolean;
begin
	Result:=FValue<>0;
end;

procedure TINIEntryInteger.DoSetAsFloat(NValue:Double);
begin
	SetValue(Trunc(NValue));
end;

procedure TINIEntryInteger.DoSetAsInteger(NValue:LongInt);
begin
	SetValue(NValue);
end;

procedure TINIEntryInteger.DoSetAsString(const NValue:string);
var FValue:LongInt;
begin
	if NValue<>'' then begin
    if AsString<>NValue then begin
      try
        FValue:=StrToInt(NValue);

        except
          on EConvertError do FValue:=IMinValue;
      end;

			SetValue(FValue);
    end;
	end;
end;

procedure TINIEntryInteger.DoSetAsBoolean(NValue:Boolean);
begin
  if NValue then
	  SetValue(IMaxValue)
  else
    SetValue(IMinValue);
end;

procedure TINIEntryInteger.SetValue(NValue:LongInt);
begin
	FValue:=LimitValue(NValue,IMinValue,IMaxValue);
end;

function TINIEntryInteger.GetCaption:string;
begin
	if IMaxValue=IMinValue then
		Result:=DisplayLabel
	else
		Result:=Format(DisplayLabel+#32+'[%d,%d]',[IMinValue,IMaxValue]);
end;

function TINIEntryInteger.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		Result:=INISec.CreateTitledControl(AOwner,TCfgSpinEdit,WParent,cInfo) as TCfgSpinEdit;

		with Result as TCfgSpinEdit do begin
			MinValue:=IMinValue;
			MaxValue:=IMaxValue;
			Increment:=IIncrement;
		end;
	end;

	(Result as TCfgSpinEdit).INIEntry:=Self;
end;

procedure TINIEntryInteger.Syncronize(Direction:TSyncronizeDirection);
begin
  if ConnectedControl is TCustomEdit then with ConnectedControl as TCustomEdit do begin
    if Direction=sdToControl then
      Text:=AsString
    else
      AsString:=Text;
  end;
end;

function TINIEntryInteger.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=inherited IsConnectionToControlSupported(WC) or (WC is TCustomEdit);
end;

{
*******************************************************
	INIEntryColor
********************************************************
}

constructor TINIEntryColor.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FValue:=clWhite;
	FDefaultValue:=clWhite;
end;

procedure TINIEntryColor.RefreshControl;
begin
	if Control<>nil then (Control as TCfgColorEdit).INIEntry:=Self;
end;

procedure TINIEntryColor.Reset;
begin
	AsColor:=DefaultValue;
end;

function TINIEntryColor.DoGetAsInteger:LongInt;
begin
	Result:=FValue;
end;

function TINIEntryColor.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryColor.DoGetAsString:string;
begin
	try
		Result:=ColorToString(FValue);

    except
    	Result:=IntToStr(FValue);
  end;
end;

procedure TINIEntryColor.DoSetAsInteger(NValue:LongInt);
begin
	SetValue(NValue);
end;

procedure TINIEntryColor.DoSetAsString(const NValue:string);
begin
	if AsString<>NValue then begin
    try
      FValue:=StringToColor(NValue);

      except
        on Exception do begin
        end;
    end;
  end;
end;

procedure TINIEntryColor.SetDefaultValue(NColor:TColor);
begin
	FDefaultValue:=LongInt(NColor);
end;

function TINIEntryColor.GetDefaultValue:TColor;
begin
	Result:=TColor(FDefaultValue);
end;

procedure TINIEntryColor.SetAsColor(NColor:TColor);
begin
	if AsColor<>NColor then AsInteger:=LongInt(NColor);
end;

function TINIEntryColor.GetAsColor:TColor;
begin
	Result:=TColor(FValue);
end;

function TINIEntryColor.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
var i:Integer;
		clName:string;
begin
	with cInfo do begin
		Result:=INISec.CreateTitledControl(AOwner,TCfgColorEdit,WParent,cInfo) as TCfgColorEdit;

		for i:=0 to CMaxAppColors-1 do begin
			clName:=ColorToString(CColors[i]);

			(Result as TCfgColorEdit).Items.Add(Copy(clName,3,Length(clName)-2));
		end;
	end;

	(Result as TCfgColorEdit).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryFloat
********************************************************
}

constructor TINIEntryFloat.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FValue:=0;
	FDefaultValue:=0;
  FDisplayFormat:='';
end;

procedure TINIEntryFloat.RefreshControl;
begin
	if Control<>nil then (Control as TCfgMaskEdit).INIEntry:=Self;
end;

procedure TINIEntryFloat.Reset;
begin
	AsFloat:=DefaultValue;
end;

function TINIEntryFloat.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=inherited IsConnectionToControlSupported(WC) or (WC is TCustomEdit);
end;

procedure TINIEntryFloat.Syncronize(Direction:TSyncronizeDirection);
begin
  if ConnectedControl is TCustomEdit then with ConnectedControl as TCustomEdit do begin
    if Direction=sdToControl then
      Text:=AsString
    else
      AsString:=Text;
  end;
end;

function TINIEntryFloat.DoGetAsFloat:Double;
begin
	Result:=FValue;
end;

function TINIEntryFloat.DoGetAsInteger:LongInt;
begin
	Result:=Trunc(FValue);
end;

function TINIEntryFloat.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryFloat.DoGetAsString:string;
begin
	if FDisplayFormat<>'' then
	  Result:=FormatFloat(FDisplayFormat,FValue)
  else
		Result:=FloatToStr(FValue);
end;

function TINIEntryFloat.DoGetAsBoolean:Boolean;
begin
	Result:=FValue<>0;
end;

procedure TINIEntryFloat.DoSetAsFloat(NValue:Double);
begin
	FValue:=NValue;
end;

procedure TINIEntryFloat.DoSetAsInteger(NValue:LongInt);
begin
	FValue:=NValue;
end;

procedure TINIEntryFloat.DoSetAsString(const NValue:string);
begin
	if NValue<>'' then begin
  	if AsString<>NValue then begin
      try
        FValue:=StrToFloat(NValue);

        except
          on EConvertError do FValue:=0;
      end;
    end;
	end;
end;

procedure TINIEntryFloat.DoSetAsBoolean(NValue:Boolean);
begin
  if NValue then
    FValue:=1
  else
    FValue:=0;
end;

procedure TINIEntryFloat.OnEditExit(Sender:TObject);
begin
	if Assigned(FExitEvent) then FExitEvent(Sender);

	(Sender as TCfgMaskEdit).Text:=AsString;
end;

function TINIEntryFloat.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		Result:=INISec.CreateTitledControl(AOwner,TCfgMaskEdit,WParent,cInfo) as TCfgMaskEdit;

		with Result as TCfgMaskEdit do begin
			EditMask:=Self.EditMask;

      FExitEvent:=OnExit;
      OnExit:=OnEditExit;
		end;
	end;

	(Result as TCfgMaskEdit).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryString
********************************************************
}

constructor TINIEntryString.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FValue:='';
	FDefaultValue:='';
	FLength:=0;
end;

procedure TINIEntryString.SetLength(len:Integer);
begin
	FLength:=Max(0,Min(255,len));
end;

procedure TINIEntryString.DoReadEntry;
begin
	with TINICustomSection(Table) do AsString:=INIStream.ReadText(RuntimeSectionName,FieldName,AsString);
end;

procedure TINIEntryString.DoWriteEntry;
begin
	with TINICustomSection(Table) do INIStream.WriteText(RuntimeSectionName,FieldName,AsString);
end;

procedure TINIEntryString.RefreshControl;
begin
	if Control<>nil then (Control as TCfgMaskEdit).INIEntry:=Self;
end;

procedure TINIEntryString.Reset;
begin
	AsString:=DefaultValue;
end;

function TINIEntryString.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=inherited IsConnectionToControlSupported(WC) or (WC is TCustomEdit) or (WC is TComboBox);
end;

procedure TINIEntryString.Syncronize(Direction:TSyncronizeDirection);
begin
  if ConnectedControl is TCustomEdit then with ConnectedControl as TCustomEdit do begin
    if Direction=sdToControl then
      Text:=AsString
    else
      AsString:=Text;
  end else if ConnectedControl is TComboBox then with ConnectedControl as TComboBox do begin
    if Direction=sdToControl then
      Text:=AsString
    else
      AsString:=Text;
  end;
end;

function TINIEntryString.DoGetAsFloat:Double;
begin
	Result:=0;

	if AllTrim(FValue)<>'' then
		try
			Result:=StrToFloat(FValue);

			except
				on EConvertError do Result:=0;
		end;
end;

function TINIEntryString.DoGetAsInteger:LongInt;
begin
	Result:=-1;
end;

function TINIEntryString.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryString.DoGetAsString:string;
begin
	Result:=FValue;
end;

function TINIEntryString.DoGetAsBoolean:Boolean;
var sL:Integer;
begin
	sL:=System.Length(string(FValue));

  if sL>0 then
		Result:=UpCase(string(FValue)[1]) in CTrue
  else
  	Result:=False;
end;

procedure TINIEntryString.DoSetAsFloat(NValue:Double);
begin
	FValue:=FloatToStr(NValue);
end;

procedure TINIEntryString.DoSetAsInteger(NValue:LongInt);
begin
	FValue:=IntToStr(NValue);
end;

procedure TINIEntryString.DoSetAsString(const NValue:string);
begin
	FValue:=NValue;
end;

procedure TINIEntryString.DoSetAsBoolean(NValue:Boolean);
begin
	FValue:=csOnOff[NValue];
end;

function TINIEntryString.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		Result:=INISec.CreateTitledControl(AOwner,TCfgMaskEdit,WParent,cInfo) as TCfgMaskEdit;

		with Result as TCfgMaskEdit do begin
			EditMask:=Self.EditMask;

			if FLength>0 then MaxLength:=FLength;
		end;
	end;

	(Result as TCfgMaskEdit).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryBoolean
********************************************************
}

constructor TINIEntryBoolean.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FValue:=False;
	FDefaultValue:=False;
end;

procedure TINIEntryBoolean.RefreshControl;
begin
	if Control<>nil then (Control as TCfgCheckBox).INIEntry:=Self;
end;

procedure TINIEntryBoolean.Reset;
begin
	AsBoolean:=DefaultValue;
end;

function TINIEntryBoolean.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=inherited IsConnectionToControlSupported(WC) or (WC is TCheckBox);
end;

procedure TINIEntryBoolean.Syncronize(Direction:TSyncronizeDirection);
begin
  if ConnectedControl is TCheckBox then with ConnectedControl as TCheckBox do begin
    if Direction=sdToControl then
      Checked:=AsBoolean
    else
      AsBoolean:=Checked;
  end;
end;

function TINIEntryBoolean.DoGetAsFloat:Double;
begin
	if FValue then
		Result:=1
	else
		Result:=0;
end;

function TINIEntryBoolean.DoGetAsInteger:LongInt;
begin
	if FValue then
		Result:=1
	else
		Result:=0;
end;

function TINIEntryBoolean.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryBoolean.DoGetAsString:string;
begin
	Result:=csOnOff[Boolean(FValue)];
end;

function TINIEntryBoolean.DoGetAsBoolean:Boolean;
begin
	Result:=FValue;
end;

procedure TINIEntryBoolean.DoSetAsFloat(NValue:Double);
begin
	FValue:=NValue<>0;
end;

procedure TINIEntryBoolean.DoSetAsInteger(NValue:LongInt);
begin
	FValue:=NValue<>0;
end;

procedure TINIEntryBoolean.DoSetAsString(const NValue:string);
begin
  if Length(NValue)>0 then
    FValue:=UpCase(NValue[1]) in CTrue
  else
    FValue:=False;
end;

procedure TINIEntryBoolean.DoSetAsBoolean(NValue:Boolean);
begin
	FValue:=NValue;
end;

function TINIEntryBoolean.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
    NoTitle;
    CtrlW:=-1;
    
		Result:=INISec.CreateTitledControl(AOwner,TCfgCheckBox,WParent,cInfo) as TCfgCheckBox;

		with Result as TCfgCheckBox do begin
			Caption:=Title;
		end;
	end;

	(Result as TCfgCheckBox).INIEntry:=Self;
end;

{
*******************************************************
	INIEntryMenuItem
********************************************************
}

procedure TINIEntryMenuItem.Change;
begin
	inherited Change;

	if MenuItem<>nil then MenuItem.Visible:=FValue;
end;

procedure TINIEntryMenuItem.RefreshControl;
begin
	if Control<>nil then (Control as TCfgCheckBoxSC).INIEntry:=Self;
end;

procedure TINIEntryMenuItem.Reset;
begin
	AsBoolean:=DefaultValue;
end;

constructor TINIEntryMenuItem.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FMenuItem:=nil;
end;

function TINIEntryMenuItem.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		NoTitle;
    CtrlW:=-1;

		Result:=INISec.CreateTitledControl(AOwner,TCfgCheckBoxSC,WParent,cInfo) as TCfgCheckBoxSC;

		with Result as TCfgCheckBoxSC do begin
			Caption:=Title;

			if Self.Tag=1 then
				Font.Color:=clWindowText
			else
				Font.Color:=clGrayText;
		end;
	end;

	(Result as TCfgCheckBoxSC).INIEntry:=Self;
end;

procedure TINIEntryMenuItem.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if (Operation=opRemove) and (AComponent=FMenuItem) then FMenuItem:=nil;
end;

{
*******************************************************
	INIEntryShortCut
********************************************************
}

procedure TINIEntryShortCut.Change;
begin
	inherited Change;

	if MenuItem<>nil then MenuItem.ShortCut:=TShortCut(FValue);
end;

function TINIEntryShortCut.GetAsShortCut:string;
begin
	Result:=ShortCutToText(AsInteger);
end;

constructor TINIEntryShortCut.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FMenuItem:=nil;
	FEntryMenuItem:=nil;
	FBorrowed:=False;
end;

procedure TINIEntryShortCut.RefreshControl;
begin
	if Control<>nil then (Control as TCfgShortCutEdit).INIEntry:=Self;
end;

procedure TINIEntryShortCut.Reset;
begin
	AsInteger:=DefaultValue;
end;

function TINIEntryShortCut.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
var ccInfo:TCreateInfo;
		lGetParentCB:TCfgCheckBoxSC;
begin
	if Borrowed then begin
		ccInfo:=cInfo;

		with ccInfo do begin
			lGetParentCB:=EntryMenuItem.Control as TCfgCheckBoxSC;

			x:=x+TextWidth(lGetParentCB.Caption)+16;
			y:=lGetParentCB.Top;

			Result:=CreateUntitledControl(AOwner,TCfgShortCutEdit,WParent,ccInfo) as TCfgShortCutEdit;

			Result.Visible:=False;

			lGetParentCB.BorrowEdit:=Result as TCfgShortCutEdit;
		end;
	end else begin
		Result:=INISec.CreateTitledControl(AOwner,TCfgShortCutEdit,WParent,cInfo) as TCfgShortCutEdit;
	end;

	(Result as TCfgShortCutEdit).INIEntry:=Self;
end;

procedure TINIEntryShortCut.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if (Operation=opRemove) and (AComponent=FMenuItem) then FMenuItem:=nil;
  if (Operation=opRemove) and (AComponent=FEntryMenuItem) then FEntryMenuItem:=nil;
end;

{
*******************************************************
	INIEntryItems
********************************************************
}

constructor TINIEntryItems.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FItems:=TStringList.Create;
	FDescriptions:=TStringList.Create;

  FItems.OnChange:=ItemsChanged;
  FDescriptions.OnChange:=DescriptionsChanged;

	FDefaultValue:='';
  
	FConnection:=nil;
end;

destructor TINIEntryItems.Destroy;
begin
	FItems.Free;
	FDescriptions.Free;

	inherited Destroy;
end;

function TINIEntryItems.IsConnected:Boolean;
begin
	Result:=FConnection<>nil;
end;

function TINIEntryItems.IsDisconnected:Boolean;
begin
	Result:=not IsConnected;
end;

function TINIEntryItems.StoreItems:Boolean;
begin
	Result:=IsDisconnected;
end;

procedure TINIEntryItems.ItemsChanged(Sender:TObject);
begin
end;

procedure TINIEntryItems.DescriptionsChanged(Sender:TObject);
begin
end;

function TINIEntryItems.GetCount:Integer;
begin
	Result:=FItems.Count;
end;

procedure TINIEntryItems.InsertItem(const Item,Description:string;Index:Integer);
begin
	Items.Insert(Index,Item);
	Descriptions.Insert(Index,Description);

  if FDefaultValue='' then FDefaultValue:=Item;
end;

procedure TINIEntryItems.AddItem(const Item,Description:string);
begin
	Items.Add(Item);
	Descriptions.Add(Description);

  if FDefaultValue='' then FDefaultValue:=Item;
end;

procedure TINIEntryItems.Clear;
begin
	Items.Clear;
	Descriptions.Clear;
end;

function TINIEntryItems.GetItem(Index:Integer):string;
begin
	if (Index>=0) and (Index<=Count-1) then begin
		Result:=FItems.Strings[Index];
	end else
  	raise ERangeError.Create('Grid index out of range!');
end;

procedure TINIEntryItems.SetItems(sList:TStringList);
begin
	FItems.Assign(sList);

  if (FDefaultValue='') and (sList.Count>0) then FDefaultValue:=sList[0];

  if FDescriptions.Count=0 then FDescriptions.Assign(sList);
end;

procedure TINIEntryItems.SetDescriptions(sList:TStringList);
begin
	FDescriptions.Assign(sList);

  if (FItems.Count=0) and (sList.Count>0) then Items:=sList;
end;

procedure TINIEntryItems.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if (Operation=opRemove) and (FConnection=AComponent) then FConnection:=nil;
end;

procedure TINIEntryItems.InternalConnectToMultiSection(MultiSection:TINICustomMultiSection;RefreshList:Boolean);
var lList:TStringList;
begin
  if (Items.Count=0) or RefreshList then begin
  	lList:=TStringList.Create;

    try
	    MultiSection.AssignToStrings(lList);

      AssignItems(lList,lList);

      finally
      	lList.Free;
    end;
  end;
end;

procedure TINIEntryItems.ReConnect;
begin
	if FConnection<>nil then ConnectToMultiSection(FConnection);
end;

procedure TINIEntryItems.ConnectToMultiSection(MultiSection:TINICustomMultiSection);
begin
	InternalConnectToMultiSection(MultiSection,True);
end;

procedure TINIEntryItems.AssignItems(slItems,slDescs:TStrings);
begin
	Items.Assign(slItems);
	Descriptions.Assign(slDescs);

	SyncronizeControls;

	Validate;
end;

procedure TINIEntryItems.Init;
begin
	inherited Init;
  
	if FConnection<>nil then InternalConnectToMultiSection(FConnection,True);
end;

{
*******************************************************
	INIEntryOneOfStrings
*******************************************************
}

constructor TINIEntryCustomOneOfStrings.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FItemIndex:=-1;
  FDisplayHeight:=0;
	FValue:='';
	FEditControl:=stComboBox;
end;

procedure TINIEntryCustomOneOfStrings.RefreshControl;
begin
	if Control<>nil then begin
		case FEditControl of
			stComboBox:(Control as TCfgComboBox).INIEntry:=Self;
			stRadioGroup:(Control as TCfgRadioGroup).INIEntry:=Self;
			stListBox:(Control as TCfgListBox).INIEntry:=Self;
		end;
	end;
end;

procedure TINIEntryCustomOneOfStrings.Reset;
begin
	AsString:=DefaultValue;
end;

procedure TINIEntryCustomOneOfStrings.Loaded;
begin
	inherited Loaded;
end;

function TINIEntryCustomOneOfStrings.DoGetAsInteger:LongInt;
begin
	Result:=FItemIndex;
end;

function TINIEntryCustomOneOfStrings.DoGetStoreIt:Boolean;
begin
	Result:=True;

	if (FDefaultValue<>'') or (FConnection=nil) then
  	Result:=FValue<>FDefaultValue
  else if FConnection<>nil then
  	Result:=(FValue<>'') or ((FConnection.KeyCount>0) and (FValue=FConnection.KeyByIndex[0]));
end;

function TINIEntryCustomOneOfStrings.DoGetAsString:string;
begin
	Result:=FValue;
end;

procedure TINIEntryCustomOneOfStrings.DoSetAsInteger(NValue:LongInt);
begin
	if (NValue>=0) and (NValue<=Count-1) and (AsInteger<>NValue) then begin
		FItemIndex:=NValue;
		FValue:=FItems.Strings[NValue];

    Change;
	end;
end;

procedure TINIEntryCustomOneOfStrings.DoSetAsString(const NValue:string);
var ii:Integer;
begin
	if FItems.Count>0 then begin
    ii:=FItems.IndexOf(NValue);

    if ii=-1 then begin
      AsInteger:=0;
    end else begin
      FValue:=NValue;
      FItemIndex:=ii;
      Change;
    end;
  end else begin
  	FValue:=NValue;
    FItemIndex:=-1;
    Change;
  end;
end;

procedure TINIEntryCustomOneOfStrings.SetDisplayHeight(DH:Integer);
begin
	if DH>=0 then FDisplayHeight:=DH;
end;

procedure TINIEntryCustomOneOfStrings.SyncronizeControls;
begin
	if Control<>nil then begin
		case FEditControl of
			stComboBox: (Control as TCfgComboBox).Items.Assign(Descriptions);
			stRadioGroup: (Control as TCfgRadioGroup).Items.Assign(Descriptions);
			stListBox: (Control as TCfgListBox).Items.Assign(Descriptions);
		end;
	end;
end;

function TINIEntryCustomOneOfStrings.GetDescription:string;
begin
	Result:='';

	if (ItemIndex>=0) and (ItemIndex<FDescriptions.Count) then Result:=FDescriptions.Strings[ItemIndex];
end;

procedure TINIEntryCustomOneOfStrings.Validate;
begin
	AsString:=AsString;
end;

function TINIEntryCustomOneOfStrings.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	if FEditControl=stComboBox then begin
    with cInfo do begin
      Result:=INISec.CreateTitledControl(AOwner,TCfgComboBox,WParent,cInfo) as TCfgComboBox;

      with Result as TCfgComboBox do begin
        Items.Assign(Descriptions);
        ItemIndex:=Self.ItemIndex;
      end;
    end;

    (Result as TCfgComboBox).INIEntry:=Self;
  end else begin
    with cInfo do begin
      if FDisplayHeight=0 then
	      CtrlH:=Items.Count*(cHeight+2)+4
      else
	      CtrlH:=FDisplayHeight*(cHeight+2)+4;

      if FEditControl=stListBox then begin
        Result:=INISec.CreateTitledControl(AOwner,TCfgListBox,WParent,cInfo) as TCfgListBox;

        with Result as TCfgListBox do begin
          Items.Assign(Descriptions);
          ItemIndex:=Self.ItemIndex;
          ItemHeight:=cHeight+2;
        end;

	    	(Result as TCfgListBox).INIEntry:=Self;
      end else begin
      	if (Items.Count mod 2)=0 then CtrlH:=CtrlH div 2;

        Result:=INISec.CreateTitledControl(AOwner,TCfgRadioGroup,WParent,cInfo) as TCfgRadioGroup;

        with Result as TCfgRadioGroup do begin
          Items.Assign(Descriptions);
          ItemIndex:=Self.ItemIndex;

          Columns:=iIf((Items.Count mod 2)=0,2,Columns);
        end;

		    (Result as TCfgRadioGroup).INIEntry:=Self;
      end;
    end;
  end;
end;

function TINIEntryManyOfStrings.DoGetAsString:string;
begin
	Result:=FValue;
end;

procedure TINIEntryManyOfStrings.DoSetAsString(const NValue:string);
begin
	FValue:=PadRight('0',Descriptions.Count,Left(NValue,Descriptions.Count));
end;

function TINIEntryManyOfStrings.DoGetStoreIt:Boolean;
begin
	Result:=FValue<>FDefaultValue;
end;

function TINIEntryManyOfStrings.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
var i,Factor:Integer;
begin
  with cInfo do begin
  	Factor:=iIf(FDisplayHeight>0,FDisplayHeight,Descriptions.Count);

    CtrlH:=0;

    Result:=INISec.CreateTitledControl(AOwner,TCfgCheckGroup,WParent,cInfo) as TCfgCheckGroup;

    with Result as TCfgCheckGroup do begin
    	Items.Clear;

    	for i:=0 to Descriptions.Count-1 do Items.Add(Descriptions[i]);

      BorderStyle:=Self.BorderStyle;
      Distance:=ControlDistance;
      AutoScroll:=True;

      if ControlCount>0 then begin
      	Height:=Factor*((Controls[0] as TWinControl).Height+ControlDistance)+(Controls[0] as TWinControl).Height-ControlDistance;

				if (FDisplayHeight>0) and (FDisplayHeight<Descriptions.Count) then
        	Height:=Height-ControlDistance-Integer(BorderStyle=bsNone)*(ControlDistance+2);
      end;
	  end;
  end;

  (Result as TCfgCheckGroup).INIEntry:=Self;
end;

constructor TINIEntryManyOfStrings.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FValue:='';
  FDefaultValue:='';
  FDisplayHeight:=0;
  FBorderStyle:=bsSingle;
end;

procedure TINIEntryManyOfStrings.RefreshControl;
begin
	if Control<>nil then (Control as TCfgCheckGroup).INIEntry:=Self;
end;

procedure TINIEntryManyOfStrings.Reset;
begin
	FValue:=FDefaultValue;
end;

procedure TINIEntryManyOfStrings.DescriptionsChanged(Sender:TObject);
begin
	FValue:=PadRight('0',Descriptions.Count,Left(FValue,Descriptions.Count));
	FDefaultValue:=PadRight('0',Descriptions.Count,Left(FDefaultValue,Descriptions.Count));
end;

function TINIEntryManyOfStrings.GetChecked(Index:Integer):Boolean;
begin
	Result:=False;

	if (Index>=0) and (Index<Descriptions.Count) then begin
  	if Length(AsString)>Index then Result:=AsString[Index+1]<>'0';
  end;
end;

procedure TINIEntryManyOfStrings.SetChecked(Index:Integer;bValue:Boolean);
var lAsString:string;
begin
	if (Index>=0) and (Index<Descriptions.Count) then begin
  	lAsString:=AsString;

  	AsString:=Left(lAsString,Index)+cIf(bValue,'1','0')+Right(lAsString,Length(lAsString)-Index-1);

    Change;
  end;
end;

procedure TINIEntryManyOfStrings.SetDisplayHeight(DH:Integer);
begin
	if DH>=0 then FDisplayHeight:=DH;
end;

procedure TINIEntryManyOfStrings.SyncronizeControls;
begin
	if Control<>nil then (Control as TCfgCheckGroup).Items.Assign(FDescriptions);
end;

{
*******************************************************
	INIEntryBinary
********************************************************
}

procedure TINIEntryBinary.FromTextToBuffer;
var lBuffer:string;
		plBuffer:Pointer;
begin
	try
		lBuffer:=ParseStringOfBytes(FValue,ibHex);

    except
    	lBuffer:='';
  end;

  if Length(lBuffer)>0 then begin
	  if AllocBuffer(Length(lBuffer)) then begin
	  	plBuffer:=@(lBuffer[1]);

		  Move(plBuffer^,FData^,FDataSize);
    end;
  end else begin
  	AllocBuffer(0);
  end;
end;

procedure TINIEntryBinary.FromBufferToText;
var plBuffer:Pointer;
		lBuffer:string;
begin
	if FData<>nil then begin
    SetLength(lBuffer,FDataSize);

    plBuffer:=@(lBuffer[1]);

    Move(FData^,plBuffer^,FDataSize);

		FValue:=DumpMemString(lBuffer);
  end;
end;

function TINIEntryBinary.AllocBuffer(Size:Integer):Boolean;
begin
	Result:=Size>0;

  if FData<>nil then begin
  	FreeMem(FData);

    FData:=nil;
  end;

  if Result then begin
    FDataSize:=Size;

    try
      GetMem(FData,FDataSize);

      except
        Result:=False;
    end;
  end;
end;

procedure TINIEntryBinary.DoReadEntry;
begin
	with TINICustomSection(Table) do begin
	  if AllocBuffer(INIStream.EntrySize(RuntimeSectionName,FieldName)) then begin
			INIStream.ReadBinary(RuntimeSectionName,FieldName,FData^,FDataSize);

			FromBufferToText;
    end;
  end;
end;

procedure TINIEntryBinary.DoWriteEntry;
begin
	FromTextToBuffer;

  if FData<>nil then with TINICustomSection(Table) do INIStream.WriteBinary(RuntimeSectionName,FieldName,FData^,FDataSize);
end;

function TINIEntryBinary.DoGetAsFloat:Double;
begin
	Result:=-1;
end;

function TINIEntryBinary.DoGetAsInteger:LongInt;
begin
	Result:=0;
end;

function TINIEntryBinary.DoGetAsString:string;
begin
	Result:=FValue;
end;

function TINIEntryBinary.DoGetAsBoolean:Boolean;
begin
	Result:=False;
end;

procedure TINIEntryBinary.DoSetAsFloat(NValue:Double);
begin
end;

procedure TINIEntryBinary.DoSetAsInteger(NValue:LongInt);
begin
end;

procedure TINIEntryBinary.DoSetAsString(const NValue:string);
begin
	FValue:=NValue;
end;

procedure TINIEntryBinary.DoSetAsBoolean(NValue:Boolean);
begin
end;

procedure TINIEntryBinary.SetLines(sLines:Integer);
begin
	if sLines>=0 then FLines:=sLines;
end;

procedure TINIEntryBinary.RefreshControl;
begin
	if Control<>nil then (Control as TCfgBinaryEdit).INIEntry:=Self;
end;

function TINIEntryBinary.GetDefaultWidth:Integer;
begin
	Result:=24;
end;

function TINIEntryBinary.DoGetStoreIt:Boolean;
begin
	FromTextToBuffer;

	Result:=(FData<>nil) and (FDataSize>0);
end;

function TINIEntryBinary.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
  	if FLines=0 then
    	CtrlH:=0
    else
	  	CtrlH:=FLines*TextHeight('0')+6;

		Result:=INISec.CreateTitledControl(AOwner,TCfgBinaryEdit,WParent,cInfo) as TCfgBinaryEdit;

   	(Result as TCfgBinaryEdit).ScrollBars:=FScrollBars;
   	(Result as TCfgBinaryEdit).WordWrap:=FWordWrap;
	end;

	(Result as TCfgBinaryEdit).INIEntry:=Self;
end;

constructor TINIEntryBinary.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FData:=nil;
  FDataSize:=0;
  FValue:='';

  FLines:=2;
  FScrollBars:=ssNone;
  FWordWrap:=True;
end;

destructor TINIEntryBinary.Destroy;
begin
  if FData<>nil then FreeMem(FData);
  
	inherited Destroy;
end;

procedure TINIEntryBinary.SetData(const Buffer;Size:Integer);
begin
	if AllocBuffer(Size) then begin
  	Move(Buffer,FData^,FDataSize);

    FromBufferToText;

    Change;
  end;
end;

function TINIEntryBinary.GetData(var Buffer;MaxSize:Integer):Integer;
begin
	Result:=0;
  
	if FData<>nil then begin
  	Result:=Min(MaxSize,FDataSize);

  	Move(FData^,Buffer,Result);
  end;
end;

end.
