{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: "INI" stream utility classes and functions
//******************************************************
//******************************************************

unit INIFile;

interface

uses SysUtils,WinProcs,Classes,IniFiles, Registry,Dialogs;

type THKEY=(hkCLASSES_ROOT,hkCURRENT_USER,hkLOCAL_MACHINE,hkUSERS,hkCURRENT_CONFIG,hkEMPTY);

const CHKEYDescriptions:array[THKEY] of string=(
						'HKEY_CLASSES_ROOT',
            'HKEY_CURRENT_USER',
            'HKEY_LOCAL_MACHINE',
            'HKEY_USERS',
            'HKEY_CURRENT_CONFIG',
            'HKEY_CURRENT_USER'
            );
			CHKEYS:array[THKEY] of HKEY=(
						HKEY_CLASSES_ROOT,
            HKEY_CURRENT_USER,
            HKEY_LOCAL_MACHINE,
            HKEY_USERS,
            HKEY_CURRENT_CONFIG,
            HKEY_CURRENT_USER
            );

	CDefaultFileStream='File';
	CDefaultRegistryStream='Registry';

	CEntrySelected='Selected';
	CEntryLast='Last';
	CEntryPrefix='Ent';

type
  TDestination=string;

	TPrivateMode=(pmPublic,pmGroupPrivate,pmUserPrivate);

	TVirtualApplication=record
		ComputerNo:Integer;
		INIFileName:string;
    UserPrivilegLevel:Integer;
		ExeDir,GroupDir,UserDir,TempDir:string;
    ExeName,GroupName,UserName:string;
	end;
  
	TSubSectionProcedure=procedure(const Section:string;Data:Pointer) of object;

  TRegINIFileExtended=class(TRegINIFile)
  protected
  	function SetKey(Key:HKEY;const Path:string;CanCreate:Boolean):Boolean;
  end;

  TEntryType=(etInteger,etBoolean,etFloat,etString,etBinary,etUnknown);
  TBaseStream=(bsFile,bsRegistry,bsOther);

	TINIStream=class(TObject)
	protected
  	FFileName,FComputerName:string;
		FInitOK:Boolean;

    function GetDescription:string; virtual; abstract;

	public
		constructor CreateStream(const ComputerName,StreamName:string); virtual; abstract;

    class function BaseStream:TBaseStream; virtual; abstract;

		function EntryType(const Section,Entry:string):TEntryType; virtual; abstract;
		function EntrySize(const Section,Entry:string):Integer; virtual; abstract;

		procedure WriteText(const Section,Entry,Value:string); virtual; abstract;
		function ReadText(const Section,Entry,Value:string):string; virtual; abstract;

		procedure WriteString(const Section,Entry,Value:string); virtual; abstract;
		function ReadString(const Section,Entry,Value:string):string; virtual; abstract;

		procedure WriteInteger(const Section,Entry:string;Value:LongInt); virtual; abstract;
		function ReadInteger(const Section,Entry:string;Value:LongInt):LongInt; virtual; abstract;

    function ReadBinary(const Section,Entry:string;var Buffer;MaxSize:Integer):Integer; virtual; abstract;
    procedure WriteBinary(const Section,Entry:string;var Buffer;Size:Integer); virtual; abstract;

    function ReadEntries(const Section:string;StringList:TStringList;var Selected:LongInt):LongInt; virtual; abstract;
    function WriteEntries(const Section:string;StringList:TStringList;Selected:LongInt):LongInt; virtual; abstract;

		procedure DeleteEntry(const Section,Entry:string); virtual; abstract;
		procedure DeleteSection(const Section:string); virtual; abstract;

    procedure ReadSections(const RootSection:string;List:TStringList); virtual; abstract;
    procedure ReadSection(const Section:string;List:TStringList); virtual; abstract;

    procedure CopySection(const FromS,ToS:string); virtual; abstract;

    property FileName:string read FFileName;
    property Description:string read GetDescription;
	end;

  TClassOfINIStream=class of TINIStream;

	TINIStreamFile=class(TINIStream)
	private
    FINIFile:TINIFile;

  protected
    function GetDescription:string; override;

	public
		constructor CreateStream(const ComputerName,StreamName:string); override;

		destructor Destroy; override;

    class function BaseStream:TBaseStream; override;

		function EntryType(const Section,Entry:string):TEntryType; override;
		function EntrySize(const Section,Entry:string):Integer; override;

		procedure WriteText(const Section,Entry,Value:string); override;
		function ReadText(const Section,Entry,Value:string):string; override;

		procedure WriteString(const Section,Entry,Value:string); override;
		function ReadString(const Section,Entry,Value:string):string; override;

		procedure WriteInteger(const Section,Entry:string;Value:LongInt); override;
		function ReadInteger(const Section,Entry:string;Value:LongInt):LongInt; override;

    function ReadBinary(const Section,Entry:string;var Buffer;MaxSize:Integer):Integer; override;
    procedure WriteBinary(const Section,Entry:string;var Buffer;Size:Integer); override;

    function ReadEntries(const Section:string;StringList:TStringList;var Selected:LongInt):LongInt; override;
    function WriteEntries(const Section:string;StringList:TStringList;Selected:LongInt):LongInt; override;

		procedure DeleteEntry(const Section,Entry:string); override;
		procedure DeleteSection(const Section:string); override;

    procedure ReadSections(const RootSection:string;List:TStringList); override;
    procedure ReadSection(const Section:string;List:TStringList); override;

    procedure CopySection(const FromS,ToS:string); override;
	end;

	TINIStreamRegistry=class(TINIStream)
	private
    FRegIniFile:TRegINIFile;
    FRootKey:HKEY;

		function OpenRegistry:TRegINIFile;

    function OpenSection(const Section:string;WriteMode:Boolean):Boolean;
    procedure CloseSection;

  protected
    function GetDescription:string; override;

	public
		constructor CreateStream(const ComputerName,StreamName:string); override;

		destructor Destroy; override;

    class function BaseStream:TBaseStream; override;

		function EntryType(const Section,Entry:string):TEntryType; override;
		function EntrySize(const Section,Entry:string):Integer; override;

		procedure WriteText(const Section,Entry,Value:string); override;
		function ReadText(const Section,Entry,Value:string):string; override;

		procedure WriteString(const Section,Entry,Value:string); override;
		function ReadString(const Section,Entry,Value:string):string; override;

		procedure WriteInteger(const Section,Entry:string;Value:LongInt); override;
		function ReadInteger(const Section,Entry:string;Value:LongInt):LongInt; override;

    function ReadBinary(const Section,Entry:string;var Buffer;MaxSize:Integer):Integer; override;
    procedure WriteBinary(const Section,Entry:string;var Buffer;Size:Integer); override;
    
    function ReadEntries(const Section:string;StringList:TStringList;var Selected:LongInt):LongInt; override;
    function WriteEntries(const Section:string;StringList:TStringList;Selected:LongInt):LongInt; override;

		procedure DeleteEntry(const Section,Entry:string); override;
		procedure DeleteSection(const Section:string); override;

    procedure ReadSections(const RootSection:string;List:TStringList); override;
    procedure ReadSection(const Section:string;List:TStringList); override;

    procedure CopySection(const FromS,ToS:string); override;
	end;

var VirtualApplication:TVirtualApplication;

function GetPrivatePath(PM:TPrivateMode;const EXE,GROUP,USER:string;const Name:string):string;

function FileErase(const Name:string):Boolean;
function FileRename(const fromName,toName:string):Boolean;
function FileSizeByName(const Name:string):LongInt;
function GetTmpFileName(const sDir,sExt:string):string;

procedure WriteLog(const Msg:string);
procedure WriteDateTimeLog(const Msg:string);

procedure WriteLnTextStream(Stream:TStream;const Text:string);
procedure WriteTextStream(Stream:TStream;const Text:string);

implementation

uses Forms,INIMis;

procedure WriteLog(const Msg:string);
var f:Text;
begin
	Assign(f,'LogFile.txt');

  if FileExists('LogFile.txt') then
	  Append(f)
  else
  	Rewrite(f);

  WriteLn(f,Msg);

  Close(f);
end;

procedure WriteDateTimeLog(const Msg:string);
var f:Text;
		lErrorFile:string;
begin
	lErrorFile:=MergePaths(ExtractFilePath(Application.ExeName),'Errors.txt');

	Assign(f,lErrorFile);

  if FileExists(lErrorFile) then
	  Append(f)
  else
  	Rewrite(f);

  WriteLn(f,Format('%s %s -> %s',[TimeToStr(Time),DateToStr(Date),Msg]));

  Close(f);
end;

function FileSizeByName(const Name:string):LongInt;
var f:file of Char;
begin
	Result:=0;

	if not FileExists(Name) then Exit;

	Assign(f,Name);

	try
		Reset(f);
		Result:=FileSize(f);
		Close(f);

		except
			on EInOutError do Result:=0;
	end;
end;

function FileErase(const Name:string):Boolean;
var f:file;
begin
	Result:=True;

	Assign(f,Name);

	try
		Erase(f);

		except
			on EInOutError do begin
				Result:=False;

				Exit;
			end;

			on Exception do raise;
	end;
end;

function FileRename(const fromName,toName:string):Boolean;
var f:file;
begin
	Result:=True;

	Assign(f,fromName);

	try
		Rename(f,toName);
		Close(f);

		except
			on EInOutError do begin
				Result:=False;

				Exit;
			end;

			on Exception do raise;
	end;
end;

function GetTmpFileName(const sDir,sExt:string):string;
const CMaxTries=4;
var i,Tries:Integer;
		OK:Boolean;
begin
	Tries:=0;

	repeat
		Result:='';

		for i:=1 to 8 do begin
			Result:=Result+Chr(RND(65,90));
		end;

		Result:=Result+'.'+sExt;

		OK:=not FileExists(MergePaths(sDir,Result));

		Tries:=Tries+1;
	until OK or (Tries>CMaxTries);

	if not OK then Result:='';
end;

procedure INIDeleteEntry(INIFile:TINIFile;const Section,Entry:string);
var lSec,lEntry,lFile:string;
begin
  lSec:=Section+#0;
  lEntry:=Entry+#0;
  lFile:=INIFile.FileName+#0;

  try
    WritePrivateProfileString(@(lSec[1]),@(lEntry[1]),nil,@(lFile[1]));

    except
      on Exception do begin
      end;
  end;
end;

function INIReadEntries(INIFile:TINIFile;const Section,EntryPrefix:string;StringList:TStringList;var Selected:LongInt):LongInt;
var eNo,eLast:LongInt;
		eLine:string;
begin
	StringList.BeginUpdate;

  try
    eLast:=INIFile.ReadInteger(Section,CEntryLast,0);

    for eNo:=0 to eLast-1 do begin
      eLine:=INIFile.ReadString(Section,Format('%s%d',[EntryPrefix,eNo]),'');

      StringList.Add(eLine);
    end;

    Selected:=INIFile.ReadInteger(Section,CEntrySelected,0);

    if eLast=0 then Selected:=-1;

		Result:=eLast-1;

    finally
    	StringList.EndUpdate;
  end;
end;

function INIWriteEntries(INIFile:TINIFile;const Section,EntryPrefix:string;StringList:TStringList;Selected:LongInt):LongInt;
var i,eNo:LongInt;
		Item:string;
begin
	eNo:=0;

  try
    INIFile.EraseSection(Section);

    eNo:=StringList.Count;

    if eNo>0 then begin
      for i:=0 to eNo-1 do begin
        Item:=StringList.Strings[i];

        INIFile.WriteString(Section,Format('%s%d',[EntryPrefix,i]),#34+Item+#34);
      end;

      if eNo=0 then
	      INIDeleteEntry(INIFile,Section,CEntryLast)
      else
	      INIFile.WriteInteger(Section,CEntryLast,eNo);

      if Selected=0 then
	      INIDeleteEntry(INIFile,Section,CEntrySelected)
      else
	      INIFile.WriteInteger(Section,CEntrySelected,Selected);
    end;

    except
      on Exception do begin
      end;
  end;

	Result:=eNo;
end;

function RegistryReadEntries(RINIFile:TRegINIFile;const Section,EntryPrefix:string;StringList:TStringList;var Selected:LongInt):LongInt;
var eNo,eLast:LongInt;
		Entry,eLine:string;
begin
	StringList.BeginUpdate;

  try
    eLast:=RINIFile.ReadInteger(Section,CEntryLast,0);

    for eNo:=0 to eLast-1 do begin
      Entry:=Format('%s%d',[EntryPrefix,eNo]);
      eLine:=RINIFile.ReadString(Section,Entry,'');

      StringList.Add(eLine);
    end;

    Selected:=RINIFile.ReadInteger(Section,CEntrySelected,0);

    if eLast=0 then Selected:=-1;

    Result:=eLast-1;

    finally
    	StringList.EndUpdate;
  end;
end;

function RegistryWriteEntries(RINIFile:TRegINIFile;const Section,EntryPrefix:string;StringList:TStringList;Selected:LongInt):LongInt;
var i,eNo,oldLast:LongInt;
		Entry,Item:string;
begin
  eNo:=StringList.Count;
  oldLast:=RINIFile.ReadInteger(Section,CEntryLast,0);

  for i:=eNo to oldLast-1 do RINIFile.DeleteValue(Format('%s%d',[EntryPrefix,i]));

  RINIFile.DeleteValue(CEntrySelected);
  RINIFile.DeleteValue(CEntryLast);

  try
    if eNo>0 then begin
      for i:=0 to eNo-1 do begin
        Entry:=Format('%s%d',[EntryPrefix,i]);

        Item:=StringList.Strings[i];

        RINIFile.WriteString(Section,Entry,Item);
      end;

      if eNo=0 then
	      RINIFile.DeleteKey(Section,CEntryLast)
      else
	      RINIFile.WriteInteger(Section,CEntryLast,eNo);

      if Selected=0 then
	      RINIFile.DeleteKey(Section,CEntrySelected)
      else
	      RINIFile.WriteInteger(Section,CEntrySelected,Selected)
    end;

    except
      on Exception do begin
      end;
  end;

  if eNo=0 then RINIFile.EraseSection(Section);

	Result:=eNo;
end;

function TRegINIFileExtended.SetKey(Key:HKEY;const Path:string;CanCreate:Boolean):Boolean;
begin
	ChangeKey(Key,'');
  
  Result:=OpenKey(Path,CanCreate);
end;

constructor TINIStreamFile.CreateStream(const ComputerName,StreamName:string);
begin
	inherited Create;

  FComputerName:=ComputerName;
  FFileName:=StreamName;

 	FINIFile:=TINIFile.Create(StreamName)
end;

destructor TINIStreamFile.Destroy;
begin
 	FINIFile.Free;

	inherited Destroy;
end;

function TINIStreamFile.GetDescription:string;
begin
	Result:=CDefaultFileStream;
end;

class function TINIStreamFile.BaseStream:TBaseStream;
begin
	Result:=bsFile;
end;

function TINIStreamFile.EntryType(const Section,Entry:string):TEntryType;
var lValue:string;
begin
	Result:=etString;

  try
		lValue:=ReadText(Section,Entry,'');

  	if Length(lValue)>0 then begin
    	if lValue[1] in ['0'..'9'] then begin
      	Result:=etInteger;

        if Pos('.',lValue)>0 then Result:=etFloat;
      end;
    end;

    except
    	Result:=etString;
  end;
end;

function TINIStreamFile.EntrySize(const Section,Entry:string):Integer;
begin
	try
	  Result:=ReadInteger(Section,Format('%s@L',[Entry]),0);

    except
    	Result:=0;
	end;
end;

procedure TINIStreamFile.WriteText(const Section,Entry,Value:string);
begin
	FINIFile.WriteString(Section,Entry,#34+Value+#34);
end;

function TINIStreamFile.ReadText(const Section,Entry,Value:string):string;
begin
 	Result:=FINIFile.ReadString(Section,Entry,Value);
end;

procedure TINIStreamFile.WriteString(const Section,Entry,Value:string);
begin
	FINIFile.WriteString(Section,Entry,Value);
end;

function TINIStreamFile.ReadString(const Section,Entry,Value:string):string;
begin
 	Result:=FINIFile.ReadString(Section,Entry,Value);
end;

procedure TINIStreamFile.WriteInteger(const Section,Entry:string;Value:LongInt);
begin
 	FINIFile.WriteInteger(Section,Entry,Value);
end;

function TINIStreamFile.ReadInteger(const Section,Entry:string;Value:LongInt):LongInt;
begin
 	Result:=FINIFile.ReadInteger(Section,Entry,Value);
end;

function TINIStreamFile.ReadBinary(const Section,Entry:string;var Buffer;MaxSize:Integer):Integer;
var lSList:TStringList;
		lSelected:LongInt;
    i:Integer;
    lBuffer:string;
    plBuffer:Pointer;
begin
	lSList:=TStringList.Create;

  try
  	INIReadEntries(FINIFile,Format('%s@%s',[Section,Entry]),'',lSList,lSelected);

    lBuffer:='';

    for i:=0 to lSList.Count-1 do begin
    	lBuffer:=lBuffer+lSList[i]+#32;
    end;

    lBuffer:=ParseStringOfBytes(lBuffer,ibHex);
    lBuffer:=Left(lBuffer,MaxSize);

    Result:=Length(lBuffer);

    plBuffer:=@(lBuffer[1]);

    Move(plBuffer^,Buffer,MaxSize);

    finally
    	lSList.Free;
  end;
end;

procedure TINIStreamFile.WriteBinary(const Section,Entry:string;var Buffer;Size:Integer);
var lSList:TStringList;
    lBuffer:string;
    plBuffer:Pointer;
begin
  SetLength(lBuffer,Size);

  plBuffer:=@(lBuffer[1]);

  Move(Buffer,plBuffer^,Size);

  lBuffer:=DumpMemString(lBuffer);

  lSList:=TStringList.Create;

  try
  	repeat
    	lSList.Add(AllTrim(Left(lBuffer,49)));
      
    	lBuffer:=Copy(lBuffer,50,Length(lBuffer));
    until Length(lBuffer)=0;

  	INIWriteEntries(FINIFile,Format('%s@%s',[Section,Entry]),'',lSList,0);

  	WriteInteger(Section,Format('%s@L',[Entry]),Size);

  	finally
    	lSList.Free;
  end;
end;

function TINIStreamFile.ReadEntries(const Section:string;StringList:TStringList;var Selected:LongInt):LongInt;
begin
 	Result:=INIReadEntries(FINIFile,Section,'',StringList,Selected);
end;

function TINIStreamFile.WriteEntries(const Section:string;StringList:TStringList;Selected:LongInt):LongInt;
begin
 	Result:=INIWriteEntries(FINIFile,Section,'',StringList,Selected);
end;

procedure TINIStreamFile.DeleteEntry(const Section,Entry:string);
begin
	INIDeleteEntry(FINIFile,Section,Entry);
end;

procedure TINIStreamFile.ReadSections(const RootSection:string;List:TStringList);
var i,sLen:Integer;
		rSection:string;
    lList:TStringList;
begin
	sLen:=0;
  
	lList:=TStringList.Create;

  try
    FINIFile.ReadSections(lList);

    if RootSection<>'' then begin
      sLen:=Length(RootSection)+1;

      for i:=lList.Count-1 downto 0 do begin
        if not (StrICmp(RootSection,lList[i]) or StrNICmp(RootSection+'\',lList[i],sLen)) then lList.Delete(i);
      end;
    end;

    for i:=0 to lList.Count-1 do begin
      rSection:=LeftToCh(Copy(lList[i],sLen+1,Length(lList[i])-sLen),'\');

      if List.IndexOf(rSection)=-1 then List.Add(rSection);
    end;

    finally
    	lList.Free;
  end;
end;

procedure TINIStreamFile.ReadSection(const Section:string;List:TStringList);
begin
	FINIFile.ReadSection(Section,List);
end;

procedure TINIStreamFile.CopySection(const FromS,ToS:string);
var Sections:TStringList;
		i,sLen:Integer;
    Sufix:string;

	procedure JustCopySection(const jFromS,jToS:string);
  var Entries:TStringList;
  		i:Integer;
  begin
  	Entries:=TStringList.Create;

    try
	    ReadSection(jFromS,Entries);

      for i:=0 to Entries.Count-1 do WriteText(jToS,Entries[i],ReadString(jFromS,Entries[i],''));

      finally
      	Entries.Free;
    end;
  end;

begin
 	JustCopySection(FromS,ToS);

	sLen:=Length(FromS);

	Sections:=TStringList.Create;

  try
  	FINIFile.ReadSections(Sections);

    for i:=0 to Sections.Count-1 do begin
      if StrNICmp(FromS+'\',Sections[i],sLen+1) then begin
      	Sufix:=Copy(Sections[i],sLen+1+1,Length(Sections[i]));

        JustCopySection(Sections[i],MergePaths(ToS,Sufix));
      end;
    end;

    finally
    	Sections.Free;
  end;
end;

procedure TINIStreamFile.DeleteSection(const Section:string);
var Sections:TStringList;
		i:Integer;

  procedure JustDeleteSection(const jSection:string);
  begin
    FINIFile.EraseSection(jSection);
  end;

begin
	Sections:=TStringList.Create;

  try
  	FINIFile.ReadSections(Sections);

    for i:=0 to Sections.Count-1 do begin
      if StrNICmp(Section+'\',Sections[i],Length(Section)+1) then begin
      	JustDeleteSection(Sections[i]);
      end;
    end;

    finally
    	Sections.Free;
  end;

  JustDeleteSection(Section);
end;

constructor TINIStreamRegistry.CreateStream(const ComputerName,StreamName:string);
begin
	inherited Create;

  FComputerName:=ComputerName;
  FFileName:=StreamName;

 	FRegINIFile:=OpenRegistry;
end;

destructor TINIStreamRegistry.Destroy;
begin
 	FRegINIFile.Free;

	inherited Destroy;
end;

function TINIStreamRegistry.OpenRegistry:TRegINIFile;
var hk:THKEY;

	function MyComputerName:string;
  var LocalComputerName:array[0..128] of Char;
      NameSize:DWord;
  begin
	  NameSize:=127;

    GetComputerName(LocalComputerName,NameSize);

    Result:=StrPas(LocalComputerName);
  end;

	function FindRootKey(const hKey:string):THKEY;
  begin
    if hKey<>'' then begin
      for Result:=Low(THKEY) to High(THKEY) do begin
        if AnsiCompareText(CHKEYDescriptions[Result],hKey)=0 then Exit;
      end;
    end;

    Result:=hkEMPTY;
  end;

begin
  FRootKey:=HKEY_CURRENT_USER;

  hk:=FindRootKey(GetNthToken(FFileName,1,['\']));

  if hk<>hkEMPTY then begin
	  FRootKey:=CHKEYS[hk];

  	FFileName:=RightFromCh(FFileName,'\');
  end;

  Result:=TRegINIFile.Create('');

  if (FComputerName<>'') and (AnsiCompareText(FComputerName,MyComputerName)<>0) then begin
  	Result.RegistryConnect(FComputerName);

	  TRegINIFileExtended(Result).SetKey(FRootKey,FFileName,True);
  end else begin
	  TRegINIFileExtended(Result).SetKey(FRootKey,FFileName,True);
  end;
end;

function TINIStreamRegistry.GetDescription:string;
begin
	Result:=CDefaultRegistryStream;
end;

class function TINIStreamRegistry.BaseStream:TBaseStream;
begin
	Result:=bsRegistry;
end;

function TINIStreamRegistry.EntryType(const Section,Entry:string):TEntryType;
var RegType:TRegDataType;
begin
	Result:=etUnknown;

  if OpenSection(Section,False) then begin
    RegType:=FRegINIFile.GetDataType(Entry);

    case RegType of
    	rdString,rdExpandString: Result:=etString;
      rdBinary: Result:=etBinary;
      rdInteger: Result:=etInteger;
    else
    	Result:=etString;
    end;
  end;

  CloseSection;
end;

function TINIStreamRegistry.EntrySize(const Section,Entry:string):Integer;
begin
	Result:=1;
  
  if OpenSection(Section,False) then begin
    Result:=FRegINIFile.GetDataSize(Entry);
  end;

  CloseSection;
end;

procedure TINIStreamRegistry.WriteText(const Section,Entry,Value:string);
begin
 	FRegINIFile.WriteString(Section,Entry,Value);
end;

function TINIStreamRegistry.ReadText(const Section,Entry,Value:string):string;
begin
	Result:=ReadString(Section,Entry,Value);
end;

procedure TINIStreamRegistry.WriteString(const Section,Entry,Value:string);
begin
 	FRegINIFile.WriteString(Section,Entry,Value);
end;

function TINIStreamRegistry.ReadString(const Section,Entry,Value:string):string;
begin
	try
	 	Result:=FRegINIFile.ReadString(Section,Entry,Value);

    except
    	on Exception do Result:=Value;
  end;
end;

procedure TINIStreamRegistry.WriteInteger(const Section,Entry:string;Value:LongInt);
begin
 	FRegINIFile.WriteInteger(Section,Entry,Value);
end;

function TINIStreamRegistry.ReadInteger(const Section,Entry:string;Value:LongInt):LongInt;
begin
	try
	 	Result:=FRegINIFile.ReadInteger(Section,Entry,Value);

    except
    	on Exception do Result:=Value;
  end;
end;

function TINIStreamRegistry.ReadBinary(const Section,Entry:string;var Buffer;MaxSize:Integer):Integer;
begin
	Result:=0;
  
  if OpenSection(Section,False) then begin
		Result:=FRegINIFile.ReadBinaryData(Entry,Buffer,MaxSize);
  end;

 	CloseSection;
end;

procedure TINIStreamRegistry.WriteBinary(const Section,Entry:string;var Buffer;Size:Integer);
begin
  if OpenSection(Section,True) then begin
		FRegINIFile.WriteBinaryData(Entry,Buffer,Size);
  end;

 	CloseSection;
end;

function TINIStreamRegistry.ReadEntries(const Section:string;StringList:TStringList;var Selected:LongInt):LongInt;
begin
	Result:=0;

  if OpenSection(Section,False) then begin
    Result:=RegistryReadEntries(FRegINIFile,'','',StringList,Selected);
  end;

  CloseSection;
end;

function TINIStreamRegistry.WriteEntries(const Section:string;StringList:TStringList;Selected:LongInt):LongInt;
begin
	Result:=0;

	if OpenSection(Section,True) then begin
    Result:=RegistryWriteEntries(FRegINIFile,'','',StringList,Selected);
  end;

  CloseSection;
end;

procedure TINIStreamRegistry.DeleteEntry(const Section,Entry:string);
begin
	if OpenSection(Section,False) then begin
		FRegINIFile.DeleteValue(Entry);
  end;

 	CloseSection;
end;

procedure TINIStreamRegistry.DeleteSection(const Section:string);
begin
 	FRegINIFile.EraseSection(Section);
end;

procedure TINIStreamRegistry.ReadSections(const RootSection:string;List:TStringList);
begin
	if OpenSection(RootSection,False) then begin
		FRegINIFile.ReadSections(List);
  end;

 	CloseSection;
end;

procedure TINIStreamRegistry.ReadSection(const Section:string;List:TStringList);
begin
	FRegINIFile.ReadSection(Section,List);
end;

procedure TINIStreamRegistry.CopySection(const FromS,ToS:string);
begin
	FRegINIFile.MoveKey(FromS,ToS,False);
end;

function TINIStreamRegistry.OpenSection(const Section:string;WriteMode:Boolean):Boolean;
begin
  Result:=TRegINIFileExtended(FRegINIFile).SetKey(FRootKey,MergePaths(FFileName,Section),WriteMode);
end;

procedure TINIStreamRegistry.CloseSection;
begin
	TRegINIFileExtended(FRegINIFile).SetKey(FRootKey,FFileName,False);
end;

procedure GetExePath;
begin
	with VirtualApplication do begin
    ExeDir:=ExtractFilePath(ParamStr(0));

    GroupDir:=ExeDir;
    UserDir:=ExeDir;
    TempDir:=ExeDir;
  end;
end;

function GetPrivatePath(PM:TPrivateMode;const EXE,GROUP,USER:string;const Name:string):string;
var Relative:Boolean;
begin
	Result:=Name;

  Relative:=not (((Length(Result)>0) and (Result[1]='\')) or ((Length(Result)>1) and (Result[1+1]=':')));

  if Relative then case PM of
		pmPublic: Result:=MergePaths(EXE,Result);
		pmGroupPrivate: Result:=MergePaths(GROUP,Result);
		pmUserPrivate: Result:=MergePaths(USER,Result);
	end;
end;

procedure WriteLnTextStream(Stream:TStream;const Text:string);
begin
	WriteTextStream(Stream,Text+#13+#10);
end;

procedure WriteTextStream(Stream:TStream;const Text:string);
var Buffer:PChar;
begin
	Buffer:=@(Text[1]);

	Stream.WriteBuffer(Buffer^,Length(Text));
end;

initialization
	with VirtualApplication do begin
	  UserPrivilegLevel:=0;

    ExeName:=ExtractFileName(Application.ExeName);
    GroupName:='DefaultGroup';
    UserName:='DefaultUser';
  end;

  GetExePath;
end.
