{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: The core unit. Defines main types and classes.
//******************************************************
//******************************************************

unit INISec;

interface

uses
	SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
	Forms, Dialogs, INIMis,ExtCtrls,StdCtrls,ComCtrls,INIFile,INIUtil,
  Buttons,DsgnIntf,CommCtrl;

const
	CCDistance=2;
  CMaxEntryImageIndex=14;
  CDefDisplayWidth=8;
	CMaxPrivilegLevel=128;
  
  CUntitled='Untitled';
  CDefGroup=CUntitled;

  CResStreamNotRegistered=20101;
  CResStreamNotAssigned=20102;
  CResINIDatabaseMissing=20103;
  CResFieldNotFound=20104;
  CResDuplicateFieldName=20105;
  CResReadOnlyMode=20106;
  CResSectionNotFound=20107;
  CResPVFCalled=20108;
  CResSectionIndexOutOfRange=20109;
  CResCannotRegisterStream=20111;
  CResCannotRegisterSection=20112;
  CResCannotRegisterEntry=20113;

type
	EINIError=class(Exception);
  EINIStackError=class(Exception);

  TStreamInfo=record
  	Stream:TStream;
    Level:Integer;
  end;

	TPrivateModes=set of TPrivateMode;

  TMultiSectionButton=(isSave,isClear,isDelete,isChange,isExchange);
  TEntryImageIndex=(eiiBoolean,eiiButton,eiiColor,eiiFloat,eiiInteger,eiiNextColumn,
  								  eiiPaintBox,eiiSeparator,eiiString,eiiCombo,eiiCheckGroup,eiiPrinter,
                    eiiFont,eiiDate,eiiBinary);

  TCaptionStyle=(csLeftTop,csLeftCenter,csLeftBottom,csTopLeft,csTopCenter,csTopRight,csWithout);
  TCreateDirection=(cdTopToBottom,cdLeftToRight);

  TINIField=class;
  TINIDatabase=class;

  TForEachProcedure=procedure(Item:TObject;Data:Pointer) of object;

  TINIFieldNotifyEvent=procedure(Sender:TINIField) of object;
	TStoreItNotifyEvent=procedure(Sender:TINIField;var Store:Boolean) of object;
	TButtonClickEvent=procedure(Sender:TObject;Button:TMultiSectionButton) of object;

  TINITable=class(TComponent)
  private
    FFields:TList;
    FTableNo:Integer;
    FModified,FReadOnly:Boolean;
    FFileName,FCopyFromFileName:string;
    FDestination:TDestination;
		FParentDestination:Boolean;
    FParentField:TINIField;
		FPrivateMode:TPrivateMode;
    FDatabase:TINIDatabase;
  	FINIStream:TINIStream;
    FDirection:TCreateDirection;
    FOnFieldChange:TINIFieldNotifyEvent;

		procedure AdjustParentDestination;

    procedure SetPrivateMode(PM:TPrivateMode);
    procedure SetFileName(const FName:string);
    procedure SetDestination(DT:TDestination);
    procedure SetParentDestination(PD:Boolean);

    function StoredIfNotParentDestination:Boolean;
    function IsDestinationStored:Boolean;

		function GetINIStream:TINIStream;
		function GetReadOnly:Boolean;

    function IsSon:Boolean;

    function FieldByNumber(FieldNo:Integer):TINIField;
    function GetField(Index:Integer):TINIField;
    function GetFieldCount:Integer;

    procedure DestroyFields;

    procedure AddField(Field:TINIField);
    procedure RemoveField(Field:TINIField);
    procedure SetField(Index:Integer;Value:TINIField);

    function GetDestination:TDestination;

    function GetFileName:string;
    function GetEditOnlyMode:Boolean;

    procedure SetTableParent(AParent:TComponent);
    function GetTableParent:TComponent;

    function GetRuntimeFileName:string;
    function GetRuntimePrivateMode:TPrivateMode;
    function GetRuntimeDestination:TDestination;

    property EditOnlyMode:Boolean read GetEditOnlyMode;

    property RuntimeFileName:string read GetRuntimeFileName;
    property RuntimePrivateMode:TPrivateMode read GetRuntimePrivateMode;
    property RuntimeDestination:TDestination read GetRuntimeDestination;

  protected
    procedure Loaded; override;

  	function Connected:Boolean;

    procedure AnyFieldChange(Field:TINIField);

		procedure CheckCanModify;
		procedure CheckFieldName(const FieldName:string;Field:TINIField);

    procedure SetName(const Value:TComponentName); override;
    procedure GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif}); override;
    procedure SetChildOrder(Component:TComponent;Order:Integer); override;

		procedure ReadState(Reader:TReader); override;
    procedure SetParentComponent(AParent:TComponent); override;
    function GetParentComponent:TComponent; override;
    function HasParent:Boolean; override;

    procedure SetDatabase(DB:TINIDatabase); virtual;
    function GetDatabase:TINIDatabase; virtual;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    function FieldByName(const FieldName:string):TINIField;
    function FindField(const FieldName:string):TINIField;
    procedure GetFieldNames(List:TStrings);

    property INIStream:TINIStream read GetINIStream;
    property Database:TINIDatabase read GetDatabase write SetDatabase;
    property FieldCount:Integer read GetFieldCount;
    property Fields[Index:Integer]:TINIField read GetField write SetField;
    property Modified:Boolean read FModified;
    property Parent:TComponent read GetTableParent write SetTableParent;
    property TableNo:Integer read FTableNo;
		property Son:Boolean read IsSon;
		property ParentDestination:Boolean read FParentDestination write SetParentDestination default True;
		property PrivateMode:TPrivateMode read FPrivateMode write SetPrivateMode default pmPublic;

  published
    property ReadOnly:Boolean read GetReadOnly write FReadOnly stored True default False;
  	property FileName:string read GetFileName write SetFileName stored StoredIfNotParentDestination;
    property Destination:TDestination read GetDestination write SetDestination stored IsDestinationStored;
    
    property OnFieldChange:TINIFieldNotifyEvent read FOnFieldChange write FOnFieldChange;
  end;

  { TGroupInfo }

  TGroupInfo=class(TObject)
  private
  	FValue:Integer;

    function GetImageIndex:Integer;
    function GetColumns:Integer;
    function GetDirection:TCreateDirection;

    procedure SetImageIndex(II:Integer);
    procedure SetColumns(sColumns:Integer);
    procedure SetDirection(sDirection:TCreateDirection);

  public
  	constructor Create(Value:Integer);
  	constructor CreateInfo(sColumns,II:Integer;sDirection:TCreateDirection);

    property ImageIndex:Integer read GetImageIndex write SetImageIndex;
    property Columns:Integer read GetColumns write SetColumns;
    property Direction:TCreateDirection read GetDirection write SetDirection;
  end;

{ TINIField }

  TINIField = class(TComponent)
  private
    FTable:TINITable;
    FFieldName:string;
    FReadOnly:Boolean;
    FCalculated:Boolean;
    FVisible:Boolean;
    FFieldNo:Integer;
    FDisplayLabel:string;
    FEditMask:string;
    FOnChange,FOnFieldNameChange:TINIFieldNotifyEvent;

    function GetCalculated:Boolean;
    function GetDisplayLabel:string;
    function GetDisplayWidth:Integer;
    function GetIndex:Integer;
    function IsDisplayLabelStored:Boolean;
    function IsDisplayWidthStored:Boolean;

    procedure SetCalculated(Value:Boolean);
    procedure SetTable(ATable:TINITable);

    procedure SetDisplayLabel(Value:string);
    procedure SetDisplayWidth(Value:Integer);

    procedure SetEditMask(const Value:string);
    procedure SetFieldName(const Value:string);

    procedure SetIndex(Value:Integer);
    procedure SetVisible(Value:Boolean);

  protected
    FValue,FDefaultValue:Variant;
    FDisplayWidth:Integer;
    FIsDirty:Boolean;

    procedure Change; virtual;

		procedure DoOnFieldNameChange; virtual;
    procedure DoOnConnected; virtual;
    procedure DoOnDisconnected; virtual;

    function DoGetAsBoolean:Boolean; virtual; abstract;
    function DoGetAsFloat:Double; virtual; abstract;
    function DoGetAsInteger:LongInt; virtual; abstract;
    function DoGetAsString:string; virtual; abstract;

    procedure DoSetAsBoolean(Value:Boolean); virtual; abstract;
    procedure DoSetAsFloat(Value:Double); virtual; abstract;
    procedure DoSetAsInteger(Value:LongInt); virtual; abstract;
    procedure DoSetAsString(const Value:string); virtual; abstract;

    function GetAsBoolean:Boolean;
    function GetAsFloat:Double;
    function GetAsInteger:LongInt;
    function GetAsString:string;

    procedure SetAsBoolean(Value:Boolean);
    procedure SetAsFloat(Value:Double);
    procedure SetAsInteger(Value:LongInt);
    procedure SetAsString(const Value:string);

    function IsDirty:Boolean; virtual;

    function GetCanModify:Boolean;
    function GetDefaultWidth:Integer; virtual;

    function GetDefaultAsBoolean:Boolean; virtual;
    function GetDefaultAsFloat:Double; virtual;
    function GetDefaultAsInteger:LongInt; virtual;
    function GetDefaultAsString:string; virtual;

    procedure SetDefaultAsBoolean(Value:Boolean); virtual;
    procedure SetDefaultAsFloat(Value:Double); virtual;
    procedure SetDefaultAsInteger(Value:LongInt); virtual;
    procedure SetDefaultAsString(const Value:string); virtual;

    procedure GetTableChildren(Proc:TGetChildProc); virtual;

		procedure ReadState(Reader:TReader); override;
    procedure SetParentComponent(AParent:TComponent); override;
    function GetParentComponent:TComponent; override;
    function HasParent:Boolean; override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    property DefaultAsBoolean:Boolean read GetDefaultAsBoolean write SetDefaultAsBoolean;
    property DefaultAsFloat:Double read GetDefaultAsFloat write SetDefaultAsFloat;
    property DefaultAsInteger:LongInt read GetDefaultAsInteger write SetDefaultAsInteger;
    property DefaultAsString:string read GetDefaultAsString write SetDefaultAsString;

    property AsBoolean:Boolean read GetAsBoolean write SetAsBoolean;
    property AsFloat:Double read GetAsFloat write SetAsFloat;
    property AsInteger:LongInt read GetAsInteger write SetAsInteger;
    property AsString:string read GetAsString write SetAsString;
    property CanModify:Boolean read GetCanModify;
    property EditMask:string read FEditMask write SetEditMask;
    property FieldNo:Integer read FFieldNo;
    property Table:TINITable read FTable write SetTable;

  public
    property Calculated:Boolean read GetCalculated write SetCalculated default False;
    property ReadOnly:Boolean read FReadOnly write FReadOnly default False;
    property OnChange:TINIFieldNotifyEvent read FOnChange write FOnChange;
    property OnFieldNameChange:TINIFieldNotifyEvent read FOnFieldNameChange write FOnFieldNameChange;
    property DisplayWidth:Integer read GetDisplayWidth write SetDisplayWidth stored IsDisplayWidthStored;

  published
    property DisplayLabel:string read GetDisplayLabel write SetDisplayLabel stored IsDisplayLabelStored;
    property FieldName:string read FFieldName write SetFieldName;
    property Index:Integer read GetIndex write SetIndex stored False;
    property Visible:Boolean read FVisible write SetVisible default True;
  end;

	TFControl=class(TControl);

  TCompressKeyMode=(ciYes,ciNo,ciExcludeBlanks);
	TGroupType=(gtTile,gtCascade);

  TCreateInfoState=record
  	Title:string;
  	X,Y,CtrlW,CtrlH,MaxCtrlH,fcCol,fcRow,cCol,cRow:Integer;
  end;

	TCreateInfo=class(TObject)
  private
		fTitle:string;
    fCaptionStyle:TCaptionStyle;
		fX,fY,fW,fH:Integer;
		fCtrlW,fCtrlH:Integer;
    fMaxCtrlH:Integer;
		fCalculated,fRO,fTitled:Boolean;
		fCols,fcCol,fcRow,fcWidth,fcHeight:Integer;
		fXBor,fYBor,fInitX,fInitY,fControlDistance:Integer;
    fDirection:TCreateDirection;
		fCanvas:TCanvas;
    fNeighborhoodControl:TControl;

    procedure SetCurrentColumn(cColumn:Integer);

    function GetState:TCreateInfoState;
    procedure SetState(State:TCreateInfoState);

  public
		constructor Create(CI:TCreateInfo);

    function TextWidth(const Msg:string):Integer;
    function TextHeight(const Msg:string):Integer;

  	procedure FirstColumn;
  	procedure FirstRow;

  	function NextColumn:Boolean;
  	function NextRow:Boolean;

  	procedure NoTitle;

  	property Title:string read fTitle;
		property X:Integer read fX write fX;
		property Y:Integer read fY write fY;
		property W:Integer read fW;
		property H:Integer read fH;
    property CtrlW:Integer read fCtrlW write fCtrlW;
    property CtrlH:Integer read fCtrlH write fCtrlH;
    property Calculated:Boolean read fCalculated;
    property RO:Boolean read fRO;
    property Direction:TCreateDirection read fDirection;
    property Cols:Integer read fCols;
    property cCol:Integer read fcCol;
    property cRow:Integer read fcRow;
    property cWidth:Integer read fcWidth;
    property cHeight:Integer read fcHeight;
    property XBor:Integer read fXBor;
    property YBor:Integer read fYBor;
    property InitX:Integer read fInitX;
    property InitY:Integer read fInitY;
    property CaptionStyle:TCaptionStyle read fCaptionStyle;
    property ControlDistance:Integer read fControlDistance;
    property Titled:Boolean read fTitled;
    property NeighborhoodControl:TControl read fNeighborhoodControl;
	end;

  TSectionState=(ssLoading,ssReading,ssWriting,ssDeleting,ssEdit);
  TSectionStates=set of TSectionState;
  
  TSyncronizeDirection=(sdToControl,sdToEntry);
  TGroupTitlePosition=(gtpTop,gtpBottom);

	TINICustomSection=class(TINITable)
	private
  	FOnCalcFields,FOnInit,FOnBeforeEdit,FOnAfterEdit,FOnAfterRead,FOnBeforeWrite:TNotifyEvent;
		FPrivilegLevel:Integer;
		FRows:Integer;
		FGroups:TStringList;
		FSectionName,FPrefix:string;
    FHeight:Integer;
		FGroupType:TGroupType;
		FPanel:TPanel;
    FAlwaysShowGroup:Boolean;
    FSectionState:TSectionStates;
    FImageIndex:Integer;
    FGroupTitlePosition:TGroupTitlePosition;

		procedure ArrangeIfVScroll(WControl:TWinControl);

    procedure SetPanel(pnl:TPanel);

    procedure FEPRefresh(Item:TObject;Data:Pointer);
    procedure FEPClear(Item:TObject;Data:Pointer);
    procedure FEPInit(Item:TObject;Data:Pointer);
    procedure FEPReset(Item:TObject;Data:Pointer);
    procedure FEPShow(Item:TObject;Data:Pointer);
    procedure FEPEnable(Item:TObject;Data:Pointer);
    procedure FEPSync(Item:TObject;Data:Pointer);

    procedure FEPWriteHumanReadable(Item:TObject;Data:Pointer);
    procedure FEPReadHumanReadable(Item:TObject;Data:Pointer);    

    procedure FEPWrite(Item:TObject;Data:Pointer);
    procedure FEPRead(Item:TObject;Data:Pointer);
    procedure FEPCopy(Item:TObject;Data:Pointer);
    procedure FEPDelete(Item:TObject;Data:Pointer);

    procedure SetHeight(h:Integer);
    procedure SetPrivilegLevel(Level:Integer);

    procedure SetImageIndex(II:Integer);

		procedure SetRows(Row:Integer);
		procedure SetGroups(gList:TStringList);
		procedure SetColumnsInGroup(Index:Integer;Clmn:Integer);
		procedure SetGroupImageIndex(Index:Integer;II:Integer);
		procedure SetDirectionInGroup(Index:Integer;CD:TCreateDirection);
    procedure SetSectionName(const sName:string);

		function GetGroupCount:Integer;
		function GetGroup(Index:Integer):string;
		function GetColumnsInGroup(Index:Integer):Integer;
		function GetGroupImageIndex(Index:Integer):Integer;
		function GetDirectionInGroup(Index:Integer):TCreateDirection;

		function GetPrivatePrefix:string;
		function GetValidFileName:string;

		function GetRuntimeSectionName:string; virtual;

		procedure PreparePrefix;

		procedure ForEachEntry(FEP:TForEachProcedure;Data:Pointer);
		procedure ForEachEntryStreamed(FEP:TForEachProcedure;Data:Pointer);

    property Panel:TPanel read FPanel write SetPanel;

		function CreateTileGroups(AOwner:TComponent;WParent:TAutoAlignPanel):Integer;
		function CreateCascadeGroups(AOwner:TComponent;WParent:TAutoAlignPanel):Integer;
		function CountVisibleControls(const GroupName:string):Integer;

	protected
		procedure Loaded; override;

    procedure OpenLocalGroupForEdit(const OpenGroup:string); virtual;

    procedure WriteGroups(Writer:TWriter);
    procedure DefineProperties(Filer:TFiler); override;
    procedure ReadGroups(Reader:TReader);

		function CreateControls(AOwner:TComponent;WParent:TWinControl;GroupNo:Integer;const GroupName:string):Integer; virtual;
		function InternalCreatePage(AOwner:TComponent;WParent:TWinControl;var NoCreated:Integer):TControl; virtual;

		procedure EnablePanel(State:Boolean);

		procedure Enable(State:Boolean);
		procedure Show(Show:Boolean);

		procedure ReadState(Reader:TReader); override;
    procedure SetParentComponent(AParent:TComponent); override;

    procedure SetDatabase(DB:TINIDatabase); override;

		procedure DoWriteHumanReadable(Stream:TStream;Level:Integer); virtual;
		procedure DoReadHumanReadable(Stream:TStream;Level:Integer); virtual;

		procedure DoWriteSection; virtual;
 		procedure DoReadSection; virtual;
		procedure DoDeleteSection; virtual;
		procedure DoCopySection(const CopyFrom:string); virtual;

    function StoreGroupInfo:Boolean; virtual;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure WriteSection;
		procedure ReadSection;
		procedure DeleteSection;
		procedure CopySection(const CopyFrom:string);

		procedure ShowControls; virtual;
		procedure HideControls; virtual;
		procedure EnableControls; virtual;
		procedure DisableControls; virtual;
		procedure RefreshControls; virtual;
		procedure Init; virtual;
		procedure Reset; virtual;
		procedure ClearControls; virtual;
		procedure Syncronize(Direction:TSyncronizeDirection); virtual;

    procedure RenameGroup(const oldName,newName:string);
    procedure RenameGroupByIndex(Index:Integer;const newName:string);

    procedure OpenGroupForEdit(const OpenGroup:string);

		function FieldsInGroup(const gName:string):Integer;
		function GetGroupIndex(const gName:string):Integer;
		function InsertGroup(const gName:string;Columns,ImageIndex:Integer;Direction:TCreateDirection;Index:Integer):Boolean;
    function DeleteGroup(const gName:string):Boolean;
    function ExchangeGroups(const gNameF,gNameL:string):Boolean;

    function FindSection(const sName:string):TINICustomSection;
    function SectionByName(const sName:string):TINICustomSection;

		property RuntimeSectionName:string read GetRuntimeSectionName;
    property State:TSectionStates read FSectionState write FSectionState;

    property GroupExpandable:Boolean read StoreGroupInfo;
    property GroupCount:Integer read GetGroupCount;
		property Group[Index:Integer]:string read GetGroup;
		property ColumnsInGroup[Index:Integer]:Integer read GetColumnsInGroup write SetColumnsInGroup;
		property GroupImageIndex[Index:Integer]:Integer read GetGroupImageIndex write SetGroupImageIndex;
		property DirectionInGroup[Index:Integer]:TCreateDirection read GetDirectionInGroup write SetDirectionInGroup;

  	property Height:Integer read FHeight write SetHeight default 0;
		property Rows:Integer read FRows write SetRows default 1;
		property GroupType:TGroupType read FGroupType write FGroupType default gtTile;
    property AlwaysShowGroup:Boolean read FAlwaysShowGroup write FAlwaysShowGroup default False;
    property ImageIndex:Integer read FImageIndex write SetImageIndex default -1;

    property GroupTitlePosition:TGroupTitlePosition read FGroupTitlePosition write FGroupTitlePosition default gtpTop;  

	published
		property ParentDestination;
		property PrivateMode;

    property SectionName:string read FSectionName write SetSectionName;
		property PrivilegLevel:Integer read FPrivilegLevel write SetPrivilegLevel default 0;

		property OnInit:TNotifyEvent read FOnInit write FOnInit;
    property OnCalcFields:TNotifyEvent read FOnCalcFields write FOnCalcFields;
    property OnBeforeEdit:TNotifyEvent read FOnBeforeEdit write FOnBeforeEdit;
    property OnAfterEdit:TNotifyEvent read FOnAfterEdit write FOnAfterEdit;
    property OnBeforeWrite:TNotifyEvent read FOnBeforeWrite write FOnBeforeWrite;
    property OnAfterRead:TNotifyEvent read FOnAfterRead write FOnAfterRead;
	end;

	TINISection=class(TINICustomSection)
  protected
  	function StoreGroupInfo:Boolean; override;

  published
  	property GroupType;
    property ImageIndex;
    property AlwaysShowGroup;
		property Rows;
  	property Height;
    property GroupTitlePosition;
  end;

  TClassOfINISection=class of TINICustomSection;
  TMultiSectionButtons=set of TMultiSectionButton;

  TSectionProcedure=procedure(Data:Pointer) of object;

	TINICustomMultiSection=class;

  TFormSearchMultiSection=class(TForm)
  public
  	function Go(MultiSection:TINICustomMultiSection;ParentControl:TControl):Word; virtual; abstract;
  end;

  TClassOfFormSearchMultiSection=class of TFormSearchMultiSection;

	TINICustomMultiSection=class(TINICustomSection)
	private
		FOnAfterReadKeys:TNotifyEvent;
		FOnChange:TNotifyEvent;
    FOnButtonClick:TButtonClickEvent;
		FPanelList:TPanelList;
		FSections,FStaticSections:TStringList;
		FSelected,FDropDownCount:LongInt;
		FStatic:Boolean;
		FCompressKey:TCompressKeyMode;
    FButtons:TMultiSectionButtons;
    FEOL,FIsDirty:Boolean;
    FWalking:Boolean;
    FSaveIndex:Integer;
    FSorted,FNewAsCopy:Boolean;
		FAllowListChange,FUseMainSection:Boolean;

	  function CanDelete(Index:Integer):Boolean;

    procedure SPRead(Data:Pointer);
    procedure SPWrite(Data:Pointer);
    procedure SPDelete(Data:Pointer);
    procedure SPCopy(Data:Pointer);

		procedure SetStaticSections(sSections:TStringList);

		procedure WriteOnlySelected;
		procedure ReadOnlySelected;

		function GetKey:string;
		function GetKeyIndex:Integer;
		function GetKeyObject:TObject;

    function GetIndexOfKey(const sKey:string):Integer;
    function GetSectionCount:Integer;

		function GetSectionByIndex(Index:Integer):string;
		function GetKeyByIndex(Index:Integer):string;

		procedure SetSorted(bSorted:Boolean);

		procedure SetKey(const sKey:string);
		procedure SetKeyIndex(Index:Integer);

    procedure SetPanelList(PL:TPanelList);

		procedure ShowList;
		procedure HideList;
		procedure ClearList;
		procedure EnableList;
		procedure DisableList;

		procedure DeleteListItem(Index:Integer);

		procedure SectionsChanged(Sender:TObject);

		procedure ButtonSaveClick;
		procedure ButtonDeleteClick;
		procedure ButtonClearClick;

		procedure DoOnListChange(Sender:TObject);
		procedure DoOnNameChange(Sender:TObject);
		procedure DoOnButtonSaveClick(Sender:TObject);
		procedure DoOnButtonDeleteClick(Sender:TObject);
		procedure DoOnButtonClearClick(Sender:TObject);

		procedure NextItem;
		procedure PrevItem;
		procedure MoveDown;
		procedure MoveUp;

		procedure SenderNextItem(Sender:TObject);
		procedure SenderPrevItem(Sender:TObject);
		procedure SenderMoveDown(Sender:TObject);
		procedure SenderMoveUp(Sender:TObject);

		procedure ComboKeyDown(Sender:TObject;var Key:Word;Shift:TShiftState);
		procedure EditKeyDown(Sender:TObject;var Key:Word;Shift:TShiftState);

		procedure SaveSection(const sName:string);

	protected
    FPanelMode:TPanelMode;

    procedure OpenLocalGroupForEdit(const OpenGroup:string); override;

		function GetRuntimeSectionName:string; override;

    function InEditMode:Boolean;

		function InternalCreatePage(AOwner:TComponent;WParent:TWinControl;var NoCreated:Integer):TControl; override;

		procedure AssignStatic;
		procedure ExpandWithStatic;

		procedure AssignSections(sList:TStrings);

		procedure InsertSelected(const secName:string);

		procedure Loaded; override;

		procedure ListChange;
		procedure NameChange;

		procedure WriteKeysAlways;
		procedure ReadKeysAlways;

		procedure DeleteKeys;

    procedure DoForThisSection(Index:Integer;SP:TSectionProcedure;Data:Pointer);

		procedure DoWriteHumanReadable(Stream:TStream;Level:Integer); override;
		procedure DoReadHumanReadable(Stream:TStream;Level:Integer); override;

		procedure WriteThisSection(Index:Integer);
		procedure ReadThisSection(Index:Integer);
		procedure DeleteThisSection(Index:Integer);
		procedure CopyThisSection(const CopyFrom:string;Index:Integer);
    
		procedure CopySectionFromTo(fromS,toS:Integer);

		function DoCompressKey(const sKey:string):string; virtual;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

		procedure RefreshList;
		procedure RefreshEnablers;
		procedure ReReadSection;

		procedure DoWriteSection; override;
		procedure DoReadSection; override;
		procedure DoDeleteSection; override;
		procedure DoCopySection(const CopyFrom:string); override;

 		property SectionCount:Integer read GetSectionCount;
    
		property Selected:LongInt read FSelected write FSelected;
		property List:TPanelList read FPanelList write SetPanelList;
    property PanelMode:TPanelMode read FPanelMode write FPanelMode default pmComboList;

		property SectionByIndex[Index:Integer]:string read GetSectionByIndex;
		property Sections:TStringList read FSections;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure Reset; override;
		procedure RefreshControls; override;
		procedure ClearControls; override;
		procedure ShowControls; override;
		procedure HideControls; override;
		procedure EnableControls; override;
		procedure DisableControls; override;

		property Key:string read GetKey write SetKey;
		property KeyIndex:Integer read GetKeyIndex write SetKeyIndex;
		property KeyObject:TObject read GetKeyObject;
		property KeyByIndex[Index:Integer]:string read GetKeyByIndex;
		property IndexOfKey[const sKey:string]:Integer read GetIndexOfKey;

		procedure DeleteKey(const sKey:string);
		procedure DeleteKeyByIndex(Index:Integer);

    procedure WriteCurrentSection;

		function KeyExists(const sKey:string):Boolean;
		function FindKey(const sKey:string):Boolean;
		function AddKey(const sKey:string):Boolean;

		procedure AssignToStrings(sList:TStrings);
		procedure AssignToSections(sMultiSection:TINICustomMultiSection);

		procedure WriteKeys;
		procedure ReadKeys;

    procedure BeginWalk;
    procedure EndWalk;

    procedure First;
    procedure Last;
    procedure Next;
    procedure Prev;

 		property KeyCount:Integer read GetSectionCount;
    property EOL:Boolean read FEOL;

		property OnButtonClick:TButtonClickEvent read FOnButtonClick write FOnButtonClick;

    property StaticSections:TStringList read FStaticSections write SetStaticSections;
		property CompressKey:TCompressKeyMode read FCompressKey write FCompressKey default ciNo;
		property Static:Boolean read FStatic write FStatic default False;
    property Sorted:Boolean read FSorted write SetSorted default False;
		property DropDownCount:LongInt read FDropDownCount write FDropDownCount default 8;
    property Buttons:TMultiSectionButtons read FButtons write FButtons default [isSave,isDelete,isClear,isChange,isExchange];
    property NewAsCopy:Boolean read FNewAsCopy write FNewAsCopy stored True default False;

	published
		property OnAfterReadKeys:TNotifyEvent read FOnAfterReadKeys write FOnAfterReadKeys;
		property OnChange:TNotifyEvent read FOnChange write FOnChange;
	end;

	TINIMultiSection=class(TINICustomMultiSection)
  protected
  	function StoreGroupInfo:Boolean; override;

  published
  	property GroupType;
    property GroupTitlePosition;
    property AlwaysShowGroup;
		property Rows;
  	property Height;
    property ImageIndex;
		property OnButtonClick;

    property StaticSections;
		property CompressKey;
		property Static;
    property Sorted;
		property DropDownCount;
    property Buttons;
    property NewAsCopy;
  end;

{ TINEntry }

	TINIEntry=class(TINIField)
	private
		FOnAfterRead,FOnBeforeWrite:TINIFieldNotifyEvent;
		FOnStoreIt:TStoreItNotifyEvent;
		FGroupNo:Integer;
    FForceWrite:Boolean;
		FGroupName:string;
		FPrivilegLevel:Integer;
		FLabel:TControl;
		FOnInit:TINIFieldNotifyEvent;
 		FHint:string;
    FHelpContext:THelpContext;
    FCaptionStyle:TCaptionStyle;
    FConnectedControl:TWinControl;

		procedure ShowError(const FunctionName:string);
		procedure SetGroupNo(gNo:Integer);

		function GetGroupName:string;
		procedure SetGroupName(const gName:string);

    procedure SetControl(C:TControl);
    procedure SetConnectedControl(WC:TWinControl);

		function GetStoreIt:Boolean;
    
		function IsPrivilegLevelStored:Boolean;
    procedure SetPrivilegLevel(Level:Integer);

    function IsControlEnabled:Boolean;
    
	protected
		FControl:TControl;

		procedure Loaded; override;

  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; virtual;
    
		function GetPrivilegLevel:Integer; virtual;
		function CheckPrivilegLevel(Level:Integer):Boolean; virtual;
		function DoGetStoreIt:Boolean; virtual;
    
		procedure DoWriteHumanReadable(Stream:TStream;Level:Integer); virtual;
		procedure DoReadHumanReadable(Stream:TStream;Level:Integer); virtual;

    procedure DoDeleteEntry; virtual;
    procedure DoWriteEntry; virtual;
    procedure DoReadEntry; virtual;
    procedure DoCopyEntry; virtual;

    procedure DeleteEntry;
    procedure WriteEntry;
    procedure ReadEntry;
    procedure CopyEntry;

    function DoGetAsBoolean:Boolean; override;
    function DoGetAsFloat:Double; override;
    function DoGetAsInteger:LongInt; override;
    function DoGetAsString:string; override;

    procedure DoSetAsBoolean(Value:Boolean); override;
    procedure DoSetAsFloat(Value:Double); override;
    procedure DoSetAsInteger(Value:LongInt); override;
    procedure DoSetAsString(const Value:string); override;

		procedure SetCaption(const Caption:string); virtual;

		function GetCaption:string; virtual;
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; virtual;

    procedure Assign(Source:TPersistent); override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

		property StoreIt:Boolean read GetStoreIt;
    
	public
		constructor Create(AOwner:TComponent); override;

		property Caption:string read GetCaption write SetCaption;
		property Control:TControl read FControl write SetControl;
    property ConnectedControl:TWinControl read FConnectedControl write SetConnectedControl;

		procedure ShowControl(Show:Boolean); virtual;
		procedure EnableControl(State:Boolean); virtual;
		procedure RefreshControl; virtual;
		procedure ClearControl; virtual;
		procedure Init; virtual;
		procedure Reset; virtual;
		procedure Validate; virtual;
    procedure Syncronize(Direction:TSyncronizeDirection); virtual;

		function CreateControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;

    property Enabled:Boolean read IsControlEnabled write EnableControl;

    property CaptionStyle:TCaptionStyle read FCaptionStyle write FCaptionStyle default csLeftCenter;

		property OnStoreIt:TStoreItNotifyEvent read FOnStoreIt write FOnStoreIt;
		property OnAfterRead:TINIFieldNotifyEvent read FOnAfterRead write FOnAfterRead;
		property OnBeforeWrite:TINIFieldNotifyEvent read FOnBeforeWrite write FOnBeforeWrite;

    property HelpContext:THelpContext read FHelpContext write FHelpContext default 0;
    property ForceWrite:Boolean read FForceWrite write FForceWrite stored True default False;

	published
		property GroupIndex:Integer read FGroupNo write SetGroupNo default 0;
		property GroupName:string read GetGroupName write SetGroupName;
		property PrivilegLevel:Integer read GetPrivilegLevel write SetPrivilegLevel stored IsPrivilegLevelStored default 0;
		property Hint:string read FHint write FHint;

		property OnInit:TINIFieldNotifyEvent read FOnInit write FOnInit;
	end;

  TClassOfINIEntry=class of TINIEntry;

	TINIEntryStreamed=class(TINIEntry)
	private
    FShowInGrid:Boolean;

	protected
		function DoGetStoreIt:Boolean; override;

    procedure DoDeleteEntry; override;
    procedure DoWriteEntry; override;
    procedure DoReadEntry; override;
    procedure DoCopyEntry; override;

    procedure Assign(Source:TPersistent); override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;

	published
  	property Calculated;
    property ReadOnly;
    property ForceWrite;
    property CaptionStyle;
		property DisplayWidth;
    property HelpContext;

  	property OnChange;
    property OnFieldNameChange;
		property OnAfterRead;
		property OnBeforeWrite;
		property OnStoreIt;

    property ShowInGrid:Boolean read FShowInGrid write FShowInGrid default False;
	end;

  TClassOfINIEntryStreamed=class of TINIEntryStreamed;

	TINIEntryLink=class(TINIEntry)
	private
		FSubSection:TINICustomSection;

		procedure SetSubSection(sSection:TINICustomSection);

	protected
		function DoGetStoreIt:Boolean; override;

    function IsDirty:Boolean; override;

		procedure DoWriteHumanReadable(Stream:TStream;Level:Integer); override;
		procedure DoReadHumanReadable(Stream:TStream;Level:Integer); override;

    procedure DoDeleteEntry; override;
    procedure DoWriteEntry; override;
    procedure DoReadEntry; override;
    procedure DoCopyEntry; override;

		procedure ShowControl(Show:Boolean); override;
		procedure EnableControl(State:Boolean); override;
		procedure RefreshControl; override;
		procedure ClearControl; override;
		procedure Init; override;
		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

    procedure GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif}); override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		property SubSection:TINICustomSection read FSubSection write SetSubSection;
	end;

	TINIEntrySaveList=class(TINIEntryLink)
  private
  	FPanelMode:TPanelMode;
    FLineCount:Integer;

    procedure SetLineCount(LC:Integer);

    procedure SetStaticList(sList:TStringList);
    function GetStaticList:TStringList;

  	function GetList:TStringList;
  	function GetKey:string;
  	function GetKeyIndex:Integer;

  	procedure SetKey(const selS:string);

  	procedure Initialize;

  protected
  	function IsConnectionToControlSupported(WC:TWinControl):Boolean; override;

		procedure Loaded; override;
		procedure DoOnFieldNameChange; override;
		procedure DoOnConnected; override;

		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

  public
  	constructor Create(AOwner:TComponent); override;
  	destructor Destroy; override;

		property List:TStringList read GetList;
    property Key:string read GetKey write SetKey;
    property KeyIndex:Integer read GetKeyIndex;

    procedure AddKey(const aKey:string);

		procedure Reset; override;
    procedure Syncronize(Direction:TSyncronizeDirection); override;

	published
  	property CaptionStyle;
		property DisplayWidth;
		property StaticList:TStringList read GetStaticList write SetStaticList;
    property PanelMode:TPanelMode read FPanelMode write FPanelMode default pmComboList;
    property LineCount:Integer read FLineCount write SetLineCount default 8;
	end;

	TINIEntrySubSection=class(TINIEntryLink)
  protected
		procedure Notification(AComponent:TComponent;Operation:TOperation); override;

    procedure GetTableChildren(Proc:TGetChildProc); override;
		function GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl; override;

  public
  	destructor Destroy; override;
	end;

  TGlyphDescriptor=class(TPersistent)
  private
  	FGlyph:TBitmap;
    FNumGlyphs:Integer;
    FCaption:string;

    procedure SetGlyph(bGlyph:TBitmap);
    procedure SetNumGlyphs(nGlyphs:Integer);

  public
		constructor Create;
		destructor Destroy; override;

    procedure GAssignTo(BB:TBitBtn;bVisible:Boolean);
    procedure GAssign(GD:TGlyphDescriptor);

  published
  	property Glyph:TBitmap read FGlyph write SetGlyph;
    property NumGlyphs:Integer read FNumGlyphs write SetNumGlyphs default 1;
    property Caption:string read FCaption write FCaption;
  end;

	TGlyphs=class(TPersistent)
	protected
  	FGlyphDescriptors:array[0..2] of TGlyphDescriptor;

    procedure SetGlyphDescriptor(Index:Integer;GD:TGlyphDescriptor);
    function GetGlyphDescriptor(Index:Integer):TGlyphDescriptor;

	public
		constructor Create;
		destructor Destroy; override;

  published
		property Save:TGlyphDescriptor index 0 read GetGlyphDescriptor write SetGlyphDescriptor;
		property Delete:TGlyphDescriptor index 1 read GetGlyphDescriptor write SetGlyphDescriptor;
		property Clear:TGlyphDescriptor index 2 read GetGlyphDescriptor write SetGlyphDescriptor;
	end;

	TNames=class(TPersistent)
	private
    FUserName,FGroupName,FComputerName:string;
    FOnUserNameChange,FOnGroupNameChange:TNotifyEvent;
    FUseWindowsUserInstead:Boolean;

    procedure SetWindowsUserInstead(UWUInstead:Boolean);

    procedure SetGroupName(const sGroup:string);
    procedure SetUserName(const sUser:string);

    function IsUserNameStored:Boolean;
    function IsGroupNameStored:Boolean;

	public
		constructor Create;

    property OnUserNameChange:TNotifyEvent read FOnUserNameChange write FOnUserNameChange;
    property OnGroupNameChange:TNotifyEvent read FOnGroupNameChange write FOnGroupNameChange;

		procedure AssignWindowsUser;

  published
    property User:string read FUserName write SetUserName stored IsUserNameStored;
    property Group:string read FGroupName write SetGroupName stored IsGroupNameStored;
    property Computer:string read FComputerName write FComputerName;
    property UseWindowsUserInstead:Boolean read FUseWindowsUserInstead write SetWindowsUserInstead stored True default False;
	end;

	TDirs=class(TPersistent)
	private
    FPublic,FGroup,FUser:string;
    FUseRunDirInstead,FUseWindowsDirInstead:Boolean;

    function IsGroupStored:Boolean;
    function IsUserStored:Boolean;

    function GetPublic:string;
    function GetData:string;
    function GetUser:string;

    procedure SetPublic(const sPublic:string);
    procedure SetGroup(const sGroup:string);
    procedure SetUser(const sUser:string);

    procedure SetUseRunDirInstead(URDInstead:Boolean);
    procedure SetUseWindowsDirInstead(UWDInstead:Boolean);    

	public
		constructor Create;

		procedure AssignRunDir;
		procedure AssignWindowsDir;

  published
    property Public:string read GetPublic write SetPublic;
    property Group:string read GetData write SetGroup stored IsGroupStored;
    property User:string read GetUser write SetUser stored IsUserStored;
    property UseRunDirInstead:Boolean read FUseRunDirInstead write SetUseRunDirInstead stored True default True;
    property UseWindowsDirInstead:Boolean read FUseWindowsDirInstead write SetUseWindowsDirInstead  stored True default False;
	end;

	TINIDatabase=class(TComponent)
	private
		FOnAfterRead,FOnBeforeWrite,FOnInit:TNotifyEvent;
  	FSections:TList;
		FMaster:TComponent;
    FDestination:TDestination;
    FDistance:Integer;
		FConstCreateInfo:TCreateInfo;
    FNames:TNames;
    FDirs:TDirs;
    FFileName:string;
    FUserPrivilegLevel:Integer;
    FEditOnlyMode:Boolean;
    FCreatePageOnChange:Boolean;
    FormGetSection:TFormSearchMultiSection;
    FTabControl:TTabControl;
    FImageList:TImageList;
    FOnFieldChange:TINIFieldNotifyEvent;

		procedure OnTabChange(Sender:TObject);

    procedure UserNameChanged(Sender:TObject);
    procedure GroupNameChanged(Sender:TObject);

    procedure SetDestination(DT:TDestination);
    function IsDestinationStored:Boolean;
    
    function SectionByNumber(SectionNo:Integer):TINICustomSection;
    function GetSection(Index:Integer):TINICustomSection;
    function GetSectionOfPrivateModeCount:Integer;
    function GetSectionCount:Integer;

    procedure AddSection(Section:TINICustomSection);
    procedure RemoveSection(Section:TINICustomSection);
    procedure SetSection(Index:Integer;Value:TINICustomSection);

    procedure FEPWriteHumanReadable(Item:TObject;Data:Pointer);
    procedure FEPReadHumanReadable(Item:TObject;Data:Pointer);

    procedure FEPCopy(Item:TObject;Data:Pointer);
    procedure FEPInit(Item:TObject;Data:Pointer);
    procedure FEPReset(Item:TObject;Data:Pointer);
    procedure FEPRead(Item:TObject;Data:Pointer);
    procedure FEPWrite(Item:TObject;Data:Pointer);
    procedure FEPClear(Item:TObject;Data:Pointer);
    procedure FEPRefresh(Item:TObject;Data:Pointer);

		procedure SetDistance(CD:Integer);
    procedure SetUserPrivilegLevel(UserLevel:Integer);
    procedure SetNames(ns:TNames);
    procedure SetDirs(ds:TDirs);
    procedure SetGlyphs(gs:TGlyphs);
    procedure SetMaster(frm:TComponent);
    procedure SetParent(WParent:TWinControl);

    function GetMaster:TComponent;

		procedure DestroyTables;

		procedure GoSearchSection(MultiSection:TINICustomMultiSection);

	protected
  	FGlyphs:TGlyphs;
    FWhatPM:TPrivateModes;
    FStartPage:string;
    FParent:TWinControl;

    procedure AnyFieldChange(Field:TINIField);
    
		procedure DoWriteHumanReadable(Stream:TStream;Level:Integer); virtual;
		procedure DoReadHumanReadable(Stream:TStream;Level:Integer); virtual;

		procedure ForEachSection(FEP:TForEachProcedure;Data:Pointer;PMSet:TPrivateModes;Extend:Boolean);

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
    procedure GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif}); override;
    procedure Loaded; override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure Init;
		procedure Reset;
		procedure Read;
		procedure Write;
		procedure Clear;
		procedure Refresh;
		procedure Copy(const CopyFrom:string);

		procedure InitCreateInfo(BorrowCanvas:TCanvas);

    function SectionByName(const SectionName:string):TINICustomSection;
    function FindSection(const SectionName:string):TINICustomSection;
    procedure GetSectionNames(List:TStrings);

		procedure OpenSectionForEdit(Section:TINICustomSection);

		function CreatePage(AOwner:TComponent;Section:TINICustomSection;WParent:TWinControl):Integer;
		function CreatePages:Integer;

		procedure WriteHumanReadable(const FileName:string);
 		procedure ReadHumanReadable(const FileName:string);

		property SectionOfPrivateModeCount:Integer read GetSectionOfPrivateModeCount;
		property SectionCount:Integer read GetSectionCount;
    property Sections[Index:Integer]:TINICustomSection read GetSection write SetSection;
    property EditOnlyMode:Boolean read FEditOnlyMode write FEditOnlyMode;
    property Master:TComponent read GetMaster write SetMaster stored True default nil;

  published
    property PrivateModes:TPrivateModes read FWhatPM write FWhatPM default [pmPublic,pmGroupPrivate,pmUserPrivate];
    property Parent:TWinControl read FParent write SetParent;
    property UserPrivilegLevel:Integer read FUserPrivilegLevel write SetUserPrivilegLevel default 0;
    property FileName:string read FFileName write FFileName;
    property Distance:Integer read FDistance write SetDistance stored True default CCDistance;
    property Destination:TDestination read FDestination write SetDestination stored IsDestinationStored;
    property CreatePageOnChange:Boolean read FCreatePageOnChange write FCreatePageOnChange default True;
    property ImageList:TImageList read FImageList write FImageList;

    property Dirs:TDirs read FDirs write SetDirs;
    property Names:TNames read FNames write SetNames;
		property Glyphs:TGlyphs read FGlyphs write SetGlyphs;

    property OnInit:TNotifyEvent read FOnInit write FOnInit;
    property OnAfterRead:TNotifyEvent read FOnAfterRead write FOnAfterRead;
    property OnBeforeWrite:TNotifyEvent read FOnBeforeWrite write FOnBeforeWrite;
    
    property OnFieldChange:TINIFieldNotifyEvent read FOnFieldChange write FOnFieldChange;
	end;

	TDestinationProperty=class(TStringProperty)
  protected
  	function GetAttributes:TPropertyAttributes; override;

  	procedure GetValues(Proc:TGetStrProc); override;
  end;

  { Register classes }

  TObjectClassDescription=class(TObject)
  protected
    FDescription:string;
  public
    property Description:string read FDescription;
  end;

  TEntryObjectClass=class(TObjectClassDescription)
  private
  	FClass:TClassOfINIEntry;
    FImageIndex:Integer;

  public
  	constructor Create(ObjectClass:TClassOfINIEntry;const ClassDescription:string;Index:Integer);

    property ObjectClass:TClassOfINIEntry read FClass;
    property ImageIndex:Integer read FImageIndex;
  end;

  TSectionObjectClass=class(TObjectClassDescription)
  private
  	FClass:TClassOfINISection;

  public
  	constructor Create(ObjectClass:TClassOfINISection;const ClassDescription:string);

    property ObjectClass:TClassOfINISection read FClass;
  end;

  TStreamObjectClass=class(TObjectClassDescription)
  private
  	FClass:TClassOfINIStream;

  public
  	constructor Create(ObjectClass:TClassOfINIStream;const ClassDescription:string);

    property ObjectClass:TClassOfINIStream read FClass;
  end;

	TCfgSubPanel=class(TPanel)
	public
		constructor Create(AOwner:TComponent); override;
	end;

var DebugMode:Boolean;
    AutoRename,WarnOnDuplicateFieldName:Boolean;
    DefaultGroupName:string;
    FormSearchMultiSectionClass:TClassOfFormSearchMultiSection;

function CreateUntitledControl(AOwner:TComponent;TCC:TControlClass;WParent:TControl;cInfo:TCreateInfo):TControl;
function CreateTitledControl(AOwner:TComponent;TCC:TControlClass;WParent:TControl;cInfo:TCreateInfo):TControl;

function CreateStream(const ComputerName,Name,StreamDescription:string):TINIStream;

procedure RegisterINIStream(const StreamDescription:string;StreamClass:TClassOfINIStream);
procedure RegisterINISection(const SectionDescription:string;SectionClass:TClassOfINISection);
procedure RegisterINIEntry(const EntryDescription:string;ImageIndex:TEntryImageIndex;EntryClass:TClassOfINIEntry);

procedure UnRegisterINISection(SectionClass:TClassOfINISection);
procedure UnRegisterINIEntry(EntryClass:TClassOfINIEntry);

function GetEntryClassesCount:Integer;
function GetSectionClassesCount:Integer;
function GetStreamClassesCount:Integer;

function GetEntryClassDescription(Index:Integer):string;
function GetSectionClassDescription(Index:Integer):string;
function GetStreamClassDescription(Index:Integer):string;

function GetEntryClassIndex(const ClassName:string):Integer;
function GetSectionClassIndex(const ClassName:string):Integer;

function GetEntryClass(Index:Integer):TClassOfINIEntry;
function GetSectionClass(Index:Integer):TClassOfINISection;

function GetEntryClassImageIndex(Index:Integer):Integer;
function GetEntryClassImageIndexByName(const EntryClassName:string):Integer;

function IsStreamFileBased(const StreamDescription:string):Boolean;

procedure ShiftDown(List:TStrings);
procedure PromoteToLeader(sList:TStrings;Index:LongInt);

implementation

{$R INIRes.RES}

var EntryClasses,SectionClasses,StreamClasses:TStringList;

function IsINIStreamRegistered(const StreamDescription:string):Boolean;
var i:Integer;
begin
  for i:=0 to StreamClasses.Count-1 do begin
    if AnsiCompareText((StreamClasses.Objects[i] as TStreamObjectClass).Description,StreamDescription)=0 then begin
    	Result:=True;

      Exit;
    end;
  end;

  Result:=False;
end;

{ error functions }

procedure INIError(const Message:string);
begin
  raise EINIError.Create(Message);
end;

procedure INIErrorFmt(const Message:string;Params:array of const);
begin
  raise EINIError.CreateFmt(Message,Params);
end;

{ misc. functions }

procedure ShiftDown(List:TStrings);
var i,lCount:LongInt;
		lSaveItem:string;
begin
	lCount:=List.Count;

	lSaveItem:=List.Strings[0];

	for i:=0 to lCount-2 do begin
		List.Strings[i]:=List.Strings[i+1];
	end;

	List.Strings[lCount-1]:=lSaveItem;
end;

procedure PromoteToLeader(sList:TStrings;Index:LongInt);
var i:LongInt;
begin
	for i:=1 to Index do begin
		ShiftDown(sList);
	end;
end;

constructor TCfgSubPanel.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	BevelInner:=bvNone;
	BevelOuter:=bvNone;
	BorderWidth:=0;
	Caption:='';
end;

{ TCreateInfo }

constructor TCreateInfo.Create(CI:TCreateInfo);
begin
	inherited Create;

  fMaxCtrlH:=0;

  if CI<>nil then begin
    fTitle:=CI.fTitle;
    fX:=CI.fX;
    fY:=CI.fY;
    fW:=CI.fW;
    fH:=CI.fH;
    fCtrlW:=CI.fCtrlW;
    fCtrlH:=CI.fCtrlH;
    fMaxCtrlH:=CI.fMaxCtrlH;
    fCalculated:=CI.fCalculated;
    fRO:=CI.fRO;
    fDirection:=CI.fDirection;
    fCols:=CI.fCols;
    fcCol:=CI.fcCol;
    fcRow:=CI.fcRow;
    fcWidth:=CI.fcWidth;
    fcHeight:=CI.fcHeight;
    fXBor:=CI.fXBor;
    fYBor:=CI.fYBor;
    fInitX:=CI.fInitX;
    fInitY:=CI.fInitY;
    fControlDistance:=CI.fControlDistance;
    fCanvas:=CI.fCanvas;
    fTitled:=CI.fTitled;
  end;
end;

procedure TCreateInfo.SetCurrentColumn(cColumn:Integer);
begin
	if (cColumn>=0) and (cColumn<fCols) then fcCol:=cColumn;
end;

procedure TCreateInfo.NoTitle;
begin
	fTitled:=False;
end;

function TCreateInfo.TextWidth(const Msg:string):Integer;
begin
	Result:=fCanvas.TextWidth(Msg);
end;

function TCreateInfo.TextHeight(const Msg:string):Integer;
begin
	Result:=fCanvas.TextHeight(Msg);
end;

procedure TCreateInfo.FirstColumn;
begin
	fcCol:=0;
  
	fX:=fInitX+fXBor;
end;

procedure TCreateInfo.FirstRow;
begin
	fcRow:=0;
  
	fY:=fInitY+fYBor;
end;

function TCreateInfo.NextColumn:Boolean;
begin
	Result:=fcCol<fCols-1;

  if Result then begin
	  fX:=fX+fW+fXBor;
  	fcCol:=fcCol+1;
  end;
end;

function TCreateInfo.NextRow:Boolean;
var Advance:Integer;
begin
	Advance:=Max(fCtrlH,fMaxCtrlH);

	Result:=Advance>0;

  if Result then begin
		fY:=fY+Advance+fControlDistance;
	  fcRow:=fcRow+1;
  end;

  fMaxCtrlH:=0;
end;

function TCreateInfo.GetState:TCreateInfoState;
begin
	Result.X:=fX;
  Result.Y:=fY;
  Result.CtrlW:=fCtrlW;
  Result.CtrlH:=fCtrlH;
  Result.MaxCtrlH:=fMaxCtrlH;
  Result.cCol:=fcCol;
  Result.cRow:=fcRow;
  Result.Title:=fTitle;
end;

procedure TCreateInfo.SetState(State:TCreateInfoState);
begin
	fX:=State.X;
  fY:=State.Y;
  fCtrlW:=State.CtrlW;
  fCtrlH:=State.CtrlH;
  fMaxCtrlH:=State.MaxCtrlH;
  fcCol:=State.cCol;
  fcRow:=State.cRow;
  fTitle:=State.Title;
end;

{ TINITable }

constructor TINITable.Create(AOwner:TComponent);
begin
  FDestination:=CDefaultFileStream;

  inherited Create(AOwner);

  FFileName:='';
  FCopyFromFileName:='';

  FINIStream:=nil;
  FParentField:=nil;
  FDatabase:=nil;

  FReadOnly:=False;
	FParentDestination:=True;
	FPrivateMode:=pmPublic;
  FDirection:=cdTopToBottom;
  
  FOnFieldChange:=nil;

  FTableNo:=0;

  FFields:=TList.Create;
end;

destructor TINITable.Destroy;
begin
	DestroyFields;

  FFields.Free;

  inherited Destroy;
end;

procedure TINITable.SetParentDestination(PD:Boolean);
begin
	FParentDestination:=PD;

  if Connected and FParentDestination then begin
    FFileName:=RuntimeFileName;
    FDestination:=RuntimeDestination;
  end;
end;

procedure TINITable.SetPrivateMode(PM:TPrivateMode);
begin
	FPrivateMode:=PM;
end;

procedure TINITable.AdjustParentDestination;
begin
	if Connected then begin
  	FParentDestination:=True;
	 	FParentDestination:=(FDestination=RuntimeDestination) and StrICmp(FFileName,RuntimeFileName);
  end;
end;

procedure TINITable.SetFileName(const FName:string);
begin
	FFileName:=FName;

  AdjustParentDestination;
end;

procedure TINITable.AnyFieldChange(Field:TINIField);
begin
	if Database<>nil then Database.AnyFieldChange(Field);

	if Assigned(FOnFieldChange) then FOnFieldChange(Field);
end;

procedure TINITable.SetDestination(DT:TDestination);
begin
	if IsINIStreamRegistered(DT) then
  	FDestination:=DT
  else
  	raise Exception.Create(Format(ResStr(CResStreamNotRegistered),[DT]));
    
  AdjustParentDestination;
end;

function TINITable.IsDestinationStored:Boolean;
begin
	Result:=StoredIfNotParentDestination and (AnsiCompareText(FDestination,CDefaultFileStream)<>0);
end;

function TINITable.StoredIfNotParentDestination:Boolean;
begin
	Result:=not FParentDestination;
end;

procedure TINITable.SetParentComponent(AParent:TComponent);
begin
	if not (csLoading in ComponentState) and (AParent is TINIDatabase) then Database:=TINIDatabase(AParent);
end;

function TINITable.GetParentComponent:TComponent;
begin
	if FParentField<>nil then
  	Result:=FParentField
  else
  	Result:=FDatabase;
end;

procedure TINITable.SetTableParent(AParent:TComponent);
begin
	SetParentComponent(AParent);
end;

function TINITable.GetTableParent:TComponent;
begin
	Result:=GetParentComponent;
end;

function TINITable.HasParent:Boolean;
begin
	Result:=True;
end;

function TINITable.GetINIStream:TINIStream;
begin
	Result:=FINIStream;

  if Result=nil then raise Exception.Create(ResStr(CResStreamNotAssigned));
end;

function TINITable.GetFileName:string;
begin
 	Result:=FFileName;

  if not (csLoading in ComponentState) then Result:=RuntimeFileName;
end;

function TINITable.GetDestination:TDestination;
begin
 	Result:=FDestination;

  if not (csLoading in ComponentState) then Result:=RuntimeDestination;
end;

function TINITable.Connected:Boolean;
begin
	Result:=(([csReading,csLoading]*ComponentState)=[]) and ((FParentField<>nil) or (FDatabase<>nil)); 
end;

function TINITable.IsSon:Boolean;
begin
	Result:=FParentField<>nil;
end;

procedure TINITable.SetDatabase(DB:TINIDatabase);
begin
end;

function TINITable.GetDatabase:TINIDatabase;

	function FindDatabase:TINIDatabase;
	var i,cc:Integer;
  begin
  	Result:=nil;

  	if Owner<>nil then begin
    	if Owner is TINIField then
      	Result:=(Owner as TINIField).Table.Database
    	else if Owner is TINITable then
      	Result:=(Owner as TINITable).Database
      else begin
        cc:=Owner.ComponentCount-1;

        for i:=0 to cc do begin
          if Owner.Components[i] is TINIDatabase then begin
            Result:=Owner.Components[i] as TINIDatabase;

            Exit;
          end;
        end;
      end;
    end;
  end;
  
begin
	Result:=nil;

	if FParentField<>nil then
  	Result:=FParentField.Table.Database
  else if FDatabase<>nil then
  	Result:=FDatabase;

  if Result=nil then raise Exception.Create(ResStr(CResINIDatabaseMissing));
end;

function TINITable.GetEditOnlyMode:Boolean;
begin
	Result:=Database.FEditOnlyMode;
end;

procedure TINITable.SetName(const Value:TComponentName);
var i:Integer;
  	OldName,FieldName,NamePrefix:TComponentName;
  	Field:TINIField;
begin
  OldName:=Name;

  inherited SetName(Value);
  
  if (csDesigning in ComponentState) and (Name<>OldName) and (Name<>'') and AutoRename then begin
    for i:=0 to FFields.Count-1 do begin
      Field:=FFields[i];

      if Field.Owner=Owner then begin
        FieldName:=Field.Name;
        NamePrefix:=FieldName;

        if Length(NamePrefix)>Length(OldName) then begin
          SetLength(NamePrefix,Length(OldName));

          if AnsiCompareText(OldName,NamePrefix) = 0 then begin
            System.Delete(FieldName,1,Length(OldName));
            System.Insert(Value,FieldName,1);
            
            try
              Field.Name:=FieldName;

	            except
  	            on EComponentError do begin
                end;
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure TINITable.GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif});
var i:Integer;
  	Field:TINIField;
begin
  for i:=0 to FFields.Count-1 do begin
    Field:=FFields[i];

    if Field.Owner<>Self then Proc(Field);
  end;
end;

procedure TINITable.SetChildOrder(Component:TComponent; Order:Integer);
begin
  if FFields.IndexOf(Component)>=0 then (Component as TINIField).Index:=Order;
end;

procedure TINITable.Loaded;
begin
  inherited Loaded;
end;

procedure TINITable.DestroyFields;
var Field:TINIField;
begin
  while FFields.Count>0 do begin
    Field:=FFields.Last;
    RemoveField(Field);
    Field.Free;
  end;
end;

procedure TINITable.AddField(Field:TINIField);
begin
  FFields.Add(Field);
  Field.FTable:=Self;
end;

procedure TINITable.RemoveField(Field:TINIField);
begin
  Field.FTable:=nil;
  FFields.Remove(Field);
end;

function TINITable.GetFieldCount:Integer;
begin
  Result:=FFields.Count;
end;

function TINITable.GetField(Index:Integer):TINIField;
begin
  Result:=FFields[Index];
end;

procedure TINITable.SetField(Index:Integer; Value:TINIField);
begin
  TINIField(FFields[Index]).Assign(Value);
end;

function TINITable.FieldByName(const FieldName:string):TINIField;
begin
  Result:=FindField(FieldName);

  if Result=nil then INIErrorFmt(ResStr(CResFieldNotFound), [FieldName]);
end;

function TINITable.FieldByNumber(FieldNo:Integer):TINIField;
var i:Integer;
begin
  for i:=0 to FFields.Count-1 do begin
    Result:=Fields[i];

    if Result.FieldNo=FieldNo then Exit;
  end;

  Result:=nil;
end;

function TINITable.FindField(const FieldName:string):TINIField;
var i:Integer;
begin
  for i:=0 to FFields.Count-1 do begin
    Result:=FFields[i];

    if AnsiCompareText(Result.FFieldName,FieldName)=0 then Exit;
  end;

  Result:=nil;
end;

procedure TINITable.CheckFieldName(const FieldName:string;Field:TINIField);
var fField:TINIField;
begin
	if csDesigning in ComponentState then begin
  	fField:=FindField(FieldName);

    if (fField is TINIEntryStreamed) and (fField<>Field) and WarnOnDuplicateFieldName and (FieldName<>'') then
	  	ShowMessage(Format(ResStr(CResDuplicateFieldName), [FieldName]));
  end;
end;

procedure TINITable.GetFieldNames(List:TStrings);
var i:Integer;
begin
  List.BeginUpdate;
  
  try
    List.Clear;

    for i:=0 to FFields.Count-1 do List.Add(TINIField(FFields[i]).FFieldName);

	  finally
  	  List.EndUpdate;
  end;
end;

function TINITable.GetReadOnly:Boolean;
begin
  Result:=FReadOnly;

  if FParentField is TINIEntryLink then with TINIEntryLink(FParentField) do begin
  	if SubSection<>nil then Result:=Result and SubSection.ReadOnly;
  end;
end;

procedure TINITable.CheckCanModify;
begin
  if ReadOnly then INIError(ResStr(CResReadOnlyMode));
end;

procedure TINITable.ReadState(Reader:TReader);
begin
  inherited ReadState(Reader);

  if Reader.Parent is TINIDatabase then Database:=TINIDatabase(Reader.Parent);
end;

function TINITable.GetRuntimeFileName:string;
begin
 	Result:=FFileName;

	if FParentDestination or (Result='') then begin
  	if Son then
    	Result:=FParentField.Table.RuntimeFileName
    else
	  	Result:=Database.FileName;
  end;
end;

function TINITable.GetRuntimePrivateMode:TPrivateMode;
begin
 	Result:=FPrivateMode;
end;

function TINITable.GetRuntimeDestination:TDestination;
begin
	Result:=FDestination;

	if FParentDestination or (Result='') then begin
  	if Son then
    	Result:=FParentField.Table.RuntimeDestination
    else
	  	Result:=Database.Destination;
  end;
end;

{ TINIField }

constructor TINIField.Create(AOwner:TComponent);
begin
  FVisible:=True;
  FIsDirty:=False;
  FTable:=nil;
  FOnChange:=nil;
  FOnFieldNameChange:=nil;

  inherited Create(AOwner);
end;

destructor TINIField.Destroy;
begin
  if FTable<>nil then FTable.RemoveField(Self);

  FTable:=nil;

  inherited Destroy;
end;

procedure TINIField.DoOnConnected;
begin
end;

procedure TINIField.DoOnDisconnected;
begin
end;

function TINIField.IsDirty:Boolean;
begin
	Result:=FIsDirty;
end;

procedure TINIField.Change;
begin
	FIsDirty:=True;

  if FTable<>nil then FTable.AnyFieldChange(Self);

  if Assigned(FOnChange) then FOnChange(Self);
end;

function TINIField.GetAsBoolean:Boolean;
begin
	Result:=DoGetAsBoolean;
end;

function TINIField.GetAsFloat:Double;
begin
	Result:=DoGetAsFloat;
end;

function TINIField.GetAsInteger:LongInt;
begin
	Result:=DoGetAsInteger;
end;

function TINIField.GetAsString:string;
begin
	Result:=DoGetAsString;
end;

procedure TINIField.SetAsBoolean(Value:Boolean);
var WasAsBoolean:Boolean;
begin
	WasAsBoolean:=GetAsBoolean;

	DoSetAsBoolean(Value);

  if WasAsBoolean<>GetAsBoolean then Change;
end;

procedure TINIField.SetAsFloat(Value:Double);
var WasAsFloat:Double;
begin
	WasAsFloat:=GetAsFloat;

	DoSetAsFloat(Value);

  if WasAsFloat<>GetAsFloat then Change;
end;

procedure TINIField.SetAsInteger(Value:LongInt);
var WasAsInteger:LongInt;
begin
	WasAsInteger:=GetAsInteger;

	DoSetAsInteger(Value);

  if WasAsInteger<>GetAsInteger then Change;
end;

procedure TINIField.SetAsString(const Value:string);
var WasAsString:string;
begin
	WasAsString:=GetAsString;

	DoSetAsString(Value);

  if WasAsString<>GetAsString then Change;
end;

procedure TINIField.GetTableChildren(Proc:TGetChildProc);
begin
end;

function TINIField.GetDefaultAsBoolean:Boolean;
begin
	if VarType(FDefaultValue)=varString then
  	Result:=False
  else
		Result:=FDefaultValue;
end;

function TINIField.GetDefaultAsFloat:Double;
begin
	Result:=StrToFloatDef(DefaultAsString,0);
end;

function TINIField.GetDefaultAsInteger:LongInt;
begin
	Result:=StrToIntDef(DefaultAsString,0);
end;

function TINIField.GetDefaultAsString:string;
begin
	Result:=FDefaultValue;
end;

procedure TINIField.SetDefaultAsBoolean(Value:Boolean);
begin
	FDefaultValue:=Value;
end;

procedure TINIField.SetDefaultAsFloat(Value:Double);
begin
	FDefaultValue:=Value;
end;

procedure TINIField.SetDefaultAsInteger(Value:LongInt);
begin
	FDefaultValue:=Value;
end;

procedure TINIField.SetDefaultAsString(const Value:string);
begin
	FDefaultValue:=Value;
end;

function TINIField.GetCalculated:Boolean;
begin
  Result:=FCalculated;
end;

function TINIField.GetCanModify:Boolean;
begin
  Result:=not ReadOnly;

  if Table<>nil then Result:=Result and not Table.ReadOnly;
end;

function TINIField.GetDefaultWidth:Integer;
begin
  Result:=CDefDisplayWidth;
end;

function TINIField.GetDisplayLabel:string;
begin
  if FDisplayLabel<>'' then
    Result:=FDisplayLabel
  else if FFieldName<>'' then
    Result:=FFieldName
  else
  	Result:=CUntitled;
end;

function TINIField.GetDisplayWidth:Integer;
begin
  Result:=FDisplayWidth;
end;

function TINIField.GetIndex:Integer;
begin
  if FTable<>nil then
    Result:=FTable.FFields.IndexOf(Self)
  else
    Result:=-1;
end;

function TINIField.HasParent:Boolean;
begin
  HasParent:=True;
end;

function TINIField.GetParentComponent:TComponent;
begin
  Result:=Table;
end;

procedure TINIField.SetParentComponent(AParent:TComponent);
begin
  if not (csLoading in ComponentState) and (AParent is TINITable) then Table:=TINITable(AParent);
end;

function TINIField.IsDisplayLabelStored:Boolean;
begin
  Result:=FDisplayLabel<>'';
end;

function TINIField.IsDisplayWidthStored:Boolean;
begin
  Result:=FDisplayWidth<>GetDefaultWidth;
end;

procedure TINIField.Notification(AComponent:TComponent;Operation:TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

procedure TINIField.ReadState(Reader:TReader);
begin
  inherited ReadState(Reader);

  if Reader.Parent is TINITable then Table:=TINITable(Reader.Parent);
end;

procedure TINIField.SetCalculated(Value:Boolean);
begin
	FCalculated:=Value;
end;

procedure TINIField.SetTable(ATable:TINITable);
var FormerTable:TINITable;
begin
	if FTable<>ATable then begin
  	FormerTable:=FTable;

    if ATable<>nil then begin
      ATable.CheckFieldName(FFieldName,Self);
    end;

    if FTable<>nil then FTable.RemoveField(Self);
    if ATable<>nil then ATable.AddField(Self);

    try
    	if (FTable<>nil) and (FormerTable<>nil) and (Name=FormerTable.Name+FieldName) and AutoRename then Name:=FTable.Name+FieldName;

      except
      	on Exception do begin
        end;
    end;

    if ATable<>nil then
	    DoOnConnected
    else
	    DoOnDisconnected;
  end;
end;

procedure TINIField.SetDisplayLabel(Value:string);
begin
  if Value=FFieldName then Value:='';

  if FDisplayLabel<>Value then FDisplaylabel:=Value;
end;

procedure TINIField.SetDisplayWidth(Value:Integer);
begin
  if FDisplayWidth<>Value then FDisplayWidth:=Max(0,Value);
end;

procedure TINIField.SetEditMask(const Value:string);
begin
  FEditMask:=Value;
end;

procedure TINIField.DoOnFieldNameChange;
begin
	if Assigned(FOnFieldNameChange) then FOnFieldNameChange(Self);
end;

procedure TINIField.SetFieldName(const Value:string);
begin
  if FTable<>nil then FTable.CheckFieldName(Value,Self);

  FFieldName:=Value;
  
  if (FDisplayLabel=Value) and (Value<>'') then FDisplayLabel:='';

  DoOnFieldNameChange;
end;

procedure TINIField.SetVisible(Value:Boolean);
begin
  if FVisible<>Value then FVisible:=Value;
end;

procedure TINIField.SetIndex(Value:Integer);
var CurIndex,Count:Integer;
begin
  CurIndex:=GetIndex;

  if CurIndex>=0 then begin
    Count:=FTable.FFields.Count;

    if Value<0 then Value:=0;
    if Value>=Count then Value:=Count-1;
    if Value<>CurIndex then begin
      FTable.FFields.Delete(CurIndex);
      FTable.FFields.Insert(Value,Self);
    end;
  end;
end;

{
*******************************************************
	TGroupInfo
********************************************************
}

function TGroupInfo.GetImageIndex:Integer;
begin
	Result:=(FValue shr 8) and $ff;

  if Result=255 then Result:=-1;
end;

function TGroupInfo.GetColumns:Integer;
begin
	Result:=FValue and $7F;
end;

function TGroupInfo.GetDirection:TCreateDirection;
begin
	if (FValue and $80)=0 then
  	Result:=cdTopToBottom
  else
  	Result:=cdLeftToRight;
end;

procedure TGroupInfo.SetImageIndex(II:Integer);
begin
	if (II>=-1) then FValue:=(FValue and $00ff) or ((II shl 8) and $ff00);
end;

procedure TGroupInfo.SetColumns(sColumns:Integer);
begin
	FValue:=(FValue and $ff80) or (sColumns and $7F);
end;

procedure TGroupInfo.SetDirection(sDirection:TCreateDirection);
begin
	if sDirection=cdTopToBottom then
		FValue:=FValue and $ff7F
  else
		FValue:=FValue or $80;
end;

constructor TGroupInfo.CreateInfo(sColumns,II:Integer;sDirection:TCreateDirection);
begin
	inherited Create;

  FValue:=$ff00;

  Columns:=sColumns;
  ImageIndex:=II;
  Direction:=sDirection;
end;

constructor TGroupInfo.Create(Value:Integer);
begin
  inherited Create;

  FValue:=Value;
end;

{
*******************************************************
	INISection
********************************************************
}

constructor TINICustomSection.Create(AOwner:TComponent);
begin
	FSectionState:=[ssLoading];

	inherited Create(AOwner);

  FHeight:=0;
	FRows:=1;
	FPrivilegLevel:=0;
  FAlwaysShowGroup:=False;

  FImageIndex:=-1;
	FGroupType:=gtTile;
  FGroupTitlePosition:=gtpTop;

	FPrefix:='';
  FSectionName:='';

	FOnInit:=nil;
	FOnCalcFields:=nil;
  FOnBeforeEdit:=nil;
  FOnAfterEdit:=nil;
  FOnAfterRead:=nil;
  FOnBeforeWrite:=nil;

	FGroups:=TStringList.Create;

	FGroups.Add(DefaultGroupName);
end;

destructor TINICustomSection.Destroy;
begin
	FGroups.Free;

  if FDatabase<>nil then FDatabase.RemoveSection(Self);
  if FParentField is TINIEntryLink then TINIEntryLink(FParentField).FSubSection:=nil;

	inherited Destroy;
end;

procedure TINICustomSection.Loaded;
begin
	inherited Loaded;

  Init;

  if FGroups.Count=0 then FGroups.Add(DefaultGroupName);

  FSectionState:=FSectionState-[ssLoading];
end;

function TINICustomSection.StoreGroupInfo:Boolean;
begin
	Result:=False;
end;

procedure TINICustomSection.WriteGroups(Writer:TWriter);
var i,Value:Integer;
begin
  Writer.WriteListBegin;

  for i:=0 to FGroups.Count-1 do begin
    if FGroups.Objects[i]<>nil then begin
    	Value:=TGroupInfo(FGroups.Objects[i]).FValue;

      if (Value and $ff00)<>$ff then Value:=Value+$0100;

	  	Writer.WriteInteger(Value);
    end else begin
	  	Writer.WriteInteger($ff01);
    end;

  	Writer.WriteString(Group[i]);
  end;

  Writer.WriteListEnd;
end;

procedure TINICustomSection.DefineProperties(Filer:TFiler);
begin
	inherited DefineProperties(Filer);

  Filer.DefineProperty('SectionGroups', ReadGroups, WriteGroups, (FGroups.Count>0) and StoreGroupInfo);
end;

procedure TINICustomSection.ReadGroups(Reader:TReader);
var gInfo:Integer;
		gName:string;
begin
  Reader.ReadListBegin;

  FGroups.Clear;

  while not Reader.EndOfList do begin
  	gInfo:=Reader.ReadInteger;
  	gName:=Reader.ReadString;

    if ((gInfo and $ff00)<>$ff) and ((gInfo and $ff00)<>$00) then
    	gInfo:=gInfo-$0100
    else
    	gInfo:=gInfo or $ff00;

  	FGroups.AddObject(gName,TGroupInfo.Create(gInfo));
  end;

  Reader.ReadListEnd;
end;

function TINICustomSection.FindSection(const sName:string):TINICustomSection;
var i:Integer;
begin
	for i:=0 to FieldCount-1 do begin
  	if Fields[i] is TINIEntrySubSection then with Fields[i] as TINIEntrySubSection do begin
    	if (SubSection<>nil) and (AnsiCompareText(SubSection.SectionName,sName)=0) then begin
      	Result:=SubSection;

      	Exit;
      end;
    end;
  end;

  Result:=nil;
end;

procedure TINICustomSection.SetImageIndex(II:Integer);
begin
	FImageIndex:=Max(-1,II);
end;

function TINICustomSection.SectionByName(const sName:string):TINICustomSection;
begin
  Result:=FindSection(sName);

  if Result=nil then INIErrorFmt(ResStr(CResSectionNotFound),[sName]);
end;

procedure TINICustomSection.SetPanel(pnl:TPanel);
begin
	FPanel:=pnl;

  if FPanel<>nil then begin
  	FSectionState:=FSectionState+[ssEdit];

    FPanel.FreeNotification(Self);

  	if Assigned(FOnBeforeEdit) then FOnBeforeEdit(Self);
  end;
end;

procedure TINICustomSection.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if Operation=opRemove then begin
  	if AComponent=FPanel then begin
    	FPanel:=nil;

      FSectionState:=FSectionState-[ssEdit];

	  	if Assigned(FOnAfterEdit) then FOnAfterEdit(Self);
    end;
  end;
end;

procedure TINICustomSection.SetDatabase(DB:TINIDatabase);
begin
	if DB<>FDatabase then begin
    if FDatabase<>nil then FDatabase.RemoveSection(Self);
    if DB<>nil then DB.AddSection(Self);
  end;
  
	inherited SetDatabase(DB);
end;

procedure TINICustomSection.SetParentComponent(AParent:TComponent);
begin
	inherited SetParentComponent(AParent);

	if not (csLoading in ComponentState) and (AParent is TINIEntryLink) then begin
  	FParentField:=nil;

    TINIEntryLink(AParent).SubSection:=Self;
  end else if FDatabase<>nil then begin
  	if FParentField is TINIEntryLink then begin
    	(FParentField as TINIEntryLink).FSubSection:=nil;

    	FParentField.Free;

      FParentField:=nil;
    end;
  end;
end;

procedure TINICustomSection.ReadState(Reader:TReader);
begin
  inherited ReadState(Reader);

  if Reader.Parent is TINIEntryLink then begin
  	FParentField:=nil;

  	TINIEntryLink(Reader.Parent).SubSection:=Self;
  end;
end;

procedure TINICustomSection.SetPrivilegLevel(Level:Integer);
begin
	FPrivilegLevel:=Max(0,Min(Level,CMaxPrivilegLevel));

  if FParentField is TINIEntry then (FParentField as TINIEntry).FPrivilegLevel:=FPrivilegLevel;
end;

procedure TINICustomSection.SetHeight(h:Integer);
begin
	FHeight:=Max(0,h);
end;

procedure TINICustomSection.SetRows(Row:Integer);
begin
	if (Row>=1) and (Row<=16) then FRows:=Row;
end;

procedure TINICustomSection.SetGroups(gList:TStringList);
var i:Integer;
begin
	FGroups.Clear;
  
	for i:=0 to gList.Count-1 do begin
  	if gList.Objects[i]<>nil then begin
    	FGroups.AddObject(gList[i],TGroupInfo.Create(TGroupInfo(gList.Objects[i]).FValue));
    end else begin
    	FGroups.AddObject(gList[i],TGroupInfo.Create($ff01));
    end;
  end;
end;

procedure TINICustomSection.RenameGroup(const oldName,newName:string);
var i,gc:Integer;

	procedure IterateFields;
  var i,fc:Integer;
  begin
  	fc:=FieldCount-1;

    for i:=0 to fc do begin
    	if (AnsiCompareText((Fields[i] as TINIEntry).GroupName,oldName)=0) then (Fields[i] as TINIEntry).GroupName:=newName;
    end;
  end;

begin
	gc:=GroupCount-1;

  for i:=0 to gc do begin
  	if StrICmp(Group[i],oldName) then begin
     	FGroups[i]:=newName;

    	IterateFields;

      Exit;
    end;
  end;
end;

procedure TINICustomSection.RenameGroupByIndex(Index:Integer;const newName:string);
begin
	RenameGroup(Group[Index],newName);
end;

function TINICustomSection.ExchangeGroups(const gNameF,gNameL:string):Boolean;
var gFI,gLI:Integer;
begin
	gFI:=GetGroupIndex(gNameF);
  gLI:=GetGroupIndex(gNameL);

  Result:=(gFI<>-1) and (gLI<>-1);

  if Result then FGroups.Exchange(gFI,gLI);
end;

function TINICustomSection.InsertGroup(const gName:string;Columns,ImageIndex:Integer;Direction:TCreateDirection;Index:Integer):Boolean;
var gI:Integer;
begin
	Result:=gName<>'';

  if Result then begin
    gI:=GetGroupIndex(gName);

    if gI=-1 then begin
      if Index=-1 then Index:=FGroups.Count;

      FGroups.InsertObject(Index,gName,TGroupInfo.CreateInfo(Columns,-1,Direction));

      Result:=False;
    end else begin
      ColumnsInGroup[gI]:=Columns;
      GroupImageIndex[gI]:=ImageIndex;
      DirectionInGroup[gI]:=Direction;
    end;
  end;
end;

procedure TINICustomSection.OpenGroupForEdit(const OpenGroup:string);
begin
  if FParentField is TINIEntry then begin
    if FParentField.Table is TINICustomSection then TINICustomSection(FParentField.Table).OpenGroupForEdit(TINIEntry(FParentField).GroupName);
  end else if Database<>nil then begin
    Database.OpenSectionForEdit(Self);
  end;

  OpenLocalGroupForEdit(OpenGroup);
end;

procedure TINICustomSection.OpenLocalGroupForEdit(const OpenGroup:string);
var tSheet:TTabSheet;

	function GetGroupPage:TTabSheet;
	var i,cc:Integer;
  		WC:TWinControl;
  begin
  	if (FPanel<>nil) and (FPanel.ControlCount>0) then begin
      WC:=FPanel.Controls[0] as TWinControl;

      cc:=WC.ControlCount-1;

      for i:=0 to cc do begin
        if WC.Controls[i] is TTabSheet then with WC.Controls[i] as TTabSheet do begin
          if AnsiCompareText(Caption,OpenGroup)=0 then begin
            Result:=WC.Controls[i] as TTabSheet;

            Exit;
          end;
        end;
      end;
    end;

    Result:=nil;
  end;

begin
	if (GroupType=gtCascade) and (GroupCount>1) then begin
  	tSheet:=GetGroupPage;

    if tSheet<>nil then (FPanel.Controls[0] as TPageControl).ActivePage:=GetGroupPage;
  end;
end;

function TINICustomSection.FieldsInGroup(const gName:string):Integer;
var i,fc:Integer;
begin
	Result:=0;

  fc:=FFields.Count-1;

  for i:=0 to fc do begin
  	if (Fields[i] is TINIEntry) and (AnsiCompareText((Fields[i] as TINIEntry).GroupName,gName)=0) then Result:=Result+1; 
  end;
end;

function TINICustomSection.DeleteGroup(const gName:string):Boolean;
var gi:Integer;
begin
	Result:=False;

	if (GroupCount>1) and (FieldsInGroup(gName)=0) then begin
    gi:=GetGroupIndex(gName);

    if gi<>-1 then begin
      RenameGroup(gName,Group[0]);

      FGroups.Delete(gi);

      Result:=True;
    end;
  end;
end;

function TINICustomSection.GetGroupCount:Integer;
begin
	Result:=FGroups.Count;
  
  if Result=0 then Result:=1;
end;

function TINICustomSection.GetGroup(Index:Integer):string;
begin
	if FGroups.Count=0 then
  	Result:=DefaultGroupName
  else
	  Result:=FGroups[Index];

  if Result='' then Result:=DefaultGroupName;
end;

function TINICustomSection.GetGroupIndex(const gName:string):Integer;
var i,gc:Integer;
begin
	gc:=GroupCount-1;

  for i:=0 to gc do begin
	  if AnsiCompareText(gName,Group[i])=0 then begin
    	Result:=i;

      Exit;
    end;
  end;

  Result:=-1;
end;

function TINICustomSection.InternalCreatePage(AOwner:TComponent;WParent:TWinControl;var NoCreated:Integer):TControl;
var pnlMain:TAutoAlignPanel;
begin
	NoCreated:=0;

	pnlMain:=TAutoAlignPanel.Create(AOwner);

	with pnlMain do begin
		Parent:=WParent;
		Align:=alClient;
		BevelInner:=bvNone;
		BevelOuter:=bvNone;
		BorderWidth:=0;
		BorderX:=0;
		BorderY:=0;
		DistanceHor:=3;
		DistanceVer:=3;
		Rows:=Self.Rows;

    Resize;
	end;

	if GroupType=gtTile then
		NoCreated:=NoCreated+CreateTileGroups(AOwner,pnlMain)
	else
		NoCreated:=NoCreated+CreateCascadeGroups(AOwner,pnlMain);

	Panel:=pnlMain;
  
  pnlMain.Resize;
  
  Result:=pnlMain;
end;

function TINICustomSection.CreateControls(AOwner:TComponent;WParent:TWinControl;GroupNo:Integer;const GroupName:string):Integer;
var LCreateInfo:TCreateInfo;

	procedure DoMain;
  var i,fc,NoCreated:Integer;
      CreatedControl:TControl;
      INIE:TINIEntry;
      State:TCreateInfoState;
      OK:Boolean;
  begin
    with LCreateInfo do begin
      fCols:=ColumnsInGroup[GroupNo];
      fDirection:=DirectionInGroup[GroupNo];

      FirstRow;
      FirstColumn;

      fW:=((WParent.ClientWidth-((fInitX shl 1)+(fCols+1)*fXBor)) div fCols);
      fH:=(WParent.ClientHeight-((fInitY+fYBor) shl 1));
    end;

    fc:=FieldCount-1;

    for i:=0 to fc do begin
      if Fields[i] is TINIEntry then begin
        INIE:=Fields[i] as TINIEntry;

        if (AnsiCompareText(INIE.GroupName,GroupName)=0) and INIE.CheckPrivilegLevel(Database.UserPrivilegLevel) and INIE.Visible then begin
          OK:=False;
          NoCreated:=0;

          State:=LCreateInfo.GetState;

          CreatedControl:=INIE.CreateControl(AOwner,WParent,LCreateInfo);

          if (INIE is TINIEntryLink) and (CreatedControl<>nil) then begin
            (INIE as TINIEntryLink).SubSection.InternalCreatePage(AOwner,CreatedControl as TWinControl,NoCreated);

            OK:=NoCreated>0;

            if not OK then begin
              INIE.FControl:=nil;

              CreatedControl.Free;
            end;
          end else if CreatedControl<>nil then begin
            NoCreated:=1;

            OK:=True;
          end;

          if not OK then begin
          	LCreateInfo.SetState(State);
          end;

          Result:=Result+NoCreated;
        end;
      end;
    end;
  end;
  
begin
	Result:=0;

  LCreateInfo:=TCreateInfo.Create(Database.FConstCreateInfo);

  try
  	DoMain;

  	finally
    	LCreateInfo.Free;
  end;
end;

function TINICustomSection.CountVisibleControls(const GroupName:string):Integer;
var i,fc:Integer;
		INIE:TINIEntry;
		iEntry:TINIField;
begin
	Result:=0;

	fc:=FieldCount-1;

	for i:=0 to fc do begin
		iEntry:=Fields[i];

		if iEntry is TINIEntry then begin
			INIE:=iEntry as TINIEntry;

			if (AnsiCompareText(INIE.GroupName,GroupName)=0) and INIE.CheckPrivilegLevel(Database.UserPrivilegLevel) and INIE.Visible then begin
				Result:=Result+1;

				Exit;
			end;
		end;
	end;
end;

procedure TINICustomSection.ArrangeIfVScroll(WControl:TWinControl);
var i,cc,VCWidth:Integer;
		WC:TControl;
begin
	if (GetWindowLong(WControl.Handle,GWL_STYLE) and WS_VSCROLL)<>0 then begin
  	cc:=WControl.ControlCount-1;

    VCWidth:=GetSystemMetrics(SM_CXVSCROLL);

  	for i:=0 to cc do begin
    	WC:=WControl.Controls[i];

      if WC.Visible then begin
        if WC.Left+WC.Width>WControl.Width-VCWidth then begin
          WC.Width:=WC.Width-VCWidth;

          if WC is TWinControl then ArrangeToFit(WC as TWinControl,5,True);
        end;
      end;
    end
  end;
end;

function TINICustomSection.CreateTileGroups(AOwner:TComponent;WParent:TAutoAlignPanel):Integer;
var i,gc:Integer;
		newG:TPanelGroup;
    sBox:TWinControl;
begin
	Result:=0;

	gc:=GroupCount-1;

	for i:=0 to gc do begin
		newG:=TPanelGroup.Create(WParent);

		with newG do begin
			Parent:=WParent as TPanel;

			Title.Font:=Font;
      Title.Font.Style:=Title.Font.Style+[fsBold];
			Title.TextFormat:=Title.TextFormat-[dtCenter];
			Title.TextFormat:=Title.TextFormat+[dtLeft];
			Title.ImageList:=Database.ImageList;
      Title.ImageIndex:=GroupImageIndex[i];

  	  Top:=(FGroupTitlePosition=gtpTop);

			TitleContext:=Group[i];

			if (GroupCount=1) and (not AlwaysShowGroup) then begin
				Title.Visible:=False;
				Line.Visible:=False;
			end;
		end;
	end;

	for i:=0 to gc do begin
		if not (CountVisibleControls(Group[i])>0) then begin
			WParent.Controls[i].Visible:=False;
		end;
	end;

	WParent.Resize;

	for i:=0 to gc do begin
		if WParent.Controls[i].Visible then begin
    	sBox:=(WParent.Controls[i] as TPanel).Controls[2] as TWinControl;

			Result:=Result+CreateControls(AOwner,sBox,i,Group[i]);
      
      ArrangeIfVScroll(sBox);
    end;
	end;
end;

function TINICustomSection.CreateCascadeGroups(AOwner:TComponent;WParent:TAutoAlignPanel):Integer;
var i,gc,gII:Integer;
		newG:TPanelGroupCascade;
    newP:TPanel;
    sBox:TScrollBox;
    tSheet,firstTab:TTabSheet;
    FImageList:TImageList;
    gName:string;
    TCI:TTCItem;
begin
	Result:=0;

	gc:=GroupCount-1;

  if gc=0 then begin
    if CountVisibleControls(Group[0])>0 then begin
      newP:=TPanel.Create(WParent);

      with newP do begin
      	Parent:=WParent;
        BevelInner:=bvNone;
        BevelOuter:=bvLowered;
        Align:=alClient;
      end;

      sBox:=TScrollBox.Create(newP);

      with sBox do begin
        Parent:=newP;
        Align:=alClient;
        BorderStyle:=bsNone;
     		HorzScrollBar.Visible:=False;
      end;

      Result:=Result+CreateControls(AOwner,sBox,gc,Group[0]);

      ArrangeIfVScroll(sBox);
    end;
	end else begin
    newG:=TPanelGroupCascade.Create(WParent);

    with newG do begin
     Parent:=WParent;
     Align:=alClient;

	   Top:=(FGroupTitlePosition=gtpTop);
    end;

    FImageList:=Database.ImageList;

    if FImageList<>nil then begin
      SendMessage(newG.Handle,TCM_SETIMAGELIST,0,FImageList.Handle);
    end;

    firstTab:=nil;
    tSheet:=nil;

    for i:=0 to gc do begin
      if CountVisibleControls(Group[i])>0 then begin
        tSheet:=TTabSheet.Create(newG);

        with tSheet do begin
         Parent:=newG;
         PageControl:=newG;
         Caption:=Group[i];
         Align:=alClient;
         Visible:=True;

         gII:=GroupImageIndex[i];

         if (FImageList<>nil) and (gII>=0) and (gII<FImageList.Count) then begin
         	gName:=Group[i];

          TCI.mask:=TCIF_TEXT or TCIF_IMAGE;
          TCI.pszText:=@(gName[1]);
          TCI.cchTextMax:=Length(gName);
          TCI.iImage:=gII;

          SendMessage(newG.Handle,TCM_SETITEM,i,LPARAM(@TCI));
         end;
        end;

        if firstTab=nil then firstTab:=tSheet;

        sBox:=TScrollBox.Create(tSheet);

        with sBox do begin
          Parent:=tSheet;
          Align:=alClient;
          BorderStyle:=bsNone;
					HorzScrollBar.Visible:=False;
        end;

        Result:=Result+CreateControls(AOwner,sBox,i,Group[i]);

        ArrangeIfVScroll(sBox);
      end;
    end;

    if firstTab<>nil then begin
      newG.ActivePage:=tSheet;
      newG.ActivePage:=firstTab;
    end else begin
    	newG.Free;
    end;
	end;
end;

function TINICustomSection.GetRuntimeSectionName:string;
begin
	Result:=GetPrivatePrefix;

  if Result<>'' then Result:=Result+'\';

  Result:=Result+FPrefix+FSectionName;
end;

function TINICustomSection.GetGroupImageIndex(Index:Integer):Integer;
begin
	Result:=-1;

	if FGroups.Objects[Index]<>nil then begin
  	Result:=TGroupInfo(FGroups.Objects[Index]).ImageIndex;
  end;
end;

function TINICustomSection.GetColumnsInGroup(Index:Integer):Integer;
begin
	Result:=1;

	if FGroups.Objects[Index]<>nil then begin
  	Result:=TGroupInfo(FGroups.Objects[Index]).Columns;
  end;
end;

function TINICustomSection.GetDirectionInGroup(Index:Integer):TCreateDirection;
begin
	Result:=cdTopToBottom;

	if FGroups.Objects[Index]<>nil then begin
  	Result:=TGroupInfo(FGroups.Objects[Index]).Direction;
  end;
end;

procedure TINICustomSection.SetGroupImageIndex(Index:Integer;II:Integer);
begin
	if (II>=-1) and (II<=255) then begin
		if FGroups.Objects[Index]<>nil then begin
  		TGroupInfo(FGroups.Objects[Index]).ImageIndex:=II;
	  end else begin
    	FGroups.Objects[Index]:=TGroupInfo.CreateInfo(1,II,cdTopToBottom);
    end;
  end;
end;

procedure TINICustomSection.SetColumnsInGroup(Index:Integer;Clmn:Integer);
begin
	if (Clmn>=1) and (Clmn<=128) then begin
		if FGroups.Objects[Index]<>nil then begin
  		TGroupInfo(FGroups.Objects[Index]).Columns:=Clmn;
	  end else begin
    	FGroups.Objects[Index]:=TGroupInfo.CreateInfo(Clmn,-1,cdTopToBottom);
    end;
  end;
end;

procedure TINICustomSection.SetSectionName(const sName:string);
begin
	FSectionName:=sName;
end;

procedure TINICustomSection.DoWriteHumanReadable(Stream:TStream;Level:Integer);
var SI:TStreamInfo;
begin
	WriteLnTextStream(Stream,Format('%sSection "%s"={',[Replicate(Level,#9),SectionName]));

	SI.Stream:=Stream;
  SI.Level:=Level+1;

	ForEachEntryStreamed(FEPWriteHumanReadable,@SI);

	WriteLnTextStream(Stream,Format('%s}',[Replicate(Level,#9)]));
end;

procedure TINICustomSection.DoReadHumanReadable(Stream:TStream;Level:Integer);
begin
end;

procedure TINICustomSection.SetDirectionInGroup(Index:Integer;CD:TCreateDirection);
begin
  if FGroups.Objects[Index]<>nil then begin
    TGroupInfo(FGroups.Objects[Index]).Direction:=CD;
  end else begin
    FGroups.Objects[Index]:=TGroupInfo.CreateInfo(1,-1,CD);
  end;
end;

function TINICustomSection.GetPrivatePrefix:string;
begin
	Result:='';

  if not Son then with Database do begin
    if RuntimePrivateMode in [pmGroupPrivate,pmUserPrivate] then Result:=MergePaths(Result,'Group\'+Names.Group);
    if RuntimePrivateMode=pmUserPrivate then Result:=MergePaths(Result,'User\'+Names.User);
  end;
end;

function TINICustomSection.GetValidFileName:string;
begin
  if IsStreamFileBased(RuntimeDestination) then begin
    with Database do begin
		  Result:=GetPrivatePath(RuntimePrivateMode,Dirs.Public,Dirs.Group,Dirs.User,RuntimeFileName);
    end;
  end else begin
  	Result:=RuntimeFileName;
  end;
end;

procedure TINICustomSection.PreparePrefix;
begin
	FPrefix:='';
  
	if Son then FPrefix:=MergePaths(TINICustomSection(FParentField.Table).RuntimeSectionName,'');
end;

procedure TINICustomSection.ReadSection;
begin
	if not EditOnlyMode then begin
  	PreparePrefix;

  	FSectionState:=FSectionState+[ssReading];

    try
			DoReadSection;

    	if Assigned(FOnAfterRead) then FOnAfterRead(Self);

      finally
		  	FSectionState:=FSectionState-[ssReading];
    end;
  end;
end;

procedure TINICustomSection.WriteSection;
begin
	if not EditOnlyMode then begin
  	PreparePrefix;

  	FSectionState:=FSectionState+[ssWriting];

    try
    	if Assigned(FOnBeforeWrite) then FOnBeforeWrite(Self);

 	  	DoWriteSection;
      
      finally
		  	FSectionState:=FSectionState-[ssWriting];
    end;
  end;
end;

procedure TINICustomSection.DoReadSection;
begin
  ForEachEntryStreamed(FEPRead,nil);

  if Assigned(FOnCalcFields) then FOnCalcFields(Self);
end;

procedure TINICustomSection.DoWriteSection;
begin
	ForEachEntryStreamed(FEPWrite,nil);
end;

procedure TINICustomSection.DeleteSection;
begin
	if not EditOnlyMode then begin
  	PreparePrefix;
    
  	FSectionState:=FSectionState+[ssDeleting];

    try
			DoDeleteSection;

      finally
		  	FSectionState:=FSectionState-[ssDeleting];
    end;
  end;
end;

procedure TINICustomSection.DoDeleteSection;
var INIFile:TINIStream;
begin
  ForEachEntryStreamed(FEPDelete,nil);

  INIFile:=CreateStream(Database.Names.Computer,GetValidFileName,RuntimeDestination);

  try
    INIFile.DeleteSection(RuntimeSectionName);

    finally
      INIFile.Free;
  end;
end;

procedure TINICustomSection.CopySection(const CopyFrom:string);
begin
	if not EditOnlyMode then begin
  	PreparePrefix;
    
  	FSectionState:=FSectionState+[ssWriting];

    try
			DoCopySection(CopyFrom);

      finally
		  	FSectionState:=FSectionState-[ssWriting];
    end;
  end;
end;

procedure TINICustomSection.DoCopySection(const CopyFrom:string);
begin
  FCopyFromFileName:=CopyFrom;

  ForEachEntryStreamed(FEPCopy,nil);
end;

procedure TINICustomSection.RefreshControls;
begin
	ForEachEntry(FEPRefresh,nil);
end;

procedure TINICustomSection.ClearControls;
begin
	ForEachEntry(FEPClear,nil);
end;

procedure TINICustomSection.Init;
begin
	if Assigned(FOnInit) then FOnInit(Self);

	ForEachEntry(FEPInit,nil);
end;

procedure TINICustomSection.Reset;
begin
	ForEachEntry(FEPReset,nil);
end;

procedure TINICustomSection.Syncronize(Direction:TSyncronizeDirection);
begin
	ForEachEntry(FEPSync,@Direction);
end;

procedure TINICustomSection.ShowControls;
begin
	ForEachEntry(FEPShow,Pointer(1));
end;

procedure TINICustomSection.HideControls;
begin
	ForEachEntry(FEPShow,nil);
end;

procedure TINICustomSection.EnableControls;
begin
	ForEachEntry(FEPEnable,Pointer(1));
end;

procedure TINICustomSection.DisableControls;
begin
	ForEachEntry(FEPEnable,nil);
end;

procedure TINICustomSection.Show(Show:Boolean);
begin
	if Show then
		ShowControls
	else
		HideControls;
end;

procedure TINICustomSection.EnablePanel(State:Boolean);
begin
	if FPanel<>nil then FPanel.Enabled:=State;
end;

procedure TINICustomSection.Enable(State:Boolean);
begin
	if State then
		EnableControls
	else
		DisableControls;
end;

procedure TINICustomSection.ForEachEntry(FEP:TForEachProcedure;Data:Pointer);
var i:Integer;
		iEntry:TINIField;
begin
	for i:=0 to FieldCount-1 do begin
		iEntry:=Fields[i];

		if iEntry is TINIEntry then FEP(iEntry,Data);
	end;
end;

procedure TINICustomSection.ForEachEntryStreamed(FEP:TForEachProcedure;Data:Pointer);
var INIFile:TINIStream;
begin
	INIFile:=CreateStream(Database.Names.Computer,GetValidFileName,RuntimeDestination);

  FINIStream:=INIFile;

	try
  	ForEachEntry(FEP,Data);

		finally
      FINIStream:=nil;

			INIFile.Free;
	end;
end;

procedure TINICustomSection.FEPRefresh(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).RefreshControl;
end;

procedure TINICustomSection.FEPClear(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).ClearControl;
end;

procedure TINICustomSection.FEPInit(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).Init;
end;

procedure TINICustomSection.FEPReset(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).Reset;
end;

procedure TINICustomSection.FEPShow(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).ShowControl(Data<>nil);
end;

procedure TINICustomSection.FEPEnable(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).EnableControl(Data<>nil);
end;

procedure TINICustomSection.FEPSync(Item:TObject;Data:Pointer);
var Direction:^TSyncronizeDirection;
begin
	Direction:=Data;

	TINIEntry(Item).Syncronize(Direction^);
end;

procedure TINICustomSection.FEPRead(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).ReadEntry;
end;

procedure TINICustomSection.FEPWrite(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).WriteEntry;
end;

procedure TINICustomSection.FEPDelete(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).DeleteEntry;
end;

procedure TINICustomSection.FEPCopy(Item:TObject;Data:Pointer);
begin
	TINIEntry(Item).CopyEntry;
end;

procedure TINICustomSection.FEPWriteHumanReadable(Item:TObject;Data:Pointer);
var pSI:^TStreamInfo;
begin
	pSI:=Data;

	TINIEntry(Item).DoWriteHumanReadable(pSI^.Stream,pSI^.Level);
end;

procedure TINICustomSection.FEPReadHumanReadable(Item:TObject;Data:Pointer);
begin
end;

function TINISection.StoreGroupInfo:Boolean;
begin
	Result:=True;
end;

constructor TINICustomMultiSection.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FPanelMode:=pmComboList;

	FOnAfterReadKeys:=nil;
	FOnChange:=nil;
	FOnButtonClick:=nil;

	List:=nil;

	FStatic:=False;
  FSorted:=False;
  FNewAsCopy:=False;
	FCompressKey:=ciNo;
  FButtons:=[isSave,isDelete,isClear,isChange,isExchange];

	FDropDownCount:=8;

	FSections:=TStringList.Create;
	FStaticSections:=TStringList.Create;

  FSections.OnChange:=SectionsChanged;

	FSelected:=-1;
  FEOL:=True;
  FIsDirty:=False;
  FWalking:=False;
  FUseMainSection:=False;

	FAllowListChange:=True;
end;

destructor TINICustomMultiSection.Destroy;
begin
	FSections.Free;
  FStaticSections.Free;

	inherited Destroy;
end;

function TINICustomMultiSection.InEditMode:Boolean;
begin
	Result:=(List<>nil);
end;

procedure TINICustomMultiSection.SetSorted(bSorted:Boolean);
begin
	FSorted:=bSorted;

  FSections.Sorted:=FSorted;

  if List<>nil then List.Combo.Sorted:=FSorted;
end;

function TINICustomMultiSection.InternalCreatePage(AOwner:TComponent;WParent:TWinControl;var NoCreated:Integer):TControl;
var pnlTop:TPanel;
		pnlMain:TAutoAlignPanel;
begin
	pnlTop:=TPanel.Create(AOwner);

	with pnlTop do begin
		Parent:=WParent;
		Align:=alClient;
		BevelInner:=bvNone;
		BevelOuter:=bvNone;
		BorderWidth:=2;
	end;

	pnlMain:=TAutoAlignPanel.Create(AOwner);

	with pnlMain do begin
		Parent:=pnlTop;
		Align:=alClient;
		BevelInner:=bvNone;
		BevelOuter:=bvNone;
		BorderWidth:=0;
		BorderX:=0;
		BorderY:=0;
		DistanceHor:=3;
		DistanceVer:=3;
		Rows:=Self.Rows;

    Resize;
	end;

  List:=TPanelList.CreatePM(AOwner,PanelMode);

  if PanelMode=pmComboList then begin
	  with List do begin
    	Parent:=pnlTop;
      Align:=alTop;

      OnListChange:=DoOnListChange;
      OnNameChange:=DoOnNameChange;
      OnButtonSaveClick:=DoOnButtonSaveClick;
      OnButtonDeleteClick:=DoOnButtonDeleteClick;
      OnButtonClearClick:=DoOnButtonClearClick;

      SpinButtonChange.OnDownClick:=SenderNextItem;
      SpinButtonChange.OnUpClick:=SenderPrevItem;

      SpinButton.OnDownClick:=SenderMoveDown;
      SpinButton.OnUpClick:=SenderMoveUp;

      SpinButtonChange.Visible:=isChange in Buttons;
      SpinButton.Visible:=(isExchange in Buttons) and not Sorted;
    end;

		List.Combo.Sorted:=Sorted;
    List.Combo.DropDownCount:=DropDownCount;
    List.Combo.OnKeyDown:=ComboKeyDown;

    if Static then
      List.Combo.Style:=csDropDownList
    else
      List.Combo.Style:=csDropDown;

    with Database do begin
	    FGlyphs.Save.GAssignTo(List.ButtonSave,isSave in Buttons);
  	  FGlyphs.Delete.GAssignTo(List.ButtonDelete,isDelete in Buttons);
    	FGlyphs.Clear.GAssignTo(List.ButtonClear,isClear in Buttons);
    end;
  end else begin
  	List.Parent:=pnlTop;
    List.Align:=alClient;
  end;

  List.Resize;

	if GroupType=gtTile then
		NoCreated:=1+CreateTileGroups(AOwner,pnlMain)
	else
		NoCreated:=1+CreateCascadeGroups(AOwner,pnlMain);

	Panel:=pnlMain;

	RefreshList;

	pnlMain.Resize;

  Result:=pnlTop;
end;

procedure TINICustomMultiSection.DoWriteHumanReadable(Stream:TStream;Level:Integer);
var SI:TStreamInfo;
begin
	WriteLnTextStream(Stream,Format('%sMultiSection "%s"={',[Replicate(Level,#9),SectionName]));

	SI.Stream:=Stream;
  SI.Level:=Level+1;

  BeginWalk;

  try
    First;

    while not EOL do begin
	    WriteLnTextStream(Stream,Format('%sItem "%s"={',[Replicate(Level+1,#9),Key]));

      SI.Level:=SI.Level+1;

      try
        ForEachEntryStreamed(FEPWriteHumanReadable,@SI);

        finally
          SI.Level:=SI.Level-1;
      end;

      WriteLnTextStream(Stream,Format('%s}',[Replicate(Level+1,#9)]));

      Next;
    end;

    finally
    	EndWalk;
  end;

	WriteLnTextStream(Stream,Format('%s}',[Replicate(Level,#9)]));
end;

procedure TINICustomMultiSection.DoReadHumanReadable(Stream:TStream;Level:Integer);
begin
end;

procedure TINICustomMultiSection.OpenLocalGroupForEdit(const OpenGroup:string);
begin
	inherited OpenLocalGroupForEdit(OpenGroup);
end;

procedure TINICustomMultiSection.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if Operation=opRemove then begin
  	if AComponent=FPanelList then FPanelList:=nil;
  end;
end;

function TINICustomMultiSection.CanDelete(Index:Integer):Boolean;
begin
	Result:=FStaticSections.IndexOf(GetKeyByIndex(Index))=-1;
end;

procedure TINICustomMultiSection.SetPanelList(PL:TPanelList);
begin
	FPanelList:=PL;

  if PL<>nil then PL.FreeNotification(Self);
end;

procedure TINICustomMultiSection.SectionsChanged(Sender:TObject);
begin
	FIsDirty:=True;
end;

procedure TINICustomMultiSection.DoOnListChange(Sender:TObject);
begin
 	ListChange;

	if Assigned(FOnChange) then FOnChange(Sender);
end;

procedure TINICustomMultiSection.DoOnNameChange(Sender:TObject);
begin
 	NameChange;
end;

procedure TINICustomMultiSection.DoOnButtonSaveClick(Sender:TObject);
begin
 	ButtonSaveClick;

	if Assigned(FOnButtonClick) then FOnButtonClick(Sender,isSave);
end;

procedure TINICustomMultiSection.DoOnButtonDeleteClick(Sender:TObject);
begin
	ButtonDeleteClick;
  
	if Assigned(FOnButtonClick) then FOnButtonClick(Sender,isDelete);
end;

procedure TINICustomMultiSection.DoOnButtonClearClick(Sender:TObject);
begin
 	ButtonClearClick;

	if Assigned(FOnButtonClick) then FOnButtonClick(Sender,isClear);
end;

procedure TINICustomMultiSection.ComboKeyDown(Sender:TObject;var Key:Word;Shift:TShiftState);
begin
	if ssCtrl in Shift then begin
		if Key=VK_DOWN then begin
			Key:=0;

			MoveDown;
		end else if Key=VK_UP then begin
			Key:=0;

			MoveUp;
		end;
	end;

  EditKeyDown(Sender,Key,Shift);
end;

procedure TINICustomMultiSection.BeginWalk;
begin
	FWalking:=True;

  FSaveIndex:=KeyIndex;
end;

procedure TINICustomMultiSection.EndWalk;
begin
	if FWalking then begin
		FWalking:=False;

    KeyIndex:=FSaveIndex;
    
    if InEditMode then begin
    	List.Items.BeginUpdate;

      try
	    	List.Items.Assign(Sections);
				List.ItemIndex:=List.Items.IndexOf(Key);

        finally
		    	List.Items.EndUpdate;
      end;

      RefreshList;
    end;
  end;
end;

procedure TINICustomMultiSection.EditKeyDown(Sender:TObject;var Key:Word;Shift:TShiftState);
begin
	if Key=VK_F2 then begin
		Key:=0;

    ButtonSaveClick;
  end else if Key=VK_F5 then begin
		Key:=0;

    if SectionCount>0 then Database.GoSearchSection(Self);
	end else if (ssCtrl in Shift) and (Key=VK_DELETE) then begin
		Key:=0;

		ButtonDeleteClick;
	end;
end;

procedure TINICustomMultiSection.SenderMoveDown(Sender:TObject);
begin
	MoveDown;
end;

procedure TINICustomMultiSection.SenderMoveUp(Sender:TObject);
begin
	MoveUp;
end;

procedure TINICustomMultiSection.MoveDown;
var ii:LongInt;
begin
	if not Sorted then begin
    ii:=List.ItemIndex;

    if ii<List.Items.Count-1 then begin
      try
        List.Items.Exchange(ii,ii+1);
        Sections.Exchange(ii,ii+1);

        except
          on EListError do begin
          end;
      end;

      List.ItemIndex:=ii+1;
      Selected:=ii+1;
    end;
  end;
end;

procedure TINICustomMultiSection.MoveUp;
var ii:LongInt;
begin
	if not Sorted then begin
    ii:=List.ItemIndex;

    if ii>0 then begin
      try
        List.Items.Exchange(ii,ii-1);
        Sections.Exchange(ii,ii-1);

        except
          on EListError do begin
          end;
      end;

      List.ItemIndex:=ii-1;
      Selected:=ii-1;
    end;
  end;
end;

procedure TINICustomMultiSection.SenderNextItem(Sender:TObject);
begin
	NextItem;
end;

procedure TINICustomMultiSection.SenderPrevItem(Sender:TObject);
begin
	PrevItem;
end;

procedure TINICustomMultiSection.NextItem;
var ii:LongInt;
begin
	ii:=List.ItemIndex;

	if ii<List.Items.Count-1 then begin
		List.ItemIndex:=ii+1;

		DoOnListChange(Self);
	end;
end;

procedure TINICustomMultiSection.PrevItem;
var ii:LongInt;
begin
	ii:=List.ItemIndex;

	if ii>0 then begin
		List.ItemIndex:=ii-1;

		DoOnListChange(Self);
	end;
end;

procedure TINICustomMultiSection.AssignStatic;
begin
  AssignSections(FStaticSections);
end;

procedure TINICustomMultiSection.ExpandWithStatic;
var i,ssc:Integer;
begin
	FSections.BeginUpdate;

  try
    ssc:=FStaticSections.Count-1;

    for i:=0 to ssc do begin
      if FSections.IndexOf(FStaticSections[i])=-1 then FSections.Add(FStaticSections[i]);
    end;

    if (FSelected=-1) and (FSections.Count>0) then FSelected:=0;

    finally
			FSections.EndUpdate;
  end;
end;

procedure TINICustomMultiSection.Loaded;
begin
	inherited Loaded;

  AssignStatic;
  
	if FStatic or (FStaticSections.Count>0) then FSelected:=0;
end;

procedure TINICustomMultiSection.SetStaticSections(sSections:TStringList);
begin
	FStaticSections.Assign(sSections);
end;

procedure TINICustomMultiSection.DeleteKey(const sKey:string);
begin
	DeleteThisSection(IndexOfKey[sKey]);
end;

procedure TINICustomMultiSection.DeleteKeyByIndex(Index:Integer);
begin
	DeleteThisSection(Index);
end;

function TINICustomMultiSection.KeyExists(const sKey:string):Boolean;
begin
	Result:=FSections.IndexOf(sKey)<>-1;
end;

function TINICustomMultiSection.FindKey(const sKey:string):Boolean;
begin
	Result:=KeyExists(sKey);

	if Result then Key:=sKey;
end;

function TINICustomMultiSection.AddKey(const sKey:string):Boolean;
begin
	Result:=KeyExists(sKey);

	if not Result then FSections.Add(sKey);

  Key:=sKey;
end;

function TINICustomMultiSection.GetIndexOfKey(const sKey:string):Integer;
begin
	Result:=Sections.IndexOf(sKey);
end;

procedure TINICustomMultiSection.AssignSections(sList:TStrings);
begin
  FSections.Assign(sList);

  FIsDirty:=True;
end;

procedure TINICustomMultiSection.AssignToSections(sMultiSection:TINICustomMultiSection);
begin
	sMultiSection.Sections.Assign(Sections);
end;

procedure TINICustomMultiSection.AssignToStrings(sList:TStrings);
begin
	sList.Assign(FSections);
end;

function TINICustomMultiSection.GetRuntimeSectionName:string;
begin
	Result:=inherited GetRuntimeSectionName;

  if not FUseMainSection then begin
  	try
	  	Result:=MergePaths(Result,SectionByIndex[FSelected]);

      except
      	on EListError do begin
        end;
    end;
  end;
end;

function TINICustomMultiSection.GetSectionCount:Integer;
begin
	Result:=FSections.Count;
end;

function TINICustomMultiSection.GetKey:string;
begin
	if (Selected>=0) and (Selected<Sections.Count) then
		Result:=Sections.Strings[Selected]
	else
		Result:='';
end;

function TINICustomMultiSection.GetKeyIndex:Integer;
begin
	Result:=Selected;
end;

procedure TINICustomMultiSection.SetKey(const sKey:string);
begin
	KeyIndex:=Sections.IndexOf(sKey);
end;

procedure TINICustomMultiSection.SetKeyIndex(Index:Integer);
begin
	if (Index>=0) and (Index<Sections.Count) then begin
		Selected:=Index;

		ReReadSection;
    
    FEOL:=False;
	end else begin
  	Reset;

    FEOL:=True;
  end;

  FIsDirty:=True;
end;

function TINICustomMultiSection.GetSectionByIndex(Index:Integer):string;
begin
	if CompressKey=ciNo then
		Result:=Sections.Strings[Index]
  else if CompressKey=ciExcludeBlanks then
		Result:=OnlyAlnum(Sections.Strings[Index])
	else
		Result:=DoCompressKey(Sections.Strings[Index]);
end;

function TINICustomMultiSection.DoCompressKey(const sKey:string):string;
begin
	Result:=MakeShortSection(sKey);
end;

function TINICustomMultiSection.GetKeyByIndex(Index:Integer):string;
begin
	if (Index>=0) and (Index<Sections.Count) then
		Result:=Sections[Index]
	else
		Result:='';
end;

function TINICustomMultiSection.GetKeyObject:TObject;
begin
	if (Selected>=0) and (Selected<Sections.Count) then
		Result:=Sections.Objects[Selected]
	else
		Result:=nil;
end;

procedure TINICustomMultiSection.ReadKeysAlways;
var INIFile:TINIStream;
		fName:string;
begin
	fName:=GetValidFileName;

  if not EditOnlyMode then begin
		Sections.Clear;

    INIFile:=CreateStream(Database.Names.Computer,fName,RuntimeDestination);

    try
    	INIFile.ReadEntries(inherited GetRuntimeSectionName,Sections,FSelected);

	    ExpandWithStatic;

      finally
      	INIFile.Free;
    end;
  end;

	if Assigned(FOnAfterReadKeys) then FOnAfterReadKeys(Self);
end;

procedure TINICustomMultiSection.WriteKeysAlways;
var INIFile:TINIStream;
		fName:string;
begin
	fName:=GetValidFileName;

  if not EditOnlyMode then begin
	  if (PanelMode=pmMemo) and InEditMode then AssignSections(List.Items);

    if FIsDirty then begin
      INIFile:=CreateStream(Database.Names.Computer,fName,RuntimeDestination);

      try
        INIFile.WriteEntries(inherited GetRuntimeSectionName,Sections,Selected);

        finally
          INIFile.Free;
      end;
    end;
  end;
end;

procedure TINICustomMultiSection.WriteOnlySelected;
var INIFile:TINIStream;
		fName:string;
begin
	fName:=GetValidFileName;

  if not EditOnlyMode then begin
    if FIsDirty then begin
      INIFile:=CreateStream(Database.Names.Computer,fName,RuntimeDestination);

      try
      	if Selected=0 then
	        INIFile.DeleteEntry(inherited GetRuntimeSectionName,CEntrySelected)
        else
	        INIFile.WriteInteger(inherited GetRuntimeSectionName,CEntrySelected,Selected);

        finally
          INIFile.Free;
      end;
    end;
  end;
end;

procedure TINICustomMultiSection.ReadOnlySelected;
var INIFile:TINIStream;
		fName:string;
begin
	fName:=GetValidFileName;

  if not EditOnlyMode then begin
    INIFile:=CreateStream(Database.Names.Computer,fName,RuntimeDestination);

    try
      FSelected:=INIFile.ReadInteger(inherited GetRuntimeSectionName,CEntrySelected,0);

      if FSelected>=Sections.Count then FSelected:=-1; 

      finally
        INIFile.Free;
    end;
  end;
end;

procedure TINICustomMultiSection.ReadKeys;
begin
	if not Static then begin
  	ReadKeysAlways;
  end else begin
  	ReadOnlySelected;
  end;
  
 	FIsDirty:=False;
end;

procedure TINICustomMultiSection.WriteKeys;
begin
	if not Static then
  	WriteKeysAlways
  else
  	WriteOnlySelected;
end;

procedure TINICustomMultiSection.DeleteKeys;
begin
	FUseMainSection:=True;

  try
		inherited DoDeleteSection;

    finally
    	FUseMainSection:=False;
  end;
end;

procedure TINICustomMultiSection.DoWriteSection;
begin
	WriteCurrentSection;
	WriteKeys;
end;

procedure TINICustomMultiSection.WriteCurrentSection;
begin
	WriteThisSection(Selected);
end;

procedure TINICustomMultiSection.DoDeleteSection;
var i,sc:Integer;
begin
  ReadKeysAlways;

  sc:=Sections.Count-1;

  for i:=sc downto 0 do begin
    DeleteThisSection(i);
  end;

  DeleteKeys;
end;

procedure TINICustomMultiSection.DoReadSection;
begin
	ReadKeys;
	ReadThisSection(Selected);
end;

procedure TINICustomMultiSection.DoCopySection(const CopyFrom:string);
var i,sc:Integer;
begin
	FIsDirty:=True;
  
	WriteThisSection(Selected);
	WriteKeys;

	sc:=Sections.Count-1;

	for i:=0 to sc do begin
		if i<>Selected then begin
			ReadThisSection(i);
			CopyThisSection(CopyFrom,i);
		end;
	end;

	ReadThisSection(Selected);
	CopyThisSection(CopyFrom,Selected);
end;

procedure TINICustomMultiSection.DoForThisSection(Index:Integer;SP:TSectionProcedure;Data:Pointer);
var SaveSelected:Integer;
begin
	if (Index>=0) and (Index<Sections.Count) then begin
	 	SaveSelected:=FSelected;

    FSelected:=Index;
    
		try
      SP(Data);

			finally
        FSelected:=SaveSelected;
		end;
	end;
end;

procedure TINICustomMultiSection.WriteThisSection(Index:Integer);
begin
	DoForThisSection(Index,SPWrite,nil);
end;

procedure TINICustomMultiSection.ReadThisSection(Index:Integer);
begin
	DoForThisSection(Index,SPRead,nil);
end;

procedure TINICustomMultiSection.DeleteThisSection(Index:Integer);
begin
  if Index>=0 then begin
    DoForThisSection(Index,SPDelete,nil);

    Sections.Delete(Selected);

    if Sections.Count=0 then
      Selected:=-1
    else if Selected>=Sections.Count then
      Selected:=Selected-1;

    FIsDirty:=True;
  end;
end;

procedure TINICustomMultiSection.CopyThisSection(const CopyFrom:string;Index:Integer);
begin
	DoForThisSection(Index,SPCopy,@CopyFrom);
end;

procedure TINICustomMultiSection.SPWrite(Data:Pointer);
begin
	inherited DoWriteSection;
end;

procedure TINICustomMultiSection.SPRead(Data:Pointer);
begin
	inherited DoReadSection;
end;

procedure TINICustomMultiSection.SPDelete(Data:Pointer);
begin
	inherited DoDeleteSection;
end;

procedure TINICustomMultiSection.SPCopy(Data:Pointer);
var CopyFrom:^string;
begin
	CopyFrom:=Data;

	inherited DoCopySection(CopyFrom^);
end;

procedure TINICustomMultiSection.CopySectionFromTo(fromS,toS:Integer);
var INIFile:TINIStream;
		fromK,toK:string;
begin
  if (fromS>=0) and (toS>=0) then begin
    INIFile:=CreateStream(Database.Names.Computer,GetValidFileName,RuntimeDestination);

    try
      fromK:=MergePaths(inherited GetRuntimeSectionName,SectionByIndex[fromS]);
      toK:=MergePaths(inherited GetRuntimeSectionName,SectionByIndex[toS]);

      INIFile.CopySection(fromK,toK);

      finally
        INIFile.Free;
    end;
  end;
end;

procedure TINICustomMultiSection.RefreshList;
begin
	if InEditMode then begin
		List.Items.Assign(Sections);
		List.ItemIndex:=List.Items.IndexOf(Key);

    if FPanelMode=pmComboList then begin
      List.SpinButton.Enabled:=List.Items.Count>0;
      List.SpinButtonChange.Enabled:=List.Items.Count>0;

      EnablePanel(List.Items.Count>0);

      if List.Items.Count>0 then begin
        List.ButtonDelete.Enabled:=CanDelete(Selected);
        List.ButtonSave.Enabled:=True;
        List.ButtonClear.Enabled:=True;
      end else begin
        List.ButtonDelete.Enabled:=False;
        List.ButtonSave.Enabled:=False;
        List.ButtonClear.Enabled:=False;
      end;
    end;

    List.AdjustControls;
	end;
end;

procedure TINICustomMultiSection.ShowControls;
begin
	inherited ShowControls;

  ShowList;
end;

procedure TINICustomMultiSection.HideControls;
begin
	inherited HideControls;

  HideList;
end;

procedure TINICustomMultiSection.EnableControls;
begin
	inherited EnableControls;

  EnableList;
end;

procedure TINICustomMultiSection.DisableControls;
begin
	inherited DisableControls;

  DisableList;
end;

procedure TINICustomMultiSection.RefreshControls;
begin
	inherited RefreshControls;

	RefreshList;
end;

procedure TINICustomMultiSection.ClearControls;
begin
	inherited ClearControls;

  ClearList;
end;

procedure TINICustomMultiSection.Reset;
begin
	inherited Reset;
end;

procedure TINICustomMultiSection.ClearList;
begin
	if InEditMode then begin
   	List.Items.Assign(FStaticSections);

    if FStaticSections.Count>0 then
	   	List.Text:=FStaticSections[0]
    else
    	List.Text:='';
  end;
end;

procedure TINICustomMultiSection.ShowList;
begin
	if InEditMode then List.Visible:=True;
end;

procedure TINICustomMultiSection.HideList;
begin
	if InEditMode then List.Visible:=False;
end;

procedure TINICustomMultiSection.EnableList;
begin
	if InEditMode then begin
		List.Enabled:=True;

		EnablePanel(List.Items.Count>0);
	end;
end;

procedure TINICustomMultiSection.DisableList;
begin
	if InEditMode then List.Enabled:=False;
end;

procedure TINICustomMultiSection.ReReadSection;
begin
	ReadThisSection(Selected);

  if (not FWalking) and InEditMode then inherited RefreshControls;
end;

procedure TINICustomMultiSection.First;
begin
	FEOL:=Sections.Count=0;

  if not FEOL then KeyIndex:=0;
end;

procedure TINICustomMultiSection.Last;
begin
	FEOL:=Sections.Count=0;

  if not FEOL then KeyIndex:=Sections.Count-1;
end;

procedure TINICustomMultiSection.Next;
begin
	FEOL:=KeyIndex=Sections.Count-1;

  if not FEOL then KeyIndex:=KeyIndex+1;
end;

procedure TINICustomMultiSection.Prev;
begin
  FEOL:=Sections.Count=0;
  
	if KeyIndex>0 then KeyIndex:=KeyIndex-1;
end;

procedure TINICustomMultiSection.RefreshEnablers;
begin
	if FPanelMode=pmComboList then begin
		List.SpinButton.Enabled:=List.Items.Count>0;
		List.SpinButtonChange.Enabled:=List.Items.Count>0;
		List.ButtonSave.Enabled:=AllTrim(List.Combo.Text)<>'';
		List.ButtonDelete.Enabled:=(List.Items.Count>0) and CanDelete(Selected);
		List.ButtonClear.Enabled:=List.ButtonSave.Enabled;

		EnablePanel(List.Items.Count>0);
	end;
end;

procedure TINICustomMultiSection.InsertSelected(const secName:string);
var pSelected:LongInt;
begin
	pSelected:=Selected;

  if Sorted then begin
  	Sections.Add(secName);
    List.Items.Add(secName);

    Selected:=Sections.IndexOf(secName);
  end else begin
    with List.Items do begin
      if Count=0 then
        Insert(0,secName)
      else
        Insert(pSelected+1,secName);
    end;

    Sections.Insert(pSelected+1,secName);

    Selected:=pSelected+1;
  end;

  List.ItemIndex:=List.Items.IndexOf(Key);
end;

procedure TINICustomMultiSection.DeleteListItem(Index:Integer);
begin
	if Index>=0 then List.Items.Delete(Index);

	FAllowListChange:=False;

  try
    List.ItemIndex:=List.Items.IndexOf(Key);

    finally
			FAllowListChange:=True;
  end;
end;

procedure TINICustomMultiSection.ListChange;
begin
	if FAllowListChange then begin
  	FIsDirty:=True;

		WriteThisSection(Selected);

    Key:=List.Text;

		ReReadSection;
		RefreshEnablers;
	end;
end;

procedure TINICustomMultiSection.NameChange;
begin
	RefreshEnablers;
end;

procedure TINICustomMultiSection.SaveSection(const sName:string);
var ii:LongInt;
    wasSelected:LongInt;
begin
  if sName<>'' then begin
    ii:=Sections.IndexOf(sName);

    if ii=-1 then begin
      WriteThisSection(Selected);

      wasSelected:=Selected;

      InsertSelected(sName);

      if FNewAsCopy then
      	CopySectionFromTo(wasSelected,Selected)
      else
				inherited ClearControls;

      ReadThisSection(Selected);
    end else begin
      Selected:=ii;
      List.ItemIndex:=List.Items.IndexOf(Key);

      WriteThisSection(Selected);
    end;
  end;

	RefreshEnablers;
end;

procedure TINICustomMultiSection.ButtonSaveClick;
begin
	WriteKeys;
	SaveSection(AllTrim(List.Text));
end;

procedure TINICustomMultiSection.ButtonDeleteClick;
var lSelected:Integer;
		sSelected:string;
begin
	lSelected:=Selected;
  sSelected:=Key;
  
	if CanDelete(lSelected) then begin
    if Selected<>-1 then begin
      DeleteThisSection(lSelected);
      DeleteListItem(List.Items.IndexOf(sSelected));

      ReReadSection;
    end;

    if Selected=-1 then ClearControls;

    RefreshEnablers;
    WriteKeys;
  end;
end;

procedure TINICustomMultiSection.ButtonClearClick;
begin
	Reset;
  RefreshControls;
end;

function TINIMultiSection.StoreGroupInfo:Boolean;
begin
	Result:=True;
end;

{
*******************************************************
	INIEntry
********************************************************
}

constructor TINIEntry.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FPrivilegLevel:=0;
	FOnInit:=nil;
	FControl:=nil;
	FLabel:=nil;
	FGroupNo:=0;
	FGroupName:='';
	FHint:='';
  FHelpContext:=0;
  FForceWrite:=False;
  FCaptionStyle:=csLeftCenter;
  FDisplayWidth:=GetDefaultWidth;
end;

procedure TINIEntry.Loaded;
begin
	inherited Loaded;

  Init;
	Reset;
end;

procedure TINIEntry.SetPrivilegLevel(Level:Integer);
begin
	FPrivilegLevel:=Max(0,Min(Level,CMaxPrivilegLevel));
end;

function TINIEntry.GetPrivilegLevel:Integer;
begin
	if Owner is TINICustomSection then
		Result:=(Owner as TINICustomSection).FPrivilegLevel
  else
		Result:=FPrivilegLevel;
end;

procedure TINIEntry.Assign(Source:TPersistent);
var INIE:TINIEntry;
begin
	if Source is TINIEntry then begin
  	INIE:=Source as TINIEntry;

    Calculated:=INIE.FCalculated;
    ReadOnly:=INIE.FReadOnly;
    Visible:=INIE.FVisible;
		GroupName:=INIE.GroupName;
		PrivilegLevel:=INIE.PrivilegLevel;
 		Hint:=INIE.FHint;
 		HelpContext:=INIE.FHelpContext;
    Index:=INIE.Index;
    CaptionStyle:=INIE.FCaptionStyle;
    DisplayLabel:=INIE.FDisplayLabel;
    DisplayWidth:=INIE.FDisplayWidth;

    DefaultAsString:=INIE.DefaultAsString;

    FOnInit:=INIE.OnInit;
    FOnChange:=INIE.OnChange;
    FOnFieldNameChange:=INIE.OnFieldNameChange;
    FOnAfterRead:=INIE.OnAfterRead;
    FOnBeforeWrite:=INIE.OnBeforeWrite;
    FOnStoreIt:=INIE.OnStoreIt;

    AsString:=INIE.AsString;
  end else inherited Assign(Source);
end;

procedure TINIEntry.ClearControl;
begin
	Reset;
	RefreshControl;
end;

procedure TINIEntry.RefreshControl;
begin
end;

procedure TINIEntry.Init;
begin
	if Assigned(FOnInit) then FOnInit(Self);
end;

procedure TINIEntry.Reset;
begin
end;

procedure TINIEntry.Validate;
begin
end;

function TINIEntry.IsControlEnabled:Boolean;
begin
	Result:=(FControl<>nil) and FControl.Enabled;
end;

function TINIEntry.IsPrivilegLevelStored:Boolean;
begin
	Result:=FPrivilegLevel<>0;
end;

procedure TINIEntry.DoWriteHumanReadable(Stream:TStream;Level:Integer);
begin
	if StoreIt then WriteLnTextStream(Stream,Format('%sField "%s"="%s"',[Replicate(Level,#9),strIf(FieldName<>'',FieldName,'Default'),AsString]));
end;

procedure TINIEntry.DoReadHumanReadable(Stream:TStream;Level:Integer);
begin
end;

function TINIEntry.DoGetAsBoolean:Boolean;
begin
	Result:=FValue;
end;

function TINIEntry.DoGetAsFloat:Double;
begin
	Result:=FValue;
end;

function TINIEntry.DoGetAsInteger:LongInt;
begin
	Result:=FValue;
end;

function TINIEntry.DoGetAsString:string;
begin
	Result:=FValue;
end;

procedure TINIEntry.DoSetAsBoolean(Value:Boolean);
begin
	FValue:=Value;
end;

procedure TINIEntry.DoSetAsFloat(Value:Double);
begin
	FValue:=Value;
end;

procedure TINIEntry.DoSetAsInteger(Value:LongInt);
begin
	FValue:=Value;
end;

procedure TINIEntry.DoSetAsString(const Value:string);
begin
	FValue:=Value;
end;

procedure TINIEntry.SetControl(C:TControl);
begin
	FControl:=C;

  if C<>nil then C.FreeNotification(Self);
end;

procedure TINIEntry.EnableControl(State:Boolean);
begin
	if FControl<>nil then begin
		if Calculated or not CanModify then
			FControl.Enabled:=False
		else
			FControl.Enabled:=State;

    if FLabel<>nil then FLabel.Enabled:=FControl.Enabled;
	end;
end;

procedure TINIEntry.ShowControl(Show:Boolean);
begin
	if FControl<>nil then begin
		if Visible then begin
    	FControl.Visible:=Show;

      if FLabel<>nil then FLabel.Visible:=Show;
    end;
	end;
end;

procedure TINIEntry.ShowError(const FunctionName:string);
begin
	raise Exception.Create(Format(ResStr(CResPVFCalled),[FunctionName]));
end;

function TINIEntry.DoGetStoreIt:Boolean;
begin
	Result:=False;
end;

function TINIEntry.GetStoreIt:Boolean;
begin
	Result:=True;

	if Assigned(FOnStoreIt) then FOnStoreIt(Self,Result);

	Result:=Result and DoGetStoreIt;
end;

procedure TINIEntry.DoReadEntry;
begin
end;

procedure TINIEntry.DoWriteEntry;
begin
end;

procedure TINIEntry.DoDeleteEntry;
begin
end;

procedure TINIEntry.DoCopyEntry;
begin
end;

procedure TINIEntry.ReadEntry;
var evEntry:TINIFieldNotifyEvent;
begin
	FIsDirty:=False;

  Reset;

  DoReadEntry;

  evEntry:=OnAfterRead;

  if Assigned(evEntry) then evEntry(Self);
end;

function TINIEntry.CheckPrivilegLevel(Level:Integer):Boolean;
begin
	Result:=Level<=PrivilegLevel;
end;

procedure TINIEntry.WriteEntry;
var evEntry:TINIFieldNotifyEvent;
begin
	if CheckPrivilegLevel(Table.Database.UserPrivilegLevel) then try
    evEntry:=FOnBeforeWrite;

    if Assigned(evEntry) then evEntry(Self);

    if IsDirty then begin
      if StoreIt or ForceWrite then
        DoWriteEntry
      else
        DoDeleteEntry;
    end;

    except
      on Exception do begin
      end;
  end;
end;

procedure TINIEntry.DeleteEntry;
begin
	DoDeleteEntry;
end;

procedure TINIEntry.CopyEntry;
begin
	FIsDirty:=True;
  
	WriteEntry;
end;

procedure TINIEntry.SetGroupNo(gNo:Integer);
var OK:Boolean;
begin
  if Table<>nil then begin
  	OK:=True;

  	try
		  FGroupName:=(Table as TINICustomSection).Group[gNo];

      except
      	on Exception do OK:=False;
    end;

    if OK then FGroupNo:=gNo;
  end else begin
  	FGroupNo:=gNo;
  end;
end;

procedure TINIEntry.SetGroupName(const gName:string);
var gNo:Integer;
begin
  if Table<>nil then begin
   	gNo:=(Table as TINICustomSection).GetGroupIndex(gName);

    if gNo<>-1 then begin
    	FGroupNo:=gNo;
      FGroupName:=gName;
    end;
  end else begin
  	FGroupName:=gName;
  end;
end;

function TINIEntry.GetGroupName:string;
var INIS:TINICustomSection;
begin
	Result:=FGroupName;

  if Result='' then begin
  	if Table is TINICustomSection then begin
      INIS:=Table as TINICustomSection;

      if FGroupNo<INIS.GroupCount then
        Result:=INIS.Group[FGroupNo]
      else if INIS.GroupCount>0 then
        Result:=INIS.Group[0]
    end;
  end;
end;

procedure TINIEntry.SetCaption(const Caption:string);
begin
	if FLabel is TLabel then (FLabel as TLabel).Caption:=Caption;
end;

function TINIEntry.GetCaption:string;
begin
	Result:=DisplayLabel;
end;

function TINIEntry.CreateControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
    fTitle:=Self.Caption;
    fCaptionStyle:=Self.CaptionStyle;
		fCalculated:=Self.Calculated;
		fRO:=ReadOnly;
    fTitled:=True;
		fCtrlW:=Min(DisplayWidth*cWidth,fW);
		fCtrlH:=0;
    fNeighborhoodControl:=nil;
	end;

  try
		Result:=GetWinControl(AOwner,WParent,cInfo);

    except
    	on Exception do Result:=nil;
  end;

  if Result<>nil then begin
    if (WParent is TWinControl) and (Result.Parent=nil) then Result.Parent:=WParent as TWinControl;

    FLabel:=cInfo.fNeighborhoodControl;

    Result.Width:=Min(Result.Width,cInfo.fW);
    Result.Left:=Max(Result.Left,cInfo.fInitX+cInfo.fXBor);

    if Result.Height>0 then with cInfo do begin
    	fCtrlH:=Result.Height+(Result.Top-fY);

      if (FLabel<>nil) and ((FLabel.Height+(FLabel.Top-fY))>fCtrlH) then fCtrlH:=FLabel.Height+(FLabel.Top-fY);

    	fMaxCtrlH:=Max(fMaxCtrlH,fCtrlH);

      if fDirection=cdLeftToRight then begin
      	if not NextColumn then begin
          NextRow;
          FirstColumn;
        end;
      end else begin
  	    NextRow;
      end;
    end;

    Result.Hint:=strIf(Hint='',DisplayLabel,Hint);

    if Result is TWinControl then (Result as TWinControl).HelpContext:=HelpContext;
  end;

	Control:=Result;
end;

function TINIEntry.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	Result:=nil;

	INIError(Format(ResStr(CResPVFCalled),['GetWinControl']));
end;

procedure TINIEntry.Syncronize(Direction:TSyncronizeDirection);
begin
end;

procedure TINIEntry.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if (Operation=opRemove) and (AComponent=FControl) then FControl:=nil;
  if (Operation=opRemove) and (AComponent=FConnectedControl) then FConnectedControl:=nil;
end;

procedure TINIEntry.SetConnectedControl(WC:TWinControl);
begin
	if IsConnectionToControlSupported(WC) then FConnectedControl:=WC;
end;

function TINIEntry.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=(WC=nil);
end;

{
*******************************************************
	INIEntryList
*******************************************************
}

constructor TINIEntryLink.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FSubSection:=nil;
end;

destructor TINIEntryLink.Destroy;
begin
	if FSubSection<>nil then begin
  	FSubSection.FParentField:=nil;
    FSubSection.Free;
    FSubSection:=nil;
  end;

	inherited Destroy;
end;

procedure TINIEntryLink.SetSubSection(sSection:TINICustomSection);
begin
	if FSubSection<>sSection then begin
    if FSubSection<>nil then begin
      if FSubSection.FParentField<>nil then FSubSection.FParentField.Free;
    end;

    FSubSection:=sSection;

    if FSubSection<>nil then begin
      FSubSection.FParentField:=Self;
      FSubSection.Database:=nil;

      FPrivilegLevel:=FSubSection.FPrivilegLevel;
    end;
  end;
end;

function TINIEntryLink.IsDirty:Boolean;
begin
	Result:=True;
end;

procedure TINIEntryLink.DoWriteHumanReadable(Stream:TStream;Level:Integer);
begin
	if SubSection<>nil then with SubSection do DoWriteHumanReadable(Stream,Level);
end;

procedure TINIEntryLink.DoReadHumanReadable(Stream:TStream;Level:Integer);
begin
end;

procedure TINIEntryLink.DoReadEntry;
begin
	if SubSection<>nil then with SubSection do ReadSection;
end;

procedure TINIEntryLink.DoDeleteEntry;
begin
	if SubSection<>nil then with SubSection do DeleteSection;
end;

procedure TINIEntryLink.DoWriteEntry;
begin
	if SubSection<>nil then with SubSection do WriteSection;
end;

procedure TINIEntryLink.DoCopyEntry;
begin
	if SubSection<>nil then with SubSection do CopySection(Table.FCopyFromFileName);
end;

function TINIEntryLink.DoGetStoreIt:Boolean;
begin
	Result:=True;
end;

procedure TINIEntryLink.GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif});
begin
	if (FSubSection<>nil) and (FSubSection.Owner<>Self) then Proc(FSubSection);
end;

procedure TINIEntryLink.RefreshControl;
begin
	if SubSection<>nil then SubSection.RefreshControls;
end;

procedure TINIEntryLink.ShowControl(Show:Boolean);
begin
	if SubSection<>nil then begin
		if Show then
			SubSection.ShowControls
		else
 			SubSection.HideControls
	end;
end;

procedure TINIEntryLink.EnableControl(State:Boolean);
begin
	if SubSection<>nil then begin
		if State then
			SubSection.EnableControls
		else
 			SubSection.DisableControls;
	end;
end;

procedure TINIEntryLink.ClearControl;
begin
	if SubSection<>nil then SubSection.ClearControls;
end;

procedure TINIEntryLink.Reset;
begin
	if SubSection<>nil then SubSection.Reset;
end;

procedure TINIEntryLink.Init;
begin
	if SubSection<>nil then SubSection.Init;
end;

procedure TINIEntryLink.Syncronize(Direction:TSyncronizeDirection);
begin
	if SubSection<>nil then SubSection.Syncronize(Direction);
end;

{
*******************************************************
	INIEntrySaveList
*******************************************************
}

procedure TINIEntrySaveList.SetStaticList(sList:TStringList);
begin
	(FSubSection as TINICustomMultiSection).StaticSections.Assign(sList);
end;

function TINIEntrySaveList.GetStaticList:TStringList;
begin
	Result:=(FSubSection as TINICustomMultiSection).StaticSections;
end;

function TINIEntrySaveList.GetList:TStringList;
begin
	Result:=(FSubSection as TINICustomMultiSection).Sections;
end;

procedure TINIEntrySaveList.AddKey(const aKey:string);
begin
	with List do if IndexOf(aKey)=-1 then Add(aKey);
end;

function TINIEntrySaveList.GetKey:string;
begin
	Result:=(FSubSection as TINICustomMultiSection).Key;
end;

function TINIEntrySaveList.GetKeyIndex:Integer;
begin
	Result:=(FSubSection as TINICustomMultiSection).KeyIndex;
end;

procedure TINIEntrySaveList.SetKey(const selS:string);
begin
	(FSubSection as TINICustomMultiSection).Key:=selS;
end;

procedure TINIEntrySaveList.Reset;
begin
	(FSubSection as TINICustomMultiSection).AssignStatic;
end;

procedure TINIEntrySaveList.DoOnFieldNameChange;
begin
	inherited DoOnFieldNameChange;

  if FSubSection<>nil then FSubSection.SectionName:=FFieldName;
end;

procedure TINIEntrySaveList.DoOnConnected;
begin
	inherited DoOnConnected;

	Initialize;
end;

procedure TINIEntrySaveList.Loaded;
begin
	inherited Loaded;
end;

procedure TINIEntrySaveList.Initialize;
begin
  with SubSection as TINICustomMultiSection do begin
    SectionName:=FieldName;

		if Table<>nil then begin
      ParentDestination:=(Table as TINICustomSection).ParentDestination;
      FileName:=(Table as TINICustomSection).FileName;
      Destination:=(Table as TINICustomSection).Destination;
      PrivateMode:=(Table as TINICustomSection).PrivateMode;
    end;

    FParentField:=Self;

    CompressKey:=ciNo;
    Buttons:=[isSave,isDelete,isChange,isExchange];
    GroupType:=gtCascade;
  end;

  (FSubSection as TINICustomMultiSection).PanelMode:=FPanelMode;
end;

constructor TINIEntrySaveList.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FSubSection:=TINICustomMultiSection.Create(Self);

  FLineCount:=8;
  FPanelMode:=pmComboList;
end;

destructor TINIEntrySaveList.Destroy;
begin
	inherited Destroy;
end;

procedure TINIEntrySaveList.SetLineCount(LC:Integer);
begin
	FLineCount:=Max(0,LC);
end;

function TINIEntrySaveList.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
var pnlSubPanel:TCfgSubPanel;
begin
	Result:=nil;

	if SubSection<>nil then begin
    with cInfo do begin
      if PanelMode=pmMemo then begin
        if LineCount<=0 then
          fCtrlH:=fH-(fY-fInitY-fYBor)
        else
          fCtrlH:=fcHeight*LineCount;

        if fCtrlH<1 then fCtrlH:=fH;
      end else begin
        fCtrlW:=Min(fCtrlW+160,fW);
        fCtrlH:=fcHeight+12;
      end;
    end;

    pnlSubPanel:=INISec.CreateTitledControl(AOwner,TCfgSubPanel,WParent,cInfo) as TCfgSubPanel;

		pnlSubPanel.SendToBack;

    Result:=pnlSubPanel;
  end;
end;

procedure TINIEntrySaveList.Syncronize(Direction:TSyncronizeDirection);
var FSS:TINICustomMultiSection;
begin
  if FConnectedControl<>nil then begin
  	FSS:=FSubSection as TINICustomMultiSection;

    if FConnectedControl is TListBox then begin
      if Direction=sdToControl then
        (FConnectedControl as TListBox).Items.Assign(FSS.Sections)
      else
        FSS.AssignSections((FConnectedControl as TListBox).Items);
    end else if FConnectedControl is TComboBox then begin
      if Direction=sdToControl then begin
        (FConnectedControl as TComboBox).Items.Assign(FSS.Sections);
        (FConnectedControl as TComboBox).Text:=Key;
      end else begin
        FSS.AssignSections((FConnectedControl as TComboBox).Items);
        Key:=(FConnectedControl as TComboBox).Text;
      end;
    end;
  end;
end;

function TINIEntrySaveList.IsConnectionToControlSupported(WC:TWinControl):Boolean;
begin
	Result:=inherited IsConnectionToControlSupported(WC) or (WC is TListBox) or (WC is TComboBox);
end;

{
*******************************************************
	INIEntrySubSection
********************************************************
}

destructor TINIEntrySubSection.Destroy;
begin
	inherited Destroy;
end;

procedure TINIEntrySubSection.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if (Operation=opRemove) and (AComponent=FSubSection) then FSubSection:=nil;
end;

procedure TINIEntrySubSection.GetTableChildren(Proc:TGetChildProc);
begin
	if (SubSection<>nil) and (SubSection.Owner<>Self) then Proc(SubSection);
end;

function TINIEntrySubSection.GetWinControl(AOwner:TComponent;WParent:TControl;cInfo:TCreateInfo):TControl;
var pnlSubPanel:TCfgSubPanel;
begin
	Result:=nil;

	if SubSection<>nil then begin
    with cInfo do begin
      fCtrlW:=fW;
      fCtrlH:=fH-(fY-fInitY-fYBor);

      if SubSection.Height>0 then fCtrlH:=SubSection.Height;
      
      if fCtrlH<1 then fCtrlH:=fH;
    end;

	  pnlSubPanel:=CreateUntitledControl(AOwner,TCfgSubPanel,WParent,cInfo) as TCfgSubPanel;

    pnlSubPanel.SendToBack;

    Result:=pnlSubPanel;
  end;
end;

{
*******************************************************
	INIEntryStreamed
********************************************************
}

constructor TINIEntryStreamed.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FShowInGrid:=False;

	FOnAfterRead:=nil;
	FOnBeforeWrite:=nil;
	FOnStoreIt:=nil;
end;

function TINIEntryStreamed.DoGetStoreIt:Boolean;
begin
	Result:=True;
end;

procedure TINIEntryStreamed.Assign(Source:TPersistent);
var INIE:TINIEntryStreamed;
begin
	inherited Assign(Source);

	if Source is TINIEntryStreamed then begin
  	INIE:=Source as TINIEntryStreamed;

    ConnectedControl:=INIE.ConnectedControl;
    ShowInGrid:=INIE.ShowInGrid;
  end;
end;

procedure TINIEntryStreamed.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);
end;

procedure TINIEntryStreamed.DoDeleteEntry;
begin
	with TINICustomSection(Table) do INIStream.DeleteEntry(RuntimeSectionName,FieldName);
end;

procedure TINIEntryStreamed.DoReadEntry;
begin
	with TINICustomSection(Table) do AsString:=INIStream.ReadString(RuntimeSectionName,FieldName,AsString);
end;

procedure TINIEntryStreamed.DoWriteEntry;
begin
	with TINICustomSection(Table) do INIStream.WriteString(RuntimeSectionName,FieldName,AsString);
end;

procedure TINIEntryStreamed.DoCopyEntry;
begin
	DoWriteEntry;
end;

{ functions }

function CreateUntitledControl(AOwner:TComponent;TCC:TControlClass;WParent:TControl;cInfo:TCreateInfo):TControl;
begin
	with cInfo do begin
		Result:=TCC.Create(AOwner);

		with Result do begin
			Parent:=WParent as TWinControl;

			if fCtrlW<=0 then fCtrlW:=Width;
			if fCtrlH<=0 then fCtrlH:=Height;

			SetBounds(fX,fY,fCtrlW,fCtrlH);
		end;
	end;
end;

function CreateTitledControl(AOwner:TComponent;TCC:TControlClass;WParent:TControl;cInfo:TCreateInfo):TControl;
var newL:TLabel;
		lH,ldY:Integer;
    cStyle:TCaptionStyle;
begin
	newL:=nil;

	with cInfo do begin
  	lH:=0;
    ldY:=0;
    cStyle:=fCaptionStyle;

    if (fCtrlW>=fW) and (cStyle<=csLeftBottom) then cStyle:=csTopLeft;

    if cStyle<>csWithout then begin
    	newL:=TLabel.Create(AOwner);

      if (cStyle in [csTopLeft,csTopCenter,csTopRight]) and fTitled then lH:=newL.Height+1;
    end;

		Result:=TCC.Create(AOwner);

		if fCtrlH>0 then Result.Height:=fCtrlH;

		if fY+Result.Height+fControlDistance+lH>fInitY+fYBor+fH then begin
    	if (fDirection=cdTopToBottom) and NextColumn then FirstRow;
    end;

    if newL<>nil then begin
      with newL do begin
        Parent:=WParent as TWinControl;
        Left:=fX;
        Top:=fY;
        Caption:=fTitle;

        if fCalculated then Font.Color:=clGrayText;

        BringToFront;
      end;

      if (cStyle in [csTopLeft,csTopCenter,csTopRight]) and fTitled then ldY:=newL.Height+1;
    end;

		with Result do begin
			Parent:=WParent as TWinControl;
			ShowHint:=True;

      if (cStyle<=csLeftBottom) and (fCtrlW>0) and fTitled then begin
      	if fCtrlW<fW then fCtrlW:=Min(fW-newL.Width-fXBor,fCtrlW);

      	Left:=fX+fW-fCtrlW;
        Width:=fCtrlW;
      end else begin
      	Left:=fX;

        if fCtrlW>0 then
        	Width:=fCtrlW
        else
        	Width:=fW;
      end;

			Top:=fY+ldY;

			Enabled:=not (fCalculated or fRO);
		end;

		Result.SetBounds(Result.Left,Result.Top,Result.Width,Result.Height);

    if newL<>nil then begin
	    if cStyle<=csLeftBottom then begin
      	if Result.Height>newL.Height then begin
          case cStyle of
            csLeftTop: newL.Top:=Result.Top;
            csLeftCenter: newL.Top:=Result.Top+(Result.Height-newL.Height) div 2;
            csLeftBottom: newL.Top:=Result.Top+Result.Height-newL.Height;
          end;
        end else begin
          case cStyle of
            csLeftBottom: Result.Top:=newL.Top;
            csLeftCenter: Result.Top:=newL.Top+(newL.Height-Result.Height) div 2;
            csLeftTop: Result.Top:=newL.Top+newL.Height-Result.Height;
          end;
        end;
      end else begin
      	if Result.Width>newL.Width then begin
          case cStyle of
            csTopLeft: newL.Left:=Result.Left;
            csTopCenter: newL.Left:=Result.Left+(Result.Width-newL.Width) div 2;
            csTopRight: newL.Left:=Result.Left+Result.Width-newL.Width;
          end;
        end else begin
          case cStyle of
            csTopLeft: Result.Left:=newL.Left;
            csTopCenter: Result.Left:=newL.Left+(newL.Width-Result.Width) div 2;
            csTopRight: Result.Left:=newL.Left+newL.Width-Result.Width;
          end;
        end;
      end;
	  end;

    fNeighborhoodControl:=newL;
	end;
end;

{ TINIDatabase }

constructor TGlyphDescriptor.Create;
begin
	inherited Create;

	FGlyph:=TBitmap.Create;
  FNumGlyphs:=1;
end;

destructor TGlyphDescriptor.Destroy;
begin
  FGlyph.Free;

	inherited Destroy;
end;

procedure TGlyphDescriptor.SetGlyph(bGlyph:TBitmap);
begin
	if bGlyph=nil then begin
{$ifndef VER90}
		FGlyph.Handle:=0;
{$endif}
{$ifdef VER90}
	with FGlyph do begin
  	Width:=0;
    Height:=0;
  end;
{$endif}
  end else begin
  	FGlyph.Assign(bGlyph);

    if (FGlyph.Width div 2)>=FGlyph.Height then FNumGlyphs:=2;
  end;
end;

procedure TGlyphDescriptor.SetNumGlyphs(nGlyphs:Integer);
begin
	if (nGlyphs>=1) and (nGlyphs<=4) then FNumGlyphs:=nGlyphs;
end;

procedure TGlyphDescriptor.GAssignTo(BB:TBitBtn;bVisible:Boolean);
begin
	BB.Glyph.Assign(Glyph);
	BB.NumGlyphs:=NumGlyphs;
  BB.Visible:=bVisible;
  BB.Caption:=Caption;
end;

procedure TGlyphDescriptor.GAssign(GD:TGlyphDescriptor);
begin
	FGlyph.Assign(GD.Glyph);
	FNumGlyphs:=GD.NumGlyphs;
  FCaption:=GD.Caption;
end;

constructor TGlyphs.Create;
begin
	inherited Create;

  FGlyphDescriptors[0]:=TGlyphDescriptor.Create;
  FGlyphDescriptors[1]:=TGlyphDescriptor.Create;
  FGlyphDescriptors[2]:=TGlyphDescriptor.Create;
end;

destructor TGlyphs.Destroy;
begin
  FGlyphDescriptors[0].Free;
  FGlyphDescriptors[1].Free;
  FGlyphDescriptors[2].Free;

	inherited Destroy;
end;

procedure TGlyphs.SetGlyphDescriptor(Index:Integer;GD:TGlyphDescriptor);
begin
	FGlyphDescriptors[Index].GAssign(GD);
end;

function TGlyphs.GetGlyphDescriptor(Index:Integer):TGlyphDescriptor;
begin
	Result:=FGlyphDescriptors[Index];
end;

constructor TNames.Create;
begin
	inherited Create;

  FUserName:='DefaultUser';
  FGroupName:='DefaultGroup';
  FComputerName:='';

  FOnUserNameChange:=nil;
  FOnGroupNameChange:=nil;

  FUseWindowsUserInstead:=False;
end;

function TNames.IsUserNameStored:Boolean;
begin
	Result:=FUserName<>'DefaultUser';
end;

function TNames.IsGroupNameStored:Boolean;
begin
	Result:=FGroupName<>'DefaultGroup';
end;

procedure TNames.AssignWindowsUser;
var lWinUser:array[0..255] of Char;
		lSize:Integer;
begin
	lSize:=254;
	GetUserName(lWinUser,lSize);

  if lWinUser[0]<>#0 then User:=lWinUser;
end;

procedure TNames.SetGroupName(const sGroup:string);
begin
	FGroupName:=sGroup;

  if Assigned(FOnGroupNameChange) then FOnGroupNameChange(Self);
end;

procedure TNames.SetUserName(const sUser:string);
begin
	FUserName:=sUser;

  if Assigned(FOnUserNameChange) then FOnUserNameChange(Self);
end;

procedure TNames.SetWindowsUserInstead(UWUInstead:Boolean);
begin
  FUseWindowsUserInstead:=UWUInstead;

  if FUseWindowsUserInstead then AssignWindowsUser;
end;

constructor TDirs.Create;
begin
	inherited Create;

  FPublic:='';
  FUser:='';
  FGroup:='';

  FUseRunDirInstead:=True;
  FUseWindowsDirInstead:=False;

  AssignRunDir;
end;

procedure TDirs.AssignRunDir;
begin
  Public:=ExtractFilePath(Forms.Application.ExeName);
end;

procedure TDirs.AssignWindowsDir;
var lWinDir:array[0..255] of Char;
begin
	GetWindowsDirectory(lWinDir,254);

  Public:=StrPas(lWinDir);
end;

function TDirs.GetPublic:string;
begin
  Result:=FPublic;
end;

function TDirs.IsGroupStored:Boolean;
begin
	Result:=True;
end;

function TDirs.IsUserStored:Boolean;
begin
	Result:=True;
end;

procedure TDirs.SetUseRunDirInstead(URDInstead:Boolean);
begin
  FUseRunDirInstead:=URDInstead;

  if FUseRunDirInstead then begin
  	AssignRunDir;

    FUseWindowsDirInstead:=False;
  end;
end;

procedure TDirs.SetUseWindowsDirInstead(UWDInstead:Boolean);
begin
  FUseWindowsDirInstead:=UWDInstead;

  if FUseWindowsDirInstead then begin
  	AssignWindowsDir;

    FUseRunDirInstead:=False;
  end;
end;

function TDirs.GetData:string;
begin
 	Result:=FGroup;
end;

function TDirs.GetUser:string;
begin
 	Result:=FUser;
end;

procedure TDirs.SetPublic(const sPublic:string);
begin
	if FGroup=FPublic then Group:=sPublic;
  
 	FPublic:=sPublic;
end;

procedure TDirs.SetGroup(const sGroup:string);
begin
	if FUser=FGroup then User:=sGroup;
  
 	FGroup:=sGroup;
end;

procedure TDirs.SetUser(const sUser:string);
begin
 	FUser:=sUser;
end;

constructor TINIDatabase.Create(AOwner:TComponent);
begin
  FDestination:=CDefaultFileStream;

	inherited Create(AOwner);

	FConstCreateInfo:=TCreateInfo.Create(nil);

	FMaster:=nil;
  FTabControl:=nil;
  FImageList:=nil;
  FStartPage:='';

  FDistance:=CCDistance;
  FWhatPM:=[pmPublic,pmUserPrivate,pmGroupPrivate];
  FParent:=nil;

  FNames:=TNames.Create;
  FDirs:=TDirs.Create;
  FGlyphs:=TGlyphs.Create;

  FNames.OnGroupNameChange:=GroupNameChanged;
  FNames.OnUserNameChange:=UserNameChanged;

  FOnInit:=nil;
  FOnAfterRead:=nil;
  FOnBeforeWrite:=nil;

  FUserPrivilegLevel:=0;
  FFileName:='';

  FEditOnlyMode:=False;
  FCreatePageOnChange:=True;

  FOnFieldChange:=nil;
  
  FormGetSection:=nil;

  FSections:=TList.Create;

  if (ComponentState=[]) and (AOwner is TForm) then InitCreateInfo((AOwner as TForm).Canvas);
end;

destructor TINIDatabase.Destroy;
begin
  FNames.Free;
  FGlyphs.Free;

  DestroyTables;

  FSections.Free;
  FConstCreateInfo.Free;

  if FormGetSection<>nil then FormGetSection.Free;

	inherited Destroy;
end;

procedure TINIDatabase.Loaded;
begin
	inherited Loaded;

  if Dirs.UseRunDirInstead then
  	Dirs.AssignRunDir
  else if Dirs.UseWindowsDirInstead then
  	Dirs.AssignWindowsDir;

  if Names.UseWindowsUserInstead then Names.AssignWindowsUser;
end;

procedure TINIDatabase.UserNameChanged(Sender:TObject);
begin
	if not FEditOnlyMode then begin
  	ForEachSection(FEPRead,nil,[pmUserPrivate],False);

    if Assigned(FOnAfterRead) then FOnAfterRead(Self);
  end;
end;

procedure TINIDatabase.GroupNameChanged(Sender:TObject);
begin
	if not FEditOnlyMode then begin
  	ForEachSection(FEPRead,nil,[pmGroupPrivate],False);

    if Assigned(FOnAfterRead) then FOnAfterRead(Self);
  end;
end;

procedure TINIDatabase.SetDestination(DT:TDestination);
begin
	if IsINIStreamRegistered(DT) then
  	FDestination:=DT
  else
  	raise Exception.Create(Format(ResStr(CResStreamNotRegistered),[DT]));
end;

procedure TINIDatabase.AnyFieldChange(Field:TINIField);
begin
	if Assigned(FOnFieldChange) then FOnFieldChange(Field);
end;

function TINIDatabase.IsDestinationStored:Boolean;
begin
	Result:=AnsiCompareText(FDestination,CDefaultFileStream)<>0;
end;

procedure TINIDatabase.GoSearchSection(MultiSection:TINICustomMultiSection);
begin
	if FormSearchMultiSectionClass<>nil then begin
		if FormGetSection=nil then FormGetSection:=FormSearchMultiSectionClass.Create(Application);

  	FormGetSection.FreeNotification(Self);

	 	FormGetSection.Go(MultiSection,MultiSection.Panel);
  end;
end;

procedure TINIDatabase.DestroyTables;
var Section:TINICustomSection;

	function DestroyTable:Boolean;
  var i:Integer;
  		INIT:TINITable;
  begin
  	for i:=0 to Master.ComponentCount-1 do begin
    	if Master.Components[i] is TINITable then begin
      	INIT:=Master.Components[i] as TINITable;

      	if (INIT.Database=Self) and not INIT.Son then begin
          INIT.Free;

          Result:=True;

          Exit;
        end;
      end;
    end;

  	Result:=False;
  end;

begin
	if Master<>nil then begin
  	repeat until not DestroyTable;
  end;

  while FSections.Count>0 do begin
    Section:=FSections.Last;
    RemoveSection(Section);
    Section.Free;
  end;
end;

procedure TINIDatabase.SetUserPrivilegLevel(UserLevel:Integer);
begin
	if (UserLevel>=0) or (UserLevel<=CMaxPrivilegLevel) then FUserPrivilegLevel:=UserLevel;
end;

procedure TINIDatabase.SetDistance(CD:Integer);
begin
	if (CD>0) and (CD<128) then begin
		FDistance:=CD;

  	FConstCreateInfo.fControlDistance:=CD;
  end;
end;

procedure TINIDatabase.InitCreateInfo(BorrowCanvas:TCanvas);
begin
  with FConstCreateInfo do begin
  	fCanvas:=BorrowCanvas;

    fcWidth:=(TextWidth('Wg') div 2)+4;
    fcHeight:=TextHeight('Wg')+4;

    fXBor:=3;
    fYBor:=1;
    fInitX:=2;
    fInitY:=3;

    fControlDistance:=FDistance;
  end;
end;

procedure TINIDatabase.SetParent(WParent:TWinControl);
begin
	FParent:=WParent;

  if WParent<>nil then begin
  	WParent.FreeNotification(Self);

    if WParent is TForm then
      InitCreateInfo((WParent as TForm).Canvas)
    else if WParent.Parent is TForm then
      InitCreateInfo((WParent.Parent as TForm).Canvas)
    else if WParent.Owner is TForm then
      InitCreateInfo((WParent.Owner as TForm).Canvas);
  end;
end;

procedure TINIDatabase.SetNames(ns:TNames);
begin
	FNames.Assign(ns);
end;

procedure TINIDatabase.SetDirs(ds:TDirs);
begin
	FDirs.Assign(ds);
end;

procedure TINIDatabase.SetGlyphs(gs:TGlyphs);
begin
	FGlyphs.Assign(gs);
end;

procedure TINIDatabase.Read;
begin
	if not FEditOnlyMode then begin
  	ForEachSection(FEPRead,nil,PrivateModes,True);

    if Assigned(FOnAfterRead) then FOnAfterRead(Self);
  end;
end;

procedure TINIDatabase.Init;
begin
 	ForEachSection(FEPInit,nil,PrivateModes,True);

  if Assigned(FOnInit) then FOnInit(Self);
end;

procedure TINIDatabase.Reset;
begin
 	ForEachSection(FEPReset,nil,PrivateModes,True);
end;

procedure TINIDatabase.Write;
begin
	if not FEditOnlyMode then begin
    if Assigned(FOnBeforeWrite) then FOnBeforeWrite(Self);

  	ForEachSection(FEPWrite,nil,PrivateModes,True);
  end;
end;

procedure TINIDatabase.Copy(const CopyFrom:string);
begin
	if not FEditOnlyMode then ForEachSection(FEPCopy,@CopyFrom,PrivateModes,True);
end;

procedure TINIDatabase.Clear;
begin
	ForEachSection(FEPClear,nil,PrivateModes,True);
end;

procedure TINIDatabase.Refresh;
begin
	ForEachSection(FEPRefresh,nil,PrivateModes,True);
end;

procedure TINIDatabase.DoWriteHumanReadable(Stream:TStream;Level:Integer);
var SI:TStreamInfo;
begin
	SI.Stream:=Stream;
  SI.Level:=Level;

	WriteLnTextStream(Stream,Format('Database "%s"={',[Self.Name]));

	ForEachSection(FEPWriteHumanReadable,@SI,[pmPublic,pmGroupPrivate,pmUserPrivate],False);

	WriteTextStream(Stream,'}');
end;

procedure TINIDatabase.DoReadHumanReadable(Stream:TStream;Level:Integer);
begin
end;

procedure TINIDatabase.WriteHumanReadable(const FileName:string);
var Stream:TFileStream;
begin
	Stream:=TFileStream.Create(FileName,fmCreate or fmOpenWrite or fmShareExclusive);

  try
		DoWriteHumanReadable(Stream,1);

    finally
    	Stream.Free;
  end;
end;

procedure TINIDatabase.ReadHumanReadable(const FileName:string);
var Stream:TFileStream;
begin
	Stream:=TFileStream.Create(FileName,fmOpenRead or fmShareDenyWrite);

  try
		DoReadHumanReadable(Stream,0);

    finally
    	Stream.Free;
  end;
end;

procedure TINIDatabase.ForEachSection(FEP:TForEachProcedure;Data:Pointer;PMSet:TPrivateModes;Extend:Boolean);
var i,mc:Integer;
		INIS:TINICustomSection;
begin
	if Extend then begin
		if pmUserPrivate in PMSet then PMSet:=PMSet+[pmGroupPrivate];
		if pmGroupPrivate in PMSet then PMSet:=PMSet+[pmPublic];
  end;

	mc:=Master.ComponentCount-1;

	for i:=0 to mc do begin
		if Master.Components[i] is TINICustomSection then begin
			INIS:=Master.Components[i] as TINICustomSection;

      if (INIS.Database=Self) and (not INIS.Son) and (INIS.RuntimePrivateMode in PMSet) then FEP(INIS,Data);
		end;
	end;
end;

procedure TINIDatabase.FEPWriteHumanReadable(Item:TObject;Data:Pointer);
var pSI:^TStreamInfo;
begin
	pSI:=Data;

	TINICustomSection(Item).DoWriteHumanReadable(pSI^.Stream,pSI^.Level);
end;

procedure TINIDatabase.FEPReadHumanReadable(Item:TObject;Data:Pointer);
begin
end;

procedure TINIDatabase.FEPClear(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).ClearControls;
end;

procedure TINIDatabase.FEPInit(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).Init;
end;

procedure TINIDatabase.FEPReset(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).Reset;
end;

procedure TINIDatabase.FEPRead(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).ReadSection;
end;

procedure TINIDatabase.FEPWrite(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).WriteSection;
end;

procedure TINIDatabase.FEPRefresh(Item:TObject;Data:Pointer);
begin
	TINICustomSection(Item).RefreshControls;
end;

procedure TINIDatabase.FEPCopy(Item:TObject;Data:Pointer);
var PCopyFrom:^string;
begin
	PCopyFrom:=Data;

	TINICustomSection(Item).CopySection(PCopyFrom^);
end;

procedure TINIDatabase.OpenSectionForEdit(Section:TINICustomSection);
var i:Integer;
begin
	if FTabControl<>nil then begin
  	for i:=0 to FTabControl.Tabs.Count-1 do begin
    	if AnsiCompareText(FTabControl.Tabs[i],Section.SectionName)=0 then begin
      	FTabControl.TabIndex:=i;

        OnTabChange(FTabControl);

        Exit;
      end;
    end;
  end;
end;

function TINIDatabase.CreatePage(AOwner:TComponent;Section:TINICustomSection;WParent:TWinControl):Integer;
begin
	Section.InternalCreatePage(AOwner,WParent,Result);
end;

procedure TINIDatabase.AddSection(Section:TINICustomSection);
begin
  FSections.Add(Section);
  Section.FDatabase:=Self;
end;

procedure TINIDatabase.RemoveSection(Section:TINICustomSection);
begin
  Section.FDatabase:=nil;
  FSections.Remove(Section);
end;

function TINIDatabase.GetSectionOfPrivateModeCount:Integer;
var i:Integer;
begin
	Result:=0;

  for i:=0 to FSections.Count-1 do if Sections[i].PrivateMode in PrivateModes then Inc(Result);
end;

function TINIDatabase.GetSectionCount:Integer;
var i:Integer;
begin
	if PrivateModes=[pmPublic,pmGroupPrivate,pmUserPrivate] then
	  Result:=FSections.Count
  else begin
  	Result:=0;
    
  	for i:=0 to FSections.Count-1 do
    	if TINICustomSection(FSections[i]).PrivateMode in PrivateModes then Result:=Result+1;
  end;
end;

function TINIDatabase.GetSection(Index:Integer):TINICustomSection;
var i:Integer;
begin
	if PrivateModes=[pmPublic,pmGroupPrivate,pmUserPrivate] then
	  Result:=FSections[Index]
  else begin
  	Result:=nil;
    
  	for i:=0 to FSections.Count-1 do begin
    	if TINICustomSection(FSections[i]).PrivateMode in PrivateModes then Index:=Index-1;

      if Index<0 then begin
      	Result:=FSections[i];

        Break;
      end;
    end;
  end;

  if Result=nil then raise EListError.Create(ResStr(CResSectionIndexOutOfRange));
end;

procedure TINIDatabase.SetSection(Index:Integer; Value:TINICustomSection);
begin
  TINICustomSection(FSections[Index]).Assign(Value);
end;

function TINIDatabase.SectionByName(const SectionName:string):TINICustomSection;
begin
  Result:=FindSection(SectionName);

  if Result=nil then INIErrorFmt(ResStr(CResSectionNotFound), [SectionName]);
end;

function TINIDatabase.SectionByNumber(SectionNo:Integer):TINICustomSection;
var i:Integer;
begin
  for i:=0 to FSections.Count - 1 do begin
    Result:=FSections[i];

    if Result.TableNo=SectionNo then Exit;
  end;
  
  Result:=nil;
end;

function TINIDatabase.FindSection(const SectionName:string):TINICustomSection;
var i:Integer;
begin
  for i:=0 to FSections.Count-1 do begin
    Result:=FSections[i];

    if AnsiCompareText(Result.FSectionName,SectionName) = 0 then Exit;
  end;
  
  Result:=nil;
end;

procedure TINIDatabase.GetSectionNames(List:TStrings);
var i:Integer;
begin
  List.BeginUpdate;

  try
    List.Clear;

    for i:=0 to FSections.Count-1 do List.Add(TINICustomSection(FSections[i]).FSectionName);

	  finally
  	  List.EndUpdate;
  end;
end;

procedure TINIDatabase.OnTabChange(Sender:TObject);
var tabIndex:Integer;
		tabPanel:TWinControl;
    i:Integer;

	function FindControlWithTag(WC:TWinControl;Tag:Integer):TWinControl;
  var i:Integer;
  begin
  	for i:=0 to WC.ControlCount-1 do begin
    	if (WC.Controls[i] is TWinControl) and ((WC.Controls[i] as TWinControl).Tag=Tag) then begin
      	Result:=WC.Controls[i] as TWinControl;

        Exit;
      end;
    end;

    Result:=nil;
  end;

begin
	tabIndex:=(Sender as TTabControl).TabIndex;

  if not FCreatePageOnChange then begin
	  tabPanel:=FindControlWithTag(Sender as TWinControl,tabIndex);

    if tabPanel<>nil then tabPanel.BringToFront;
  end else begin
	  tabPanel:=(Sender as TTabControl).Controls[0] as TWinControl;

    for i:=tabPanel.ControlCount-1 downto 0 do tabPanel.Controls[i].Free;

    if tabIndex<SectionCount then begin
      if csDesigning in ComponentState then
        tabPanel.Parent.Parent.Visible:=False
      else
        tabPanel.Visible:=False;

      try
        CreatePage(Self,Sections[tabIndex],tabPanel);

        Sections[tabIndex].RefreshControls;

        finally
          if csDesigning in ComponentState then
            tabPanel.Parent.Parent.Visible:=True
          else
            tabPanel.Visible:=True;
      end;
    end;
  end;
end;

function TINIDatabase.CreatePages:Integer;
var tabPanel:TPanel;
		i:Integer;
    sName:string;
    TCI:TTCItem;
    INIS:TINICustomSection;
begin
	Result:=0;

	if SectionCount>1 then begin
    FTabControl:=TTabControl.Create(Self);

    with FTabControl do begin
      Parent:=FParent;
      Align:=alClient;
      MultiLine:=False;
      TabIndex:=0;
      OnChange:=OnTabChange;

      if FImageList<>nil then SendMessage(FTabControl.Handle,TCM_SETIMAGELIST,0,FImageList.Handle);

      for i:=0 to SectionCount-1 do begin
      	Tabs.Add(Sections[i].SectionName);

        if (Sections[i] is TINICustomSection) and (FImageList<>nil) then begin
        	INIS:=Sections[i] as TINICustomSection;

	        if (INIS.FImageIndex>=0) and (INIS.FImageIndex<FImageList.Count) then begin
          	sName:=INIS.SectionName;

		        TCI.mask:=TCIF_TEXT or TCIF_IMAGE;
  		      TCI.pszText:=@(sName[1]);
    		    TCI.cchTextMax:=Length(sName);
						TCI.iImage:=INIS.FImageIndex;

	          SendMessage(FTabControl.Handle,TCM_SETITEM,i,LPARAM(@TCI));
  	      end;
        end;

        if (i=0) or not FCreatePageOnChange then begin
          tabPanel:=TPanel.Create(FTabControl);

          with tabPanel do begin
            Parent:=FTabControl;
            Align:=alClient;
            BevelInner:=bvNone;
            BevelOuter:=bvNone;
            Tag:=i;
          end;
          
          if not FCreatePageOnChange then begin
            CreatePage(Self,Sections[i],tabPanel);

            Sections[i].RefreshControls;
          end;
        end;
      end;

      OnTabChange(FTabControl);
    end;

    Result:=1;
  end else if SectionCount>0 then begin
	  Result:=CreatePage(Self,Sections[0],FParent);

    Sections[0].RefreshControls;
  end;
end;

procedure TINIDatabase.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if Operation=opRemove then begin
    if (FMaster=AComponent) and (FMaster<>Owner) then FMaster:=nil;

  	if FTabControl=AComponent then FTabControl:=nil;
  	if FParent=AComponent then FParent:=nil;
    if FImageList=AComponent then FImageList:=nil;
    if FormGetSection=AComponent then FormGetSection:=nil;
  end;
end;

procedure TINIDatabase.SetMaster(frm:TComponent);
begin
	FMaster:=frm;

  if frm<>nil then frm.FreeNotification(Self);
end;

function TINIDatabase.GetMaster:TComponent;
begin
	Result:=FMaster;

  if Result=nil then Result:=Owner;
end;

procedure TINIDatabase.GetChildren(Proc:TGetChildProc{$ifndef VER90};Root:TComponent{$endif});
var i:Integer;
    INIT:TINITable;
begin
	if Master<>nil then begin
  	for i:=0 to Master.ComponentCount-1 do begin
    	if Master.Components[i] is TINITable then begin
      	INIT:=Master.Components[i] as TINITable;

      	if (INIT.Database=Self) and (INIT.Owner<>Self) and not INIT.Son then Proc(INIT);
      end;
    end;
  end;
end;

function TDestinationProperty.GetAttributes:TPropertyAttributes;
begin
	Result:=[paMultiSelect,paValueList];
end;

procedure TDestinationProperty.GetValues(Proc:TGetStrProc);
var i:Integer;
begin
	for i:=0 to StreamClasses.Count-1 do Proc((StreamClasses.Objects[i] as TStreamObjectClass).Description);
end;

{ Register functions }

function GetStreamClassIndexByDescription(const StreamDescription:string):Integer;
begin
  for Result:=0 to StreamClasses.Count-1 do begin
    if AnsiCompareText((StreamClasses.Objects[Result] as TStreamObjectClass).Description,StreamDescription)=0 then Exit;
  end;

  Result:=0;
end;

function GetStreamClassByDescription(const StreamDescription:string):TClassOfINIStream;
begin
	try
		Result:=(StreamClasses.Objects[GetStreamClassIndexByDescription(StreamDescription)] as TStreamObjectClass).ObjectClass;

		except
    	on Exception do raise Exception.Create(Format(ResStr(CResStreamNotRegistered),[StreamDescription]));
  end;
end;

function CreateStream(const ComputerName,Name,StreamDescription:string):TINIStream;
begin
  Result:=GetStreamClassByDescription(StreamDescription).CreateStream(ComputerName,Name);
end;

procedure RegisterINIStream(const StreamDescription:string;StreamClass:TClassOfINIStream);
begin
  if StreamClass.ClassName=TINIStream.ClassName then raise Exception.Create(Format(ResStr(CResCannotRegisterStream),[TINIStream.ClassName]));

	if StreamClasses=nil then StreamClasses:=TStringList.Create;

  if StreamClasses.IndexOf(StreamClass.ClassName)=-1 then StreamClasses.AddObject(StreamClass.ClassName,TStreamObjectClass.Create(StreamClass,StreamDescription));
end;

procedure RegisterINIEntry(const EntryDescription:string;ImageIndex:TEntryImageIndex;EntryClass:TClassOfINIEntry);
begin
  if EntryClass.ClassName=TINIEntryLink.ClassName then raise Exception.Create(Format(ResStr(CResCannotRegisterEntry),[TINIEntryLink.ClassName]));
  if EntryClass.ClassName=TINIEntrySubSection.ClassName then raise Exception.Create(Format(ResStr(CResCannotRegisterEntry),[TINIEntrySubSection.ClassName]));

	if EntryClasses=nil then EntryClasses:=TStringList.Create;

  if EntryClasses.IndexOf(EntryClass.ClassName)=-1 then EntryClasses.AddObject(EntryClass.ClassName,TEntryObjectClass.Create(EntryClass,EntryDescription,Integer(ImageIndex)));

	RegisterClass(EntryClass);
end;

procedure UnRegisterINIEntry(EntryClass:TClassOfINIEntry);
var ii:Integer;
begin
	ii:=EntryClasses.IndexOf(EntryClass.ClassName);

  if ii<>-1 then begin
  	EntryClasses.Delete(ii);

	  UnRegisterClass(EntryClass);
  end;
end;

constructor TEntryObjectClass.Create(ObjectClass:TClassOfINIEntry;const ClassDescription:string;Index:Integer);
begin
	FClass:=ObjectClass;
  FDescription:=ClassDescription;
  FImageIndex:=max(0,min(Index,CMaxEntryImageIndex));
end;

procedure RegisterINISection(const SectionDescription:string;SectionClass:TClassOfINISection);
begin
	if SectionClasses=nil then SectionClasses:=TStringList.Create;

  if SectionClasses.IndexOf(SectionClass.ClassName)=-1 then SectionClasses.AddObject(SectionClass.ClassName,TSectionObjectClass.Create(SectionClass,SectionDescription));

	RegisterClass(SectionClass);
end;

procedure UnRegisterINISection(SectionClass:TClassOfINISection);
var ii:Integer;
begin
	ii:=SectionClasses.IndexOf(SectionClass.ClassName);

  if ii<>-1 then begin
  	SectionClasses.Delete(ii);

	  UnRegisterClass(SectionClass);
  end;
end;

constructor TSectionObjectClass.Create(ObjectClass:TClassOfINISection;const ClassDescription:string);
begin
	FClass:=ObjectClass;
  FDescription:=ClassDescription;
end;

constructor TStreamObjectClass.Create(ObjectClass:TClassOfINIStream;const ClassDescription:string);
begin
	FClass:=ObjectClass;
  FDescription:=ClassDescription;
end;

function GetEntryClassesCount:Integer;
begin
	Result:=EntryClasses.Count;
end;

function GetSectionClassesCount:Integer;
begin
	Result:=SectionClasses.Count;
end;

function GetStreamClassesCount:Integer;
begin
	Result:=StreamClasses.Count;
end;

function GetEntryClass(Index:Integer):TClassOfINIEntry;
begin
	Result:=(EntryClasses.Objects[Index] as TEntryObjectClass).ObjectClass;
end;

function GetSectionClass(Index:Integer):TClassOfINISection;
begin
	Result:=(SectionClasses.Objects[Index] as TSectionObjectClass).ObjectClass;
end;

function GetEntryClassDescription(Index:Integer):string;
begin
	Result:=(EntryClasses.Objects[Index] as TEntryObjectClass).Description;
end;

function GetSectionClassDescription(Index:Integer):string;
begin
	Result:=(SectionClasses.Objects[Index] as TSectionObjectClass).Description;
end;

function GetStreamClassDescription(Index:Integer):string;
begin
	Result:=(StreamClasses.Objects[Index] as TStreamObjectClass).Description;
end;

function IsStreamFileBased(const StreamDescription:string):Boolean;
begin
  Result:=TStreamObjectClass(StreamClasses.Objects[GetStreamClassIndexByDescription(StreamDescription)]).ObjectClass.BaseStream=bsFile;
end;

function GetEntryClassIndex(const ClassName:string):Integer;
begin
	Result:=EntryClasses.IndexOf(ClassName);
end;

function GetSectionClassIndex(const ClassName:string):Integer;
begin
	Result:=SectionClasses.IndexOf(ClassName);
end;

function GetEntryClassImageIndex(Index:Integer):Integer;
begin
  Result:=(EntryClasses.Objects[Index] as TEntryObjectClass).ImageIndex;
end;

function GetEntryClassImageIndexByName(const EntryClassName:string):Integer;
var ii:Integer;
begin
  Result:=0;

  ii:=GetEntryClassIndex(EntryClassName);

  if ii<>-1 then Result:=GetEntryClassImageIndex(ii);
end;

initialization
	FormSearchMultiSectionClass:=nil;
  
	EntryClasses:=TStringList.Create;
  SectionClasses:=TStringList.Create;
  StreamClasses:=TStringList.Create;

  DebugMode:=False;
  AutoRename:=True;
  WarnOnDuplicateFieldName:=True;
  DefaultGroupName:=CDefGroup;

	RegisterINIStream(CDefaultFileStream,TINIStreamFile);
	RegisterINIStream(CDefaultRegistryStream,TINIStreamRegistry);

  RegisterClass(TINIEntrySubSection);

finalization
	if EntryClasses<>nil then EntryClasses.Free;
	if SectionClasses<>nil then SectionClasses.Free;
	if StreamClasses<>nil then StreamClasses.Free;

end.
