{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: TINIMultiSectionDatabase
//******************************************************
//******************************************************

unit INISecDB;

interface

uses Classes,Windows,Graphics,Forms,Dialogs,ExtCtrls,INISec,INIEnt,INIEntDB,DB,DBTables;

type
	TINIMultiSectionDatabase=class(TINICustomMultiSection)
  private
  	FParameters:TINICustomMultiSection;
    
  	procedure InitDriverName(Sender:TINIField);

  public
  	constructor Create(AOwner:TComponent); override;

		function SetDatabaseProperties(const dbName:string;Database:TDatabase):Boolean;

  published
		property OnButtonClick;

		property CompressKey;
    property Static;
		property DropDownCount;
    property Buttons;
    property GroupType;
  end;
  
procedure Register;

implementation

constructor TINIMultiSectionDatabase.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  RenameGroupByIndex(0,'Database');
  InsertGroup('Parameters',1,-1,cdTopToBottom,-1);

  GroupType:=gtCascade;

  with TINIEntryDatabaseName.Create(Self) do begin
  	Table:=Self;
  	FieldName:='AliasName';
    DisplayLabel:='Alias name';
    GroupName:='Database';
  end;

  with TINIEntryOneOfStrings.Create(Self) do begin
  	Table:=Self;
  	FieldName:='DriverName';
    DisplayLabel:='Driver name';
    DisplayWidth:=16;
    GroupName:='Database';
    OnInit:=InitDriverName;
  end;

  with TINIEntryBoolean.Create(Self) do begin
  	Table:=Self;
  	FieldName:='KeepConnection';
    DisplayLabel:='Keep connection';
    DisplayWidth:=6;
    DefaultValue:=False;
    GroupName:='Database';
  end;

  with TINIEntryBoolean.Create(Self) do begin
  	Table:=Self;
  	FieldName:='LoginPrompt';
    DisplayLabel:='Login prompt';
    DefaultValue:=False;
    GroupName:='Database';
  end;

  with TINIEntrySubSection.Create(Self) do begin
  	Table:=Self;
  	FieldName:='SS';
    GroupName:='Parameters';
  end;

  FParameters:=TINIMultiSection.Create(Self);

  FParameters.RenameGroupByIndex(0,'Parameter');

  with FParameters do begin
  	Parent:=Self.FieldByName('SS');
    SectionName:='Parameters';
    Buttons:=[isSave,isDelete,isChange,isExchange];
  end;

  with TINIEntryString.Create(Self) do begin
  	Table:=FParameters;
    FieldName:='Value';
    DisplayLabel:='Parameter value';
    CaptionStyle:=csTopLeft;
    DisplayWidth:=0;
    GroupName:='Parameter';
    ShowInGrid:=True;
  end;
end;

procedure TINIMultiSectionDatabase.InitDriverName(Sender:TINIField);
var drList:TStringList;
begin
  if (Session<>nil) and (Sender is TINIEntryOneOfstrings) then begin
    drList:=TStringList.Create;

    try
      Session.GetDriverNames(drList);

      (Sender as TINIEntryOneOfStrings).AssignItems(drList,drList);

      finally
        drList.Free;
    end;
  end;
end;

function TINIMultiSectionDatabase.SetDatabaseProperties(const dbName:string;Database:TDatabase):Boolean;
begin
	Result:=KeyExists(dbName);

	if Result then with Database do begin
    Key:=dbName;

    DatabaseName:=Key;
    KeepConnection:=FieldByName('KeepConnection').AsBoolean;
    LoginPrompt:=FieldByName('LoginPrompt').AsBoolean;
    DriverName:=FieldByName('DriverName').AsString;
    AliasName:=FieldByName('AliasName').AsString;

    Params.Clear;

    FParameters.First;

    while not FParameters.EOL do begin
      Params.Values[FParameters.Key]:=FParameters.FieldByName('Value').AsString;

      FParameters.Next;
    end;
  end;
end;

procedure Register;
begin
	RegisterINISection('BDE aliases',TINIMultiSectionDatabase);

  INIEntDB.Register;
end;

end.
