{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: Component TDBTreeView. Used in INIDsgnF unit 
//******************************************************
//******************************************************

unit INITree;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, CommCtrl,
  ComCtrls,INIMis,ClipBrd,Menus;

type
  TNodeObject=class(TObject)
  private
    FString:string;
    FImageIndex:Integer;
    FChild,FDirty:Boolean;
    FData:Pointer;

  public
    constructor Create(const setString:string);
    constructor CreateImage(const setString:string;setIndex:Integer);
    constructor CreateImageComponent(const setString:string;setIndex:Integer;C:TComponent);
    constructor CreateChild(const setString:string;setChild:Boolean);
    constructor CreateImageChild(const setString:string;setIndex:Integer;setChild:Boolean);
    constructor CreateImageChildComponent(const setString:string;setIndex:Integer;setChild:Boolean;C:TComponent);
    constructor CreateComponent(C:TComponent);
    constructor CreateComponentImage(C:TComponent;setIndex:Integer);

    property StringContext:string read FString write FString;
    property ImageIndex:Integer read FImageIndex write FImageIndex;
    property Child:Boolean read FChild write FChild;
    property Dirty:Boolean read FDirty write FDirty;
    property Data:Pointer read FData write FData;
  end;

  TClipboardNodeComponent=class(TComponent)
  private
    FTree:TTreeView;
    FLevel:Integer;
    FParamName,FValue,FText,FContext:string;
    FImageIndex:Integer;
    FChild:Boolean;

    function GetAsID:LongInt;

  public
    constructor Create(AOwner:TComponent); override;
    constructor CreateTextObject(AOwner:TComponent;const pName,pValue:string);
    constructor CreateIDObject(AOwner:TComponent;const pName:string;pValue:LongInt);

  published
    property Tree:TTreeView read FTree write FTree;
    property Level:Integer read FLevel write FLevel;
    property ParamName:string read FParamName write FParamName;
    property Value:string read FValue write FValue;
    property Text:string read FText write FText;
    property Context:string read FContext write FContext;
    property ImageIndex:Integer read FImageIndex write FImageIndex;
    property Child:Boolean read FChild write FChild;

    property AsID:LongInt read GetAsID;
    property AsText:string read FValue;
  end;

  TDragNodeControl=class(TControl)
  private
    FTree:TTreeView;
    FNode:TTreeNode;
    FLevel:Integer;
    FBackLevel:Integer;
    FDragImages:TImageList;

    function GetAcceptLevel:Integer;

    function GetDragImages:TCustomImageList; override;
    
		procedure CreateDragImages;

  public
    constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

    property Tree:TTreeView read FTree write FTree;
    property Node:TTreeNode read FNode write FNode;
    property Level:Integer read FLevel write FLevel;
    property BackLevel:Integer read FBackLevel write FBackLevel;
    property AcceptLevel:Integer read GetAcceptLevel;
  end;

  TDragNodeControlObject=class(TDragControlObject)
  public
  	constructor Create(DragNodeControl:TDragNodeControl;sTree:TTreeView;sNode:TTreeNode;sLevel:Integer);
  end;

const
  WMU_DELETENODE=WM_USER+1;
  WMU_SCROLL=WM_USER+2;

  CPName='LEVEL%d';
  CIDDelim=',';
  CPathDelim:string='\';
  CDefFieldDelim=', ';
  CDefParamDelim:Char=',';
  CMacroChar:Char='@';

  CPlusDragCursorIndex=10;

type
  TSelections=class(TStringList)
  end;

  TEventOnCustomizeList=procedure(Sender:TObject;Node:TTreeNode;List:TStringList) of object;
  TEventOnDragOver=procedure(Sender:TObject;DragNode,AcceptNode:TTreeNode;const DragParamName:string;var Accept:Boolean) of object;
  TEventOnDragDrop=procedure(Sender:TObject;DropNode,AcceptNode:TTreeNode;const DropParamName:string) of object;
  TEventOnEdited=procedure(Sender:TObject;EditedNode:TTreeNode;var S:string) of object;
  TEventOnPaste=procedure(Sender:TObject;PasteToNode:TTreeNode;const PastedParam,Value,Text,Context:string) of object;
	TEventOnExecuteVerb=procedure(Sender:TObject;PopupNode:TTreeNode;Verb:Integer) of object;
	TEventOnEnableVerb=procedure(Sender:TObject;PopupNode:TTreeNode;Verb:Integer;var Enabled:Boolean) of object;
	TEventOnBeforePopup=procedure(Sender:TObject;PopupNode:TTreeNode;const PopupParam:string) of object;
	TEventOnGetNodeText=procedure(Sender:TObject;var Text:string;Fields:TList) of object;
	TEventOnGetNodeImageIndex=procedure(Sender:TObject;var ImageIndex:Integer) of object;

  TListDistributorProperty=class(TComponent)
  end;

  TListDistributor=class(TListDistributorProperty)
  private
    FImageIndex:Integer;
    FParamName:string;
    FNextLevel:TListDistributorProperty;
    FOnDragOver:TEventOnDragOver;
    FOnDragDrop:TEventOnDragDrop;
    FEditable:Boolean;
    FOnEdited:TEventOnEdited;
    FOnPaste:TEventOnPaste;
    FOnCustomizeList:TEventOnCustomizeList;
    FOnExecuteVerb:TEventOnExecuteVerb;
    FOnEnableVerb:TEventOnEnableVerb;
    FVerbs:TStringList;

    procedure SetNextLevel(LD:TListDistributorProperty);
    procedure SetNOfParameters(pNo:Integer);
    procedure SetVerbs(vList:TStringList);

    function GetParamName:string;

  protected
    FNOfParameters:Integer;
    FData:TObject;

    function DoGetList(listS:TStringList):LongInt; virtual;
    
		procedure Notification(AComponent:TComponent;Operation:TOperation); override;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    function GetList(listS:TStringList;Node,Data:TObject;emptyIt:Boolean):LongInt;

    property Editable:Boolean read FEditable write FEditable stored True default False;
    property OnEdited:TEventOnEdited read FOnEdited write FOnEdited;
    
  published
    property ParamName:string read GetParamName write FParamName;
    property NextLevel:TListDistributorProperty read FNextLevel write SetNextLevel;
    property ImageIndex:Integer read FImageIndex write FImageIndex stored True default -1;
    property NOfParameters:Integer read FNOfParameters write SetNOfParameters;
    property Verbs:TStringList read FVerbs write SetVerbs;

    property OnCustomizeList:TEventOnCustomizeList read FOnCustomizeList write FOnCustomizeList;
    property OnDragOver:TEventOnDragOver read FOnDragOver write FOnDragOver;
    property OnDragDrop:TEventOnDragDrop read FOnDragDrop write FOnDragDrop;
    property OnPaste:TEventOnPaste read FOnPaste write FOnPaste;
    property OnExecuteVerb:TEventOnExecuteVerb read FOnExecuteVerb write FOnExecuteVerb;
    property OnEnableVerb:TEventOnEnableVerb read FOnEnableVerb write FOnEnableVerb;
  end;

  TSimpleListDistributor=class(TListDistributor)
  private
    function GetSplitPos(const item:string):Integer;
    function GetID(ind:Integer):string;
    function GetContext(ind:Integer):string;

  protected
    FList:TStringList;

    procedure SetList(listS:TStringList);

    function DoGetList(listS:TStringList):LongInt; override;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    property ID[ind:Integer]:string read GetID;
    property Context[ind:Integer]:string read GetContext;

  published
    property Editable;
    property List:TStringList read FList write SetList;

    property OnEdited;
  end;

  TListOfListDistributors=class(TStringList)
  protected
    function GetLD(Index:Integer):TListDistributor;

    procedure AddLD(LD:TListDistributor);

  public
    property LD[Index:Integer]:TListDistributor read GetLD;
  end;

  TExtTreeNode=class(TTreeNode)
  public
    destructor Destroy; override;
  end;

  TDBTreeView=class(TTreeView)
  private
    FListDistributors:TListOfListDistributors;
    FFirstLevel:TListDistributor;
    FStructureChanged:Boolean;
    FPathDelimiter:string;
    FMouseDown,FCanceled,FCtrlPressed,FInSetHasChildren,FCollapsing:Boolean;
    FUpdating:Integer;
    FDownX,FDownY:Integer;
    FDownTime:LongInt;
	  FControlDragDrop,FAlwaysRebuild:Boolean;
    FDragNodeControl:TDragNodeControl;
    FScrollingOn,FInEditing:Boolean;
    FDeltaY:Integer;
    FScrollSpeed,FScrollBegin:LongInt;
    FOnBeforePopup:TEventOnBeforePopup;
    FDontBuild,FAlwaysExpand:Boolean;
    FPopupNode,FKeyDownNode,FExpandingNode,FDeletingNode:TTreeNode;
    FPopupParam:string;
    FPopupLD:TListDistributor;
    FLoadedMenuItems:Integer;
    FEnableDragScroll:Boolean;

    function FillList(level:Integer;Node:TTreeNode):LongInt;

    function GetPopupNode:TTreeNode;
    function GetKeyDownNode:TTreeNode;

    function GetLastLevel:Integer;
    function GetInitialized:Boolean;

    procedure SetHasChildren(Node:TTreeNode;Child:Boolean);

    procedure BeginUpdate;
    procedure EndUpdate;

    procedure SetFirstLevel(LD:TListDistributor);
    procedure SetScrollSpeed(sSpeed:LongInt);

		procedure RemoveExtraMenuItems;
    
  protected
    procedure WMDeleteNode(var Msg:TMessage); message WMU_DELETENODE;
    procedure WMScrollTree(var Msg:TMessage); message WMU_SCROLL;

    function GetFineHint(Node:TTreeNode):string;
    function FullPath(Node:TTreeNode;const strDelim:string):string;
    function GetPathOfNode(Node:TTreeNode):string;
    function FullSelectedPath:string;

    function GetSelection(Node:TTreeNode;level,pN:Integer):string;
    function GetSelectionByName(Node:TTreeNode;const pName:string):string;
    function GetSelections(Node:TTreeNode):TSelections;

    function CreateLDList:Boolean;

    procedure DoExit; override;
    procedure Loaded; override;
    procedure Change(Node:TTreeNode); override;

    procedure UpdateHint(Node:TTreeNode;var Msg:TMessage);

    function AlwaysExpand(Node:TTreeNode):Boolean;

    function CanExpand(Node:TTreeNode):Boolean; override;
    function CanCollapse(Node:TTreeNode):Boolean; override;

    procedure Expand(Node:TTreeNode); override;
    procedure Collapse(Node:TTreeNode); override;

    function GetLD(level:Integer):TListDistributor;
    function GetCurrentParam:string;
    function GetCurrentLD:TListDistributor;

    property Lists:TListOfListDistributors read FListDistributors;

		procedure DoOpenPopup(X,Y:Integer);
		procedure OnPopupItemClick(Sender:TObject);

		procedure MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer); override;
		procedure MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer); override;
    procedure MouseMove(Shift:TShiftState;X,Y:Integer); override;

    procedure KeyDown(var Key:Word;Shift:TShiftState); override;

    procedure BeginScroll(Direction:Integer);
    procedure EndScroll;

    procedure DoStartDrag(var DragObject:TDragObject); override;
    procedure DragOver(Source:TObject;X,Y:Integer;State:TDragState;var Accept:Boolean); override;
    procedure DragDrop(Source:TObject;X,Y:Integer); override;

    function GetDragImages:TCustomImageList; override;

		procedure CheckExpandWhileDragging(X,Y:Integer);

		function CanEdit(Node:TTreeNode):Boolean; override;
		procedure Edit(const Item:TTVItem); override;

    procedure OnIdle(Sender:TObject;var Done:Boolean);

    procedure WMMouseMove(var Msg:TWMMouseMove); message WM_MOUSEMOVE;
    
		procedure Notification(AComponent:TComponent;Operation:TOperation); override;

  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;

    property LastLevel:Integer read GetLastLevel;
    property Selection[Node:TTreeNode;level,pN:Integer]:string read GetSelection;
    property SelectionByName[Node:TTreeNode;const pName:string]:string read GetSelectionByName;
    property Path:string read FullSelectedPath;
    property PathOfNode[Node:TTreeNode]:string read GetPathOfNode;

    procedure Add(LD:TListDistributor);

    procedure Init(Restructure:Boolean); virtual;
    procedure DeleteAll; virtual;

    function ExpandIfNotExpanded(Node:TTreeNode):Boolean;
    procedure MakeDirty(Node:TTreeNode);
    procedure RefreshNode(Node:TTreeNode);
    procedure JustExpand(Node:TTreeNode);
    procedure JustCollapse(Node:TTreeNode);

    procedure Copy(Node:TTreeNode);
    procedure Paste(Node:TTreeNode);

    function GoToPath(const fPath:string):Boolean;
    function OpenChildNode(const NodeText:string):Boolean;

    procedure EditNode(Node:TTreeNode);
    procedure FocusNode(Node:TTreeNode);
    procedure DeleteNode(Node:TTreeNode);
    procedure ExpandAll(Node:TTreeNode);
    procedure RestoreNode(Node:TTreeNode);
    procedure RestoreLevel(Node:TTreeNode);

    procedure MoveAsChild(Node,ToNode:TTreeNode);

		procedure DeleteNodeWithData(Node:TTreeNode);
		function CopyNodeWithData(SourceNode,ParentNode:TTreeNode):TTreeNode;
		function MakeNodeWithData(const Text:string;NO:TNodeObject;ParentNode:TTreeNode;Child:Boolean):TTreeNode;

		procedure PreparePopup(Node:TTreeNode); virtual;
		procedure OpenPopup(X,Y:Integer;Node:TTreeNode);

    property LD[level:Integer]:TListDistributor read GetLD;
    property CurrentParam:string read GetCurrentParam;
    property PopupParam:string read FPopupParam;
    property CurrentLD:TListDistributor read GetCurrentLD;
    property CtrlPressed:Boolean read FCtrlPressed;
    property Initialized:Boolean read GetInitialized;
    property ScrollSpeed:LongInt read FScrollSpeed write SetScrollSpeed;
    property PopupNode:TTreeNode read GetPopupNode;
    property KeyDownNode:TTreeNode read GetKeyDownNode;
    property ExpandingNode:TTreeNode read FExpandingNode;
    property Editing:Boolean read FInEditing;

  published
  	property AlwaysRebuild:Boolean read FAlwaysRebuild write FAlwaysRebuild stored True default True;
    property FirstLevel:TListDistributor read FFirstLevel write SetFirstLevel;
    property PathDelimiter:string read FPathDelimiter write FPathDelimiter;
    property ControlDragDrop:Boolean read FControlDragDrop write FControlDragDrop stored True default False;
    property EnableDragScroll:Boolean read FEnableDragScroll write FEnableDragScroll stored True default False;

  	property OnBeforePopup:TEventOnBeforePopup read FOnBeforePopup write FOnBeforePopup;
  end;

function IsAvailableClipboardParam(const ParamName:string):Boolean;
function GetClipboardNodeComponent(AOwner,AParent:TComponent):TClipboardNodeComponent;
procedure InsertAndFreeClipboardNodeComponent(ClipboardNodeComponent:TClipboardNodeComponent);

function GetNodeID(Node:TTreeNode):string;
function GetNodeLevel(Node:TTreeNode):Integer;
function GetNodeParent(Node:TTreeNode;bLevel:Integer):TTreeNode;
function NodeHasInformation(Node:TTreeNode):Boolean;
function NodeIsExpanded(Node:TTreeNode):Boolean;

implementation

function IsAvailableClipboardParam(const ParamName:string):Boolean;
var nodeCComponent:TClipboardNodeComponent;
begin
	Result:=False;

	nodeCComponent:=GetClipboardNodeComponent(Application,Application);

  if nodeCComponent<>nil then begin
  	Result:=nodeCComponent.ParamName=ParamName;

  	nodeCComponent.Free;
  end;
end;

function GetClipboardNodeComponent(AOwner,AParent:TComponent):TClipboardNodeComponent;
var clpComponent:TComponent;
begin
	Result:=nil;

  if Clipboard.HasFormat(CF_COMPONENT) then begin
  	clpComponent:=Clipboard.GetComponent(AOwner,AParent);

    if clpComponent is TClipboardNodeComponent then Result:=clpComponent as TClipboardNodeComponent;
  end;
end;

procedure InsertAndFreeClipboardNodeComponent(ClipboardNodeComponent:TClipboardNodeComponent);
begin
	Clipboard.SetComponent(ClipboardNodeComponent);

  ClipboardNodeComponent.Free;
end;

function GetNodeParent(Node:TTreeNode;bLevel:Integer):TTreeNode;
begin
  while (bLevel>0) and (Node<>nil) do begin
    Node:=Node.Parent;

    bLevel:=bLevel-1;
  end;

  Result:=Node;
end;

function GetNodeID(Node:TTreeNode):string;
begin
	Result:='';

  if (Node<>nil) and (Node.Data<>nil) then Result:=TNodeObject(Node.Data).StringContext;
end;

function GetNodeLevel(Node:TTreeNode):Integer;
const CMaxLevels=128;
begin
  Result:=0;

  while (Node<>nil) and (Result<CMaxLevels) do begin
    Node:=Node.Parent;

    Result:=Result+1;
  end;
end;

function NodeIsExpanded(Node:TTreeNode):Boolean;
begin
	Result:=(Node<>nil) and (Node.GetFirstChild<>nil) and (Node.Expanded) and (Node.HasChildren); 
end;

function NodeHasInformation(Node:TTreeNode):Boolean;
begin
	Result:=(GetNodeID(Node)<>'') or (TNodeObject(Node.Data).Data<>nil); 
end;

constructor TNodeObject.Create(const setString:string);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=-1;
  FChild:=True;
  FDirty:=False;
  FData:=nil;
end;

constructor TNodeObject.CreateImage(const setString:string;setIndex:Integer);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=setIndex;
  FChild:=True;
  FDirty:=False;
  FData:=nil;
end;

constructor TNodeObject.CreateChild(const setString:string;setChild:Boolean);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=-1;
  FChild:=setChild;
  FDirty:=False;
  FData:=nil;
end;

constructor TNodeObject.CreateImageChild(const setString:string;setIndex:Integer;setChild:Boolean);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=setIndex;
  FChild:=setChild;
  FDirty:=False;
  FData:=nil;
end;

constructor TNodeObject.CreateImageChildComponent(const setString:string;setIndex:Integer;setChild:Boolean;C:TComponent);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=setIndex;
  FChild:=setChild;
  FDirty:=False;
  FData:=C;
end;

constructor TNodeObject.CreateImageComponent(const setString:string;setIndex:Integer;C:TComponent);
begin
  inherited Create;

  FString:=setString;
  FImageIndex:=setIndex;
  FChild:=True;
  FDirty:=False;
  FData:=C;
end;

constructor TNodeObject.CreateComponentImage(C:TComponent;setIndex:Integer);
begin
  inherited Create;

  FString:=C.Name;
  FImageIndex:=setIndex;
  FChild:=True;
  FDirty:=False;
  FData:=C;
end;

constructor TNodeObject.CreateComponent(C:TComponent);
begin
  inherited Create;

  FString:=C.Name;
  FImageIndex:=-1;
  FChild:=True;
  FDirty:=False;
  FData:=C;
end;

constructor TClipboardNodeComponent.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FTree:=nil;
  FLevel:=0;
  FParamName:='';
  FValue:='';
  FText:='';
  FContext:='';
  FImageIndex:=-1;
  FChild:=True;
end;

constructor TClipboardNodeComponent.CreateTextObject(AOwner:TComponent;const pName,pValue:string);
begin
	inherited Create(AOwner);

  FTree:=nil;
  FLevel:=0;
  FParamName:=pName;
  FValue:=pValue;
  FText:='';
  FContext:='';
  FImageIndex:=-1;
  FChild:=True;
end;

constructor TClipboardNodeComponent.CreateIDObject(AOwner:TComponent;const pName:string;pValue:LongInt);
begin
	inherited Create(AOwner);

  FTree:=nil;
  FLevel:=0;
  FParamName:=pName;
  FValue:=IntToStr(pValue);
  FText:='';
  FContext:='';
  FImageIndex:=-1;
  FChild:=True;
end;

function TClipboardNodeComponent.GetAsID:LongInt;
begin
	Result:=StrToIntDef(FValue,-1);
end;

constructor TDragNodeControlObject.Create(DragNodeControl:TDragNodeControl;sTree:TTreeView;sNode:TTreeNode;sLevel:Integer);
begin
	inherited Create(DragNodeControl);

  DragNodeControl.Tree:=sTree;
  DragNodeControl.Node:=sNode;
  DragNodeControl.Level:=sLevel;
  
  DragNodeControl.DragCursor:=sTree.DragCursor;

  DragNodeControl.CreateDragImages;
end;

constructor TDragNodeControl.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FTree:=nil;
  FNode:=nil;
  FLevel:=0;
  FBackLevel:=1;
  FDragImages:=nil;
end;

destructor TDragNodeControl.Destroy;
begin
	if FDragImages<>nil then FDragImages.Free;
  
	inherited Destroy;
end;

procedure TDragNodeControl.CreateDragImages;
var FBMP:TBitmap;
begin
	FDragImages:=TImageList.Create(Self);

  if (FTree<>nil) and (FTree.Images<>nil) then begin
	  FBMP:=TBitmap.Create;

    try
		  FTree.Images.GetBitmap(FNode.ImageIndex,FBMP);

  	  FDragImages.Add(FBMP,FBMP);

      finally
		    FBMP.Free;
    end;
  end;
end;

function TDragNodeControl.GetDragImages:TCustomImageList;
begin
	Result:=FDragImages;
end;

function TDragNodeControl.GetAcceptLevel:Integer;
begin
  Result:=FLevel-FBackLevel;
end;

constructor TListDistributor.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

  FParamName:=Name;
  FImageIndex:=-1;
  FNOfParameters:=1;

  FOnCustomizeList:=nil;
  FOnDragOver:=nil;
  FOnDragDrop:=nil;
  FOnPaste:=nil;
  FOnExecuteVerb:=nil;
  FOnEnableVerb:=nil;

  FNextLevel:=nil;

  FEditable:=False;
  FOnEdited:=nil;

  FVerbs:=TStringList.Create;
end;

destructor TListDistributor.Destroy;
begin
	FVerbs.Free;
  
  inherited Destroy;
end;

function TListDistributor.GetParamName:string;
begin
  if FParamName<>'' then
    Result:=FParamName
  else
    Result:=Name;
end;

procedure TListDistributor.SetNOfParameters(pNo:Integer);
begin
  if pNo>0 then FNOfParameters:=pNo;
end;

procedure TListDistributor.SetNextLevel(LD:TListDistributorProperty);
begin
  FNextLevel:=LD;
end;

procedure TListDistributor.SetVerbs(vList:TStringList);
begin
	FVerbs.Assign(vList);
end;

function TListDistributor.DoGetList(listS:TStringList):LongInt;
begin
  Result:=0;
end;

function TListDistributor.GetList(listS:TStringList;Node,Data:TObject;emptyIt:Boolean):LongInt;
begin
  if emptyIt then listS.Clear;

  FData:=Data;

  try
    Result:=DoGetList(listS);

    if Assigned(FOnCustomizeList) then begin
    	if Node is TTreeNode then
	    	FOnCustomizeList(Self,Node as TTreeNode,listS)
      else
	    	FOnCustomizeList(Self,nil,listS);

      Result:=listS.Count;
    end;

    finally
      if FData<>nil then FData.Free;
  end;
end;

procedure TListDistributor.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if (Operation=opRemove) and (AComponent=FNextLevel) then FNextLevel:=nil;
end;

procedure TListOfListDistributors.AddLD(LD:TListDistributor);
begin
  AddObject('',LD);
end;

function TListOfListDistributors.GetLD(Index:Integer):TListDistributor;
begin
  Result:=nil;

  if Objects[Index] is TListDistributor then Result:=Objects[Index] as TListDistributor;
end;

destructor TExtTreeNode.Destroy;
begin
  if Data<>nil then begin
    TObject(Data).Free;
  end;

  inherited Destroy;
end;

constructor TDBTreeView.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

 	FDragNodeControl:=nil;

  if not (csDesigning in ComponentState) then FDragNodeControl:=TDragNodeControl.Create(Self);

  FFirstLevel:=nil;
  FStructureChanged:=False;
  FControlDragDrop:=False;
  FAlwaysRebuild:=True;
  FPathDelimiter:=CPathDelim;

	FCanceled:=False;  
  FMouseDown:=False;
  FCtrlPressed:=False;
	FInSetHasChildren:=False;
  FCollapsing:=False;
  FDownX:=-1;
  FDownY:=-1;
	FDownTime:=0;

  FScrollingOn:=False;
  FScrollBegin:=DateTimeToTimeStamp(Now).Time;
  FScrollSpeed:=32;
  FDeltaY:=32;

  FLoadedMenuItems:=0;

  FDontBuild:=False;
  FAlwaysExpand:=False;
  FInEditing:=False;
  
	FEnableDragScroll:=False;
  FUpdating:=0;

  FOnBeforePopup:=nil;

  FExpandingNode:=nil;
  FPopupNode:=nil;
  FKeyDownNode:=nil;
  FDeletingNode:=nil;

  FPopupParam:='';

  FListDistributors:=TListOfListDistributors.Create;
end;

destructor TDBTreeView.Destroy;
begin
  FListDistributors.Free;

  EndScroll;

  RemoveExtraMenuItems;

  inherited Destroy;
end;

procedure TDBTreeView.OnIdle(Sender:TObject;var Done:Boolean);
var curP:TPoint;
begin
  GetCursorPos(curP);

  curP:=ScreenToClient(curP);

	CheckExpandWhileDragging(curP.X,curP.Y);
end;

procedure TDBTreeView.SetScrollSpeed(sSpeed:LongInt);
begin
	if sSpeed>=0 then FScrollSpeed:=sSpeed;
end;

procedure TDBTreeView.SetFirstLevel(LD:TListDistributor);
begin
  FFirstLevel:=LD;

  FStructureChanged:=True;
end;

procedure TDBTreeView.Loaded;
begin
  inherited Loaded;

  if PopupMenu<>nil then FLoadedMenuItems:=PopupMenu.Items.Count;
end;

function TDBTreeView.GetDragImages;
begin
	Result:=Images;
end;

procedure TDBTreeView.DoExit;
begin
  inherited DoExit;

  ReleaseCapture;

	FCanceled:=True;
  FMouseDown:=False;

  EndScroll;
end;

procedure TDBTreeView.DeleteNodeWithData(Node:TTreeNode);
begin
	FDeletingNode:=Node;

  try
    if Node<>nil then begin
      if Node.Data<>nil then TObject(Node.Data).Free;

      while Node.GetFirstChild<>nil do DeleteNodeWithData(Node.GetFirstChild);

      Node.Delete;
    end;

    finally
    	FDeletingNode:=nil;
  end;
end;

procedure TDBTreeView.Change(Node:TTreeNode);
begin
	if FUpdating=0 then begin
	  FPopupNode:=Node;
	  FKeyDownNode:=Node;
	  FExpandingNode:=Node;

  	if not FMouseDown then inherited Change(Node);
  end;
end;

procedure TDBTreeView.RemoveExtraMenuItems;
var i,ri:Integer;
begin
	if PopupMenu<>nil then begin
		ri:=PopupMenu.Items.Count-FLoadedMenuItems;

		for i:=1 to ri do PopupMenu.Items.Delete(PopupMenu.Items.Count-1);
  end;
end;

procedure TDBTreeView.WMMouseMove(var Msg:TWMMouseMove);
var Shift:TShiftState;
begin
	Shift:=[];

  if (Msg.Keys and MK_CONTROL)<>0 then Shift:=Shift+[ssCtrl];
  if (Msg.Keys and MK_SHIFT)<>0 then Shift:=Shift+[ssShift];

	MouseMove(Shift,Msg.XPos,Msg.YPos);

  UpdateHint(GetNodeAt(Msg.XPos,Msg.YPos),TMessage(Msg));
end;

procedure TDBTreeView.UpdateHint(Node:TTreeNode;var Msg:TMessage);
begin
	if ShowHint then begin
  	Hint:=GetFineHint(Node);

    Application.HintMouseMessage(Self,Msg);
  end;
end;

function TDBTreeView.CreateLDList:Boolean;
var LD:TComponent;
begin
  Lists.Clear;

  LD:=FirstLevel;

  while (LD is TListDistributor) do begin
    Lists.AddLD(LD as TListDistributor);

    LD:=(LD as TListDistributor).NextLevel;
  end;

  Result:=Lists.Count>0;
end;

function TDBTreeView.GetPopupNode:TTreeNode;
begin
	if FPopupNode=nil then FPopupNode:=Selected;

  Result:=FPopupNode;
end;

function TDBTreeView.GetKeyDownNode:TTreeNode;
begin
	if FKeyDownNode=nil then FKeyDownNode:=Selected;

  Result:=FKeyDownNode;
end;

function TDBTreeView.GetLD(level:Integer):TListDistributor;
var i:Integer;
begin
	if (Lists.Count=0) and (FirstLevel<>nil) then begin
  	Result:=FirstLevel;

    for i:=1 to level do begin
    	if Result<>nil then Result:=Result.NextLevel as TListDistributor;
    end;
  end else begin
	  Result:=Lists.LD[level];
  end;
end;

function TDBTreeView.GetCurrentParam:string;
begin
  Result:='';
  
  if (Selected<>nil) then begin
    try
      Result:=LD[GetNodeLevel(Selected)-1].ParamName;

      except
        on Exception do Result:='';
    end;
  end;
end;

function TDBTreeView.GetCurrentLD:TListDistributor;
begin
  Result:=nil;
  
  if (Selected<>nil) then begin
    try
      Result:=LD[GetNodeLevel(Selected)-1];

      except
        on Exception do Result:=nil;
    end;
  end;
end;

function TDBTreeView.GetInitialized:Boolean;
begin
	Result:=not FStructureChanged;
end;

function TDBTreeView.GetLastLevel:Integer;
begin
  Result:=Lists.Count;
end;

function TDBTreeView.GetPathOfNode(Node:TTreeNode):string;
begin
	Result:=FullPath(Node,FPathDelimiter);
end;

function TDBTreeView.FullSelectedPath:string;
begin
  Result:=FullPath(Selected,FPathDelimiter);
end;

procedure TDBTreeView.SetHasChildren(Node:TTreeNode;Child:Boolean);
begin
	FInSetHasChildren:=True;

  try
  	Node.HasChildren:=Child;

  	finally
    	FInSetHasChildren:=False;
  end;
end;

procedure TDBTreeView.MoveAsChild(Node,ToNode:TTreeNode);
begin
  if NodeIsExpanded(ToNode) then begin
  	if Node.Parent<>nil then with Node.Parent do begin
    	if (GetFirstChild<>nil) and (GetNextChild(GetFirstChild)=nil) then begin
      	JustCollapse(Node.Parent);

        SetHasChildren(Node.Parent,False);
      end;
    end;

    Node.MoveTo(ToNode,naAddChild);
  end else begin
    DeleteNode(Node);
  end;

  TNodeObject(ToNode.Data).Dirty:=True;
  RestoreNode(ToNode);
end;

function TDBTreeView.CopyNodeWithData(SourceNode,ParentNode:TTreeNode):TTreeNode;
begin
  Result:=Items.AddChildObject(ParentNode,SourceNode.Text,
                       TNodeObject.CreateImageChild(TNodeObject(SourceNode.Data).StringContext,
                       											TNodeObject(SourceNode.Data).ImageIndex,
                                            TNodeObject(SourceNode.Data).Child));

	SetHasChildren(Result,TNodeObject(SourceNode.Data).Child);
end;

function TDBTreeView.MakeNodeWithData(const Text:string;NO:TNodeObject;ParentNode:TTreeNode;Child:Boolean):TTreeNode;
begin
	Result:=nil;

	if Text<>'' then begin
    Result:=Items.AddChildObject(ParentNode,Text,NO);

    Result.ImageIndex:=NO.ImageIndex;
    Result.SelectedIndex:=NO.ImageIndex;

    SetHasChildren(Result,Child);
  end;
end;

function TDBTreeView.FullPath(Node:TTreeNode;const strDelim:string):string;
begin
  Result:='';

  while Node<>nil do begin
    if Result<>'' then
      Result:=Node.Text+strDelim+Result
    else
      Result:=Node.Text;

    Node:=Node.Parent;
  end;
end;

function TDBTreeView.GetFineHint(Node:TTreeNode):string;
var level:Integer;
begin
  Result:='';

  level:=GetNodeLevel(Node)-1;

  while Node<>nil do begin
		Result:=Space(level)+Node.Text+strIf(Result='','',#13)+Result;

    level:=level-1;

    Node:=Node.Parent;
  end;
end;

procedure TDBTreeView.MakeDirty(Node:TTreeNode);
begin
  if (Node<>nil) and (Node.Data<>nil) then TNodeObject(Node.Data).Dirty:=True;
end;

procedure TDBTreeView.RefreshNode(Node:TTreeNode);
begin
	if NodeIsExpanded(Node) or (not FAlwaysRebuild) then begin
    while Node.GetFirstChild<>nil do begin
      DeleteNodeWithData(Node.GetFirstChild);
    end;

    SetHasChildren(Node,True);
  end else begin
  	RestoreNode(Node);
  end;

  if AlwaysExpand(Node) then JustExpand(Node);

  if (Node<>nil) and (Node.Data<>nil) then TNodeObject(Node.Data).Dirty:=False;
end;

procedure TDBTreeView.EditNode(Node:TTreeNode);
begin
	if Node<>nil then begin
		if LD[GetNodeLevel(Node)-1].Editable then begin
    	FInEditing:=True;

      Node.EditText;
    end;
  end;
end;

function TDBTreeView.ExpandIfNotExpanded(Node:TTreeNode):Boolean;
begin
  if (not NodeIsExpanded(Node)) and AlwaysExpand(Node) then JustExpand(Node);

  Result:=NodeIsExpanded(Node);
end;

procedure TDBTreeView.JustExpand(Node:TTreeNode);
begin
	if (Node.GetFirstChild<>nil) and (Node.GetFirstChild.Text<>'') then begin
    FDontBuild:=True;

    try
      Node.Expanded:=True;

      finally
        FDontBuild:=False;
    end;
  end else begin
  	Node.Expanded:=False;
  end;
end;

procedure TDBTreeView.JustCollapse(Node:TTreeNode);
begin
  FDontBuild:=True;

  try
    Node.Expanded:=False;

    finally
      FDontBuild:=False;
  end;
end;

function TDBTreeView.AlwaysExpand(Node:TTreeNode):Boolean;
begin
	FAlwaysExpand:=True;

  try
	  Result:=CanExpand(Node);

    finally
    	FAlwaysExpand:=False;
  end;
end;

function TDBTreeView.CanExpand(Node:TTreeNode):Boolean;
var lc:LongInt;

  procedure DoBeforeExpand(DeleteNodes:Boolean);
  var lNode,dNode:TTreeNode;
  begin
    FExpandingNode:=Node;

    if DeleteNodes then begin
	  	lNode:=Node.GetFirstChild;

  	  while lNode<>nil do begin
    		dNode:=lNode;

	    	lNode:=Node.GetNextChild(lNode);

  	    DeleteNodeWithData(dNode);
    	end;
    end;
  end;

  procedure DoAfterExpand;
  begin
    Result:=(lc>0);

    FExpandingNode:=nil;
  end;

  procedure DoCanExpand;
  begin
  	if FCollapsing then begin
    	Result:=False;
    end else if (NodeIsExpanded(Node) or FDontBuild) and (not FAlwaysExpand) then begin
	    Result:=inherited CanExpand(Node);
    end else begin
      Result:=True;

      try
      	if FAlwaysRebuild or (Node.GetFirstChild=nil) or (TNodeObject(Node.Data).Dirty) then begin
	        DoBeforeExpand(True);

  	      lc:=FillList(GetNodeLevel(Node),Node);

          TNodeObject(Node.Data).Dirty:=False;
        end else begin
        	DoBeforeExpand(False);
        end;

     	  DoAfterExpand;

        except
          on Exception do begin
            lc:=0;

            DoAfterExpand;

            raise;
          end;
      end;
    end;
  end;

begin
	BeginUpdate;

  try
  	DoCanExpand;

    finally
    	EndUpdate;
  end;

  if not Result then SetHasChildren(Node,False);
end;

function TDBTreeView.CanCollapse(Node:TTreeNode):Boolean;
begin
	Result:=inherited CanCollapse(Node);
end;

procedure TDBTreeView.Expand(Node:TTreeNode);
begin
	if not FCollapsing then inherited Expand(Node);
end;

procedure TDBTreeView.Collapse(Node:TTreeNode);
begin
	inherited Collapse(Node);

	if (FAlwaysRebuild) and (not FDontBuild) and (not FCollapsing) and (not FInSetHasChildren) and (FDeletingNode=nil) then begin
    BeginUpdate;

    Application.ProcessMessages;

    PostMessage(Self.Handle,WMU_DELETENODE,WPARAM(Node),0);
  end;
end;

procedure TDBTreeView.Copy(Node:TTreeNode);
var level:Integer;
		CopyNode:TClipboardNodeComponent;
begin
	level:=GetNodeLevel(Node)-1;

  CopyNode:=TClipboardNodeComponent.Create(Self);

  try
    CopyNode.Tree:=Self;
    CopyNode.Level:=level;
    CopyNode.ParamName:=LD[level].ParamName;
    CopyNode.Value:=SelectionByName[Node,LD[level].ParamName];
    CopyNode.Text:=Node.Text;
    CopyNode.Context:=TNodeObject(Node.Data).StringContext;
    CopyNode.ImageIndex:=TNodeObject(Node.Data).ImageIndex;
    CopyNode.Child:=TNodeObject(Node.Data).Child;

    Clipboard.SetComponent(CopyNode);

    finally
    	CopyNode.Free;
  end;
end;

procedure TDBTreeView.Paste(Node:TTreeNode);
var evPaste:TEventOnPaste;
    level:Integer;
    pastedNode:TClipboardNodeComponent;
    clpComponent:TComponent;
begin
	if Clipboard.HasFormat(CF_COMPONENT) then begin
  	clpComponent:=Clipboard.GetComponent(Self,Self);

    try
      if clpComponent is TClipboardNodeComponent then begin
        pastedNode:=clpComponent as TClipboardNodeComponent;

        level:=GetNodeLevel(Node)-1;

        evPaste:=LD[level].OnPaste;

        if Assigned(evPaste) then with pastedNode do begin
          evPaste(Self,Node,ParamName,Value,Text,Context);
        end;
      end;

      finally
      	RemoveComponent(clpComponent);
      	clpComponent.Free;
    end;
  end;
end;

function TDBTreeView.GoToPath(const fPath:string):Boolean;
var ofs:Integer;
		Token:string;
    tokenDelim:Char;
    sNode,Node:TTreeNode;

	function FindNode(TopNode:TTreeNode;const NodeText:string):TTreeNode;
  begin
  	Result:=TopNode;

  	while (Result<>nil) and not StrICmp(Result.Text,NodeText) do begin
    	Result:=Result.GetNextSibling;
    end;
  end;

  procedure DoMain;
  begin
    sNode:=nil;
    Node:=Items.GetFirstNode;

    if Length(PathDelimiter)>0 then
      tokenDelim:=PathDelimiter[1]
    else
      tokenDelim:=CPathDelim[1];

    ofs:=1;
    Token:=GetToken(fPath,[#0..#255]-[tokenDelim],ofs);

    while (Token<>'') and (Node<>nil) do begin
      Node:=FindNode(Node,Token);

      Token:=GetToken(fPath,[#0..#255]-[tokenDelim],ofs);

      if Node<>nil then begin
        sNode:=Node;

        if Token<>'' then begin
	        if not NodeIsExpanded(Node) then begin
  	      	AlwaysExpand(Node);
    	    	JustExpand(Node);
      	  end;

        	Node:=Node.GetFirstChild;
        end;
      end;
    end;

    if sNode<>nil then begin
      sNode.MakeVisible;

      sNode.Focused:=True;
      sNode.Selected:=True;

      Selected:=sNode;

      Result:=True;
    end;
  end;

begin
  Result:=False;

	BeginUpdate;

  try
  	DoMain;

    finally
    	EndUpdate;
  end;
end;

function TDBTreeView.OpenChildNode(const NodeText:string):Boolean;
var ParentNode,Node:TTreeNode;
begin
	Result:=False;
  
	if Selected<>nil then begin
    ParentNode:=Selected;

    if not NodeIsExpanded(ParentNode) then begin
      AlwaysExpand(ParentNode);
      JustExpand(ParentNode);
    end;

    Node:=ParentNode.GetFirstChild;

    while Node<>nil do begin
      if AnsiCompareText(Node.Text,NodeText)=0 then begin
        Node.MakeVisible;

        Node.Focused:=True;
        Node.Selected:=True;

        Selected:=Node;

        Result:=True;

        Exit;
      end;

      Node:=ParentNode.GetNextChild(Node);
    end;
  end;
end;

procedure TDBTreeView.FocusNode(Node:TTreeNode);
begin
 	if Node<>nil then begin
  	Node.MakeVisible;

  	Selected:=Node;
    Node.Focused:=True;

  	FPopupNode:=Node;
    FKeyDownNode:=Node;
  end;
end;

procedure TDBTreeView.DeleteNode(Node:TTreeNode);

	procedure SetNilNode(Node:TTreeNode;var NilNode:TTreeNode);

    function IterateChildren(Node:TTreeNode):Boolean;
    var ChildNode:TTreeNode;
    begin
      Result:=True;

      if NilNode=Node then begin
        NilNode:=nil;

        Exit;
      end;

      ChildNode:=Node.GetFirstChild;

      while ChildNode<>nil do begin
        if IterateChildren(ChildNode) then Exit;

        ChildNode:=Node.GetNextChild(ChildNode);
      end;

      Result:=False;
    end;

  begin
  	IterateChildren(Node);
  end;

begin
	if Node<>nil then begin
  	if Node.Parent<>nil then with Node.Parent do begin
    	if (GetFirstChild<>nil) and (GetNextChild(GetFirstChild)=nil) then begin
      	JustCollapse(Node.Parent);

        SetHasChildren(Node.Parent,False);
      end;
    end;

  	SetNilNode(Node,FExpandingNode);
  	SetNilNode(Node,FPopupNode);
  	SetNilNode(Node,FKeyDownNode);

    if Node.GetNextSibling<>nil then
    	FocusNode(Node.GetNextSibling)
  	else if Node.GetPrevSibling<>nil then
    	FocusNode(Node.GetPrevSibling)
    else if Node.Parent<>nil then
	   	FocusNode(Node.Parent);

  	DeleteNodeWithData(Node);
  end;
end;

procedure TDBTreeView.RestoreNode(Node:TTreeNode);
begin
 	if (Node<>nil) and (not NodeIsExpanded(Node)) and (Node.GetFirstChild=nil) and (Node.Data<>nil) and TNodeObject(Node.Data).Child then begin
  	SetHasChildren(Node,True);
  end;

  MakeDirty(Node);
end;

procedure TDBTreeView.ExpandAll(Node:TTreeNode);

  procedure ExpandNode(Node:TTreeNode);
  var lNode:TTreeNode;
  begin
    if not NodeIsExpanded(Node) then AlwaysExpand(Node);

  	lNode:=Node.GetFirstChild;

    while lNode<>nil do begin
    	ExpandNode(lNode);

      lNode:=lNode.GetNextChild(lNode);
    end;

    JustExpand(Node);
  end;

begin
	BeginUpdate;

  try
  	ExpandNode(Node);

    finally
    	EndUpdate;
  end;
end;

procedure TDBTreeView.RestoreLevel(Node:TTreeNode);
begin
	while Node<>nil do begin
  	RestoreNode(Node);

  	Node:=Node.GetNextSibling;
  end;
end;

procedure TDBTreeView.WMDeleteNode(var Msg:TMessage);
var Node:TTreeNode;
begin
	FCollapsing:=True;

  try
    Node:=TTreeNode(Msg.WParam);

    if Node<>nil then begin
      while Node.GetFirstChild<>nil do begin
        DeleteNodeWithData(Node.GetFirstChild);
      end;

      SetHasChildren(Node,True);
    end;

    finally
    	EndUpdate;

      FCollapsing:=False;
  end;
end;

procedure TDBTreeView.Add(LD:TListDistributor);
begin
	if Lists.Count=0 then
  	FirstLevel:=LD
  else
	  Lists.LD[Lists.Count-1].NextLevel:=LD;

  Lists.AddLD(LD);

  FStructureChanged:=True;
end;

procedure TDBTreeView.DeleteAll;
begin
	while Items.GetFirstNode<>nil do DeleteNodeWithData(Items.GetFirstNode);

  Lists.Clear;

  FStructureChanged:=True;
end;

procedure TDBTreeView.BeginUpdate;
begin
	FUpdating:=FUpdating+1;
end;

procedure TDBTreeView.EndUpdate;
begin
  if FUpdating=1 then Change(Selected);

	if FUpdating>0 then FUpdating:=FUpdating-1;
end;

procedure TDBTreeView.Init(Restructure:Boolean);
var doFill:Boolean;
begin
  doFill:=True;

  if Restructure and (FFirstLevel<>nil) then begin
    Lists.Clear;

    doFill:=CreateLDList;
  end;

  if doFill then begin
    FillList(0,Items.GetFirstNode);

    FStructureChanged:=False;

    if Items.GetFirstNode<>nil then begin
    	Selected:=Items.GetFirstNode;

      Items.GetFirstNode.Selected:=True;
      Items.GetFirstNode.Focused:=True;

      Items.GetFirstNode.MakeVisible;
    end;
  end;
end;

function TDBTreeView.GetSelection(Node:TTreeNode;level,pN:Integer):string;
var selN:TSelections;
begin
  Result:='';

  if Node<>nil then begin
    selN:=GetSelections(Node);

    try
      if selN.Count>level then begin
        try
          Result:=selN.Values[Format('%s:%d',[LD[level].ParamName,pN+1])]

          except
            on Exception do Result:='';
        end;
      end;

      finally
        selN.Free;
    end;
  end;
end;

function TDBTreeView.GetSelectionByName(Node:TTreeNode;const pName:string):string;
var selN:TSelections;
begin
  Result:='';

  if Node<>nil then begin
    selN:=GetSelections(Node);

    try
      try
        Result:=selN.Values[pName]

        except
          on Exception do Result:='';
      end;

      finally
        selN.Free;
    end;
  end;
end;

function TDBTreeView.GetSelections(Node:TTreeNode):TSelections;
var i,ml,ofs:Integer;
    pName,pValues,pValue:string;

	procedure SetValue(pName:string;const pValue:string;Index:Integer);
  var lastV:string;
  begin
  	if Index>0 then pName:=Format('%s:%d',[pName,Index]);

  	try
    	lastV:=Result.Values[pName];

      except
      	on EListError do lastV:='';
    end;

    if lastV<>'' then
    	SetValue(pName,pValue,Index+1)
    else
    	Result.Values[pName]:=pValue;
  end;

begin
  Result:=TSelections.Create;

  ml:=GetNodeLevel(Node);

  while Node<>nil do begin
    ml:=ml-1;

    pName:=LD[ml].ParamName;

    if not (pName<>'') then pName:=Format(CPName,[ml]);

    if Node.Data<>nil then
      pValues:=TNodeObject(Node.Data).StringContext
    else
      pValues:=pName;

    ofs:=1;

    for i:=0 to LD[ml].NOfParameters-1 do begin
      pValue:=GetToken(pValues,[#0..#255]-[CDefParamDelim],ofs);

      SetValue(Format('%s:%d',[pName,i+1]),pValue,0);
    end;

    if LD[ml].NOfParameters=1 then SetValue(pName,pValue,0);

    Node:=Node.Parent;
  end;
end;

procedure TDBTreeView.OnPopupItemClick(Sender:TObject);
var evExecuteVerb:TEventOnExecuteVerb;
begin
	if (FPopupLD<>nil) and (FPopupLD.ParamName=CurrentParam) then begin
  	if (Sender as TMenuItem).Caption<>'-' then begin
      evExecuteVerb:=FPopupLD.OnExecuteVerb;

      if Assigned(evExecuteVerb) then evExecuteVerb(Self,PopupNode,(Sender as TMenuItem).Tag);
    end;
  end;
end;

procedure TDBTreeView.PreparePopup(Node:TTreeNode);
var i,ic:Integer;
		verbEnabled:Boolean;
    evEnableVerb:TEventOnEnableVerb;
    lastItem,Items:TMenuItem;
    mLevel,nLevel:Integer;

	function AddPopupItem(const mCaption:string;tValue:Integer):TMenuItem;
  var miCaption,sShortCut:string;
  		scBegin,scEnd:Integer;
  begin
  	Result:=TMenuItem.Create(nil);

    with Result do begin
    	scBegin:=Pos('[',mCaption);

    	if scBegin>0 then begin
      	scEnd:=Pos(']',mCaption);

        if scEnd=0 then scEnd:=Length(mCaption);

      	miCaption:=System.Copy(mCaption,1,scBegin-1);
        sShortCut:=System.Copy(mCaption,scBegin+1,scEnd-scBegin-1);
      end else begin
      	miCaption:=mCaption;
        sShortCut:='';
      end;

    	Caption:=LTrim(miCaption);
      ShortCut:=TextToShortCut(sShortCut);
      Tag:=tValue;

      OnClick:=OnPopupItemClick;
    end;

	  Items.Add(Result);
  end;

begin
  if (Node<>nil) and (PopupMenu<>nil) then begin
  	if Selected<>Node then begin
	    Selected:=Node;
  	 	Selected.Focused:=True;
    end;

    FPopupLD:=LD[GetNodeLevel(Node)-1];
    FPopupParam:=FPopupLD.ParamName;

    Items:=PopupMenu.Items;

    ic:=FPopupLD.Verbs.Count;

    RemoveExtraMenuItems;

    if ic>0 then begin
      lastItem:=AddPopupItem('-',0);

      mLevel:=0;

      for i:=1 to ic do begin
        verbEnabled:=True;

        evEnableVerb:=FPopupLD.OnEnableVerb;

        if Assigned(evEnableVerb) then evEnableVerb(Self,Node,i-1,verbEnabled);

        nLevel:=Blanks(FPopupLD.Verbs.Strings[i-1]);

        if mLevel>nLevel then
        	Items:=Items.Parent
        else if mLevel<nLevel then
        	Items:=lastItem;

        lastItem:=AddPopupItem(FPopupLD.Verbs.Strings[i-1],i-1);
        lastItem.Enabled:=verbEnabled;

        mLevel:=nLevel;
      end;
    end;
  end;
end;

procedure TDBTreeView.OpenPopup(X,Y:Integer;Node:TTreeNode);
begin
	FPopupNode:=Node;

	DoOpenPopup(X,Y);
end;

procedure TDBTreeView.DoOpenPopup(X,Y:Integer);
var PopupPoint:TPoint;
begin
  if (FPopupNode<>nil) and (PopupMenu<>nil) then begin
  	PreparePopup(FPopupNode);

    if Assigned(FOnBeforePopup) then FOnBeforePopup(Self,PopupNode,FPopupParam);

    PopupPoint.X:=X;
    PopupPoint.Y:=Y;

    PopupPoint:=ClientToScreen(PopupPoint);

    PopupMenu.Popup(PopupPoint.X,PopupPoint.Y);
  end;
end;

procedure TDBTreeView.MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
var ht:THitTests;
begin
	inherited MouseDown(Button,Shift,X,Y);

  if (Button=mbLeft) and FControlDragDrop and not FCanceled then begin
  	ht:=GetHitTestInfoAt(X,Y);

  	if (FUpdating=0) and (htOnItem in ht) then begin
	    FMouseDown:=True;
  	  FCtrlPressed:=ssCtrl in Shift;
    	FDownX:=X;
	    FDownY:=Y;
      FDownTime:=DateTimeToTimeStamp(Time).Time;
      
      SetCapture(Self.Handle);
    end;
  end else if (Button=mbRight) and (PopupMenu<>nil) then begin
	  FPopupNode:=GetNodeAt(X,Y);

  	DoOpenPopup(X,Y);
  end;
end;

procedure TDBTreeView.MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
	inherited MouseUp(Button,Shift,X,Y);

  if FMouseDown and FControlDragDrop then begin
  	EndDrag(False);
  end;

  ReleaseCapture;

  FMouseDown:=False;
  FCanceled:=False;

  EndScroll;
end;

procedure TDBTreeView.MouseMove(Shift:TShiftState;X,Y:Integer);
begin
	inherited MouseMove(Shift,X,Y);

  if FMouseDown then begin
    if ((Abs(FDownX-X)>3) or (Abs(FDownY-Y)>1)) and (GetNodeAt(FDownX,FDownY)=GetNodeAt(X,Y)) and ((DateTimeToTimeStamp(Time).Time-FDownTime)<500) then begin
    	FCanceled:=False;
		 	FMouseDown:=False;

      if FControlDragDrop and (GetNodeAt(FDownX,FDownY)<>nil) then begin
      	if (not FCtrlPressed) or (Screen.Cursors[CPlusDragCursorIndex]=65544) then
        	DragCursor:=crDrag
        else
        	DragCursor:=CPlusDragCursorIndex;

      	BeginDrag(True);
      end;

    	FDownX:=X;
      FDownY:=Y;
    end;
  end;
end;

procedure TDBTreeView.KeyDown(var Key:Word;Shift:TShiftState);
var curP:TPoint;
begin
  FKeyDownNode:=Selected;

 	if FPopupParam<>CurrentParam then PreparePopup(FKeyDownNode);

	inherited KeyDown(Key,Shift);

  if Key=VK_F2 then begin
  	if (not FInEditing) and (CurrentLD<>nil) and CurrentLD.Editable then begin
    	Key:=0;

    	EditNode(Selected);
    end;
  end else if (Key=VK_SPACE) and Dragging then begin
    GetCursorPos(curP);

    curP:=ScreenToClient(curP);

    CheckExpandWhileDragging(curP.X,curP.Y);
  end;
end;

procedure TDBTreeView.DoStartDrag(var DragObject:TDragObject);
var Node:TTreeNode;
begin
	inherited DoStartDrag(DragObject);

  if not (csDesigning in ComponentState) then begin
  	Node:=GetNodeAt(FDownX,FDownY);

  	if Node<>nil then begin
      DragObject:=TDragNodeControlObject.Create(FDragNodeControl,Self,Node,GetNodeLevel(Node));
    end;
  end;
end;

procedure TDBTreeView.DragOver(Source:TObject;X,Y:Integer;State:TDragState;var Accept:Boolean);
var dragNode:TDragNodeControl;
		xyNode:TTreeNode;
		nxyL,ndL:Integer;
    evDragOver:TEventOnDragOver;
begin
	inherited DragOver(Source,X,Y,State,Accept);

  if FControlDragDrop then begin
    if Source is TDragNodeControl then begin
      dragNode:=Source as TDragNodeControl;

      xyNode:=GetNodeAt(X,Y);

      if xyNode<>nil then begin
        nxyL:=GetNodeLevel(xyNode)-1;
        ndL:=GetNodeLevel(dragNode.Node)-1;

        evDragOver:=LD[nxyL].OnDragOver;

        Accept:=(not xyNode.HasAsParent(dragNode.Node)) and (xyNode<>dragNode.Node);
        Accept:=Accept and NodeHasInformation(dragNode.Node) and NodeHasInformation(xyNode);

        if Accept and Assigned(evDragOver) then evDragOver(Self,dragNode.Node,xyNode,LD[ndL].ParamName,Accept);
      end;

      if Y<=FDeltaY then begin
        FScrollingOn:=True;

        BeginScroll(-1);
      end else if Y>Height-FDeltaY then begin
        FScrollingOn:=True;

        BeginScroll(1);
      end else begin
        EndScroll;
      end;

      CheckExpandWhileDragging(X,Y);
    end;
  end;
end;

procedure TDBTreeView.CheckExpandWhileDragging(X,Y:Integer);
var expandNode:TTreeNode;
begin
	if Dragging then begin
    if [htOnButton,htOnItem]*GetHitTestInfoAt(X,Y)<>[] then begin
      if (GetKeyState(VK_SPACE) and $80)<>0 then begin
        expandNode:=GetNodeAt(X,Y);

        if expandNode<>nil then begin
          if (not NodeIsExpanded(expandNode)) and AlwaysExpand(expandNode) then JustExpand(expandNode);
        end;
      end;
    end;
  end;
end;

procedure TDBTreeView.DragDrop(Source:TObject;X,Y:Integer);
var dragNode:TDragNodeControl;
		xyNode:TTreeNode;
		nxyL,ndL:Integer;
    evDragDrop:TEventOnDragDrop;
begin
	inherited DragDrop(Source,X,Y);

  if FControlDragDrop then begin
    if Source is TDragNodeControl then begin
      dragNode:=Source as TDragNodeControl;

      xyNode:=GetNodeAt(X,Y);

      if xyNode<>nil then begin
        nxyL:=GetNodeLevel(xyNode)-1;
        ndL:=GetNodeLevel(dragNode.Node)-1;

        evDragDrop:=LD[nxyL].OnDragDrop;

        if Assigned(evDragDrop) then evDragDrop(Self,dragNode.Node,xyNode,LD[ndL].ParamName);
      end;
    end;

    EndScroll;
  end;
end;

function TDBTreeView.CanEdit(Node:TTreeNode):Boolean;
begin
	Result:=inherited CanEdit(Node) and LD[GetNodeLevel(Node)-1].Editable;
end;

procedure TDBTreeView.BeginScroll(Direction:Integer);
begin
	if FEnableDragScroll then PostMessage(Self.Handle,WMU_SCROLL,Direction,0);
end;

procedure TDBTreeView.EndScroll;
begin
	FScrollingOn:=False;
end;

procedure TDBTreeView.WMScrollTree(var Msg:TMessage);
var mvNode,tTopItem:TTreeNode;
		curP:TPoint;
    Progress:Boolean;
begin
	if DateTimeToTimeStamp(Now).Time-FScrollBegin>FScrollSpeed then begin
  	GetCursorPos(curP);

    curP:=ScreenToClient(curP);

    Progress:=False;

    mvNode:=GetNodeAt(curP.X,curP.Y);

    if mvNode<>nil then begin
      if Msg.WParam=1 then
        mvNode:=mvNode.GetNextVisible
      else
      	mvNode:=mvNode.GetPrevVisible;

      if mvNode<>nil then begin
      	tTopItem:=TopItem;

      	mvNode.MakeVisible;

        if TopItem<>tTopItem then InvalidateRect(Self.Handle,nil,False);

        FScrollBegin:=DateTimeToTimeStamp(Now).Time;

        Progress:=True;
      end;
    end;

    FScrollingOn:=Progress;
  end;

  if FScrollingOn and Dragging then begin
  	Application.ProcessMessages;

  	BeginScroll(Msg.WParam);
  end;
end;

procedure TDBTreeView.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if Operation=opRemove then begin
  	if AComponent=FFirstLevel then FFirstLevel:=nil;
  end;
end;

procedure TDBTreeView.Edit(const Item:TTVItem);
var Node:TTreeNode;
		nL:Integer;
    evEdited:TEventOnEdited;
    S:string;
begin
  FInEditing:=False;
  
  Node:=Items.GetNode(Item.hItem);

  if Node<>nil then begin
  	nL:=GetNodeLevel(Node)-1;

    if LD[nL].Editable then begin
      evEdited:=LD[nL].OnEdited;

      S:=Item.pszText;
      
      if Assigned(evEdited) then evEdited(Self,Node,S);

      Node.Text:=S;

      Selected:=Node;
      Node.Focused:=True;
    end;
  end;
end;

function TDBTreeView.FillList(level:Integer;Node:TTreeNode):LongInt;
var sList:TStringList;
    i,lc:Integer;
    gNode:TTreeNode;
begin
  sList:=TStringList.Create;

  try
    lc:=LD[level].GetList(sList,Node,GetSelections(Node),False);

    for i:=0 to lc-1 do begin
      gNode:=Items.AddChildObject(Node,sList.Strings[i],sList.Objects[i]);

      gNode.ImageIndex:=(sList.Objects[i] as TNodeObject).ImageIndex;
      gNode.SelectedIndex:=gNode.ImageIndex;

      SetHasChildren(gNode,False);

      if LastLevel>0 then begin
	      if level+1<LastLevel then SetHasChildren(gNode,True);
      end else begin
      	if (sList.Objects[i] as TNodeObject).Child then SetHasChildren(gNode,True);
      end;
    end;

    Result:=sList.Count;

    finally
      sList.Free;
  end;
end;

procedure TSimpleListDistributor.SetList(listS:TStringList);
begin
  FList.Assign(listS);
end;

function TSimpleListDistributor.DoGetList(listS:TStringList):LongInt;
var i,lc:Integer;
begin
  lc:=FList.Count-1;

  for i:=0 to lc do begin
    listS.AddObject(GetContext(i),TNodeObject.CreateImage(GetID(i),ImageIndex));
  end;

  Result:=listS.Count;
end;

constructor TSimpleListDistributor.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);

  FList:=TStringList.Create;
end;

destructor TSimpleListDistributor.Destroy;
begin
  FList.Free;

  inherited Destroy;
end;

function TSimpleListDistributor.GetSplitPos(const item:string):Integer;
var i,posD,lenI:Integer;
begin
  Result:=0;

  lenI:=Length(item);

  for i:=1 to NOfParameters do begin
    posD:=Pos(CDefParamDelim,Copy(item,Result+1,lenI));

    if posD=0 then Exit;

    lenI:=lenI-posD;
    Result:=Result+posD;
  end;
end;

function TSimpleListDistributor.GetContext(ind:Integer):string;
var posD:Integer;
begin
  Result:=FList.Strings[ind];

  posD:=GetSplitPos(Result);

  if posD>=1 then Result:=Copy(Result,posD+1,Length(Result)-posD);
end;

function TSimpleListDistributor.GetID(ind:Integer):string;
var posD:Integer;
begin
  Result:=FList.Strings[ind];

  posD:=GetSplitPos(Result);

  if posD>=1 then Result:=Copy(Result,1,posD-1);
end;

initialization
	RegisterClasses([TDragNodeControl,TClipboardNodeComponent]);
end.
