{$O+A+W-U-I+B-X+T-P+H+J+R-Q-D-L-}

//******************************************************
//******************************************************
// Project: Delphi Initialization Database
// Author: JBD
// E-mail: binel@alf.tel.hr
// Comment: Utility functions and classes
//******************************************************
//******************************************************

unit INIUtil;

interface

uses
	SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
	Forms, Dialogs, ExtCtrls,DsgnIntf,StdCtrls,Windows,CommCtrl,ComCtrls,Buttons,
  Spin,ShellApi,Menus;

type
	TDrawTextFlags=(dtBottom,dtCenter,dtExpandTabs,dtExternalLeading,dtLeft,
									dtNoClip,dtNoPrefix,dtRight,dtSingleLine,dtTabStop,dtTop,dtVCenter,dtWordBreak);

const
	CDrawTextFlagValues:array[TDrawTextFlags] of Word=(
			dt_Bottom,dt_Center,dt_ExpandTabs,dt_ExternalLeading,dt_Left,
			dt_NoClip,dt_NoPrefix,dt_Right,dt_SingleLine,dt_TabStop,dt_Top,dt_VCenter,dt_WordBreak
	);

type
	TTextFormatFlag=set of TDrawTextFlags;

  TListViewExtended=class(TListView)
  private
  	FSortOnColumn:Integer;

		procedure LVExtendedCompare(Sender: TObject; Item1,Item2: TListItem; Data: Integer; var Compare: Integer);
    
  protected
  	procedure ColClick(Column:TListColumn); override;

  public
		constructor Create(AOwner:TComponent); override;
  end;

	TPaintTextBox = class(TPaintBox)
	private
		FMsgText:string;
		FTextFormat:TTextFormatFlag;
		FXBorder,FYBorder,FShadow:Integer;
    FShadowColor:TColor;
    FImageList:TImageList;
    FImageIndex:Integer;

		procedure SetMessageText(const Msg:string);
		procedure SetTextFormat(Format:TTextFormatFlag);
		procedure SetXBorder(xBor:Integer);
		procedure SetYBorder(yBor:Integer);
		procedure SetShadow(sh:Integer);
		procedure SetShadowColor(shCol:TColor);

		function GetFormat:Word;

		property FormatValue:Word read GetFormat;

	protected
		procedure Paint; override;

    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
	public
		constructor Create(AOwner:TComponent); override;

	published
		property MessageText:string read FMsgText write SetMessageText;
		property ImageList:TImageList read FImageList write FImageList;
		property Caption:string read FMsgText write SetMessageText;
		property TextFormat:TTextFormatFlag read FTextFormat write SetTextFormat;
		property XBorder:Integer read FXBorder write SetXBorder;
		property YBorder:Integer read FYBorder write SetYBorder;
		property Shadow:Integer read FShadow write SetShadow;
    property ShadowColor:TColor read FShadowColor write SetShadowColor;
    property ImageIndex:Integer read FImageIndex write FImageIndex default -1;
	end;

	TShowHintBox = class(TPaintBox)
	private
		FHintContext:string;
		FFont1,FFont2,FFont3:TFont;

		procedure SetFont1(font:TFont);
		procedure SetFont2(font:TFont);
		procedure SetFont3(font:TFont);

		procedure SetHintContext(const hc:string);

	protected
		procedure Paint; override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

	published
		property Font1:TFont read FFont1 write SetFont1;
		property Font2:TFont read FFont2 write SetFont2;
		property Font3:TFont read FFont3 write SetFont3;

		property HintContext:string read FHintContext write SetHintContext;
	end;
  
const CMaxRows=16;
			CMaxCols=16;

type
	TWinControlsIn2D=array[1..CMaxRows,1..CMaxCols] of TControl;

	TAutoAlignPanel = class(TPanel)
	private
		NOfSB,InRow,Iterator:Integer;
    FOnlyVisible:Boolean;
    FUpdating:Boolean;

		FWinCtrls:TWinControlsIn2D;
		
		FBorders,FDistances:array[0..1] of Integer;
		FRows:Integer;

		function GetBorder(Index:Integer):Integer;
		function GetDistance(Index:Integer):Integer;

		procedure SetBorder(Index:Integer;bor:Integer);
		procedure SetDistance(Index:Integer;Distance:Integer);

		procedure SetRows(r:Integer);

		procedure SetWinCtrl(ir,ic:Integer;wCtrl:TControl);
		function GetWinCtrl(ir,ic:Integer):TControl;

		procedure ResetIterator;
		function NextControl:TControl;

	protected
		function GetControl(var Order:Integer):TControl;

		property WinCtrls[ir,ic:Integer]:TControl read GetWinCtrl write SetWinCtrl;
    
		procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure Resize; override;

		procedure SetBorders(xBor,yBor:Integer); virtual;
		procedure SetDistances(xDist,yDist:Integer); virtual;

    procedure BeginUpdate;
    procedure EndUpdate(DoResize:Boolean);

		function HasControlTag(No:LongInt):Boolean;

	published
		property BorderX:Integer index 0 read GetBorder write SetBorder default 0;
		property BorderY:Integer index 1 read GetBorder write SetBorder default 0;
		property DistanceHor:Integer index 0 read GetDistance write SetDistance default 0;
		property DistanceVer:Integer index 1 read GetDistance write SetDistance default 0;
		property Rows:Integer read FRows write SetRows default 1;
    property OnlyVisible:Boolean read FOnlyVisible write FOnlyVisible default True;
	end;

type
	TDirection=(rdHorizontal,rdVertical);

	TLineForResizing = class(TPanel)
	private
		FMinW,FMaxW,FMinH,FMaxH:Integer;
		FMinCW,FMaxCW,FMinCH,FMaxCH:Integer;
    FRuntimeMinCW,FRuntimeMinCH:Integer;
    FLastX,FLastY:Integer;
		FSaveX,FSaveY:Integer;
		FInResizing:Boolean;
		FResizeControl:TControl;
		FOnFinnished,FOnResizing:TNotifyEvent;
		FButton:TMouseButton;
		FShowResizing:Boolean;
		FDirection:TDirection;
    FWParent:TWinControl;

    procedure SetDirection(RD:TDirection);

		procedure DrawLine(x,y:Integer);
    
	protected
		procedure MouseMove(shift:TShiftState;x,y:Integer); override;
		procedure MouseDown(mb:TMouseButton;shift:TShiftState;x,y:Integer); override;
		procedure MouseUp(mb:TMouseButton;shift:TShiftState;x,y:Integer); override;

    procedure SetResizeControl(Ctrl:TControl); virtual;

		procedure ChangeControlSize(var newX,newY:Integer;Apply:Boolean); virtual;

		procedure NotifyFinnished; virtual;
		procedure NotifyResizing; virtual;

		procedure Notification(AComponent:TComponent;Operation:TOperation); override;

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		procedure SetSize(W,H:Integer);
		procedure SetWSize(W:Integer);
		procedure SetHSize(H:Integer);

	published
		property ResizeControl:TControl read FResizeControl write SetResizeControl;
		property ShowResizing:Boolean read FShowResizing write FShowResizing stored True default False;
		property Button:TMouseButton read FButton write FButton default mbLeft;
		property Direction:TDirection read FDirection write SetDirection default rdHorizontal;

		property MinW:Integer read FMinW write FMinW default 0;
		property MaxW:Integer read FMaxW write FMaxW default 0;
		property MinH:Integer read FMinH write FMinH default 0;
		property MaxH:Integer read FMaxH write FMaxH default 0;

		property MinCW:Integer read FMinCW write FMinCW default 0;
		property MaxCW:Integer read FMaxCW write FMaxCW default 0;
		property MinCH:Integer read FMinCH write FMinCH default 0;
		property MaxCH:Integer read FMaxCH write FMaxCH default 0;

		property OnFinnished:TNotifyEvent read FOnFinnished write FOnFinnished;
		property OnResizing:TNotifyEvent read FOnResizing write FOnResizing;
	end;

const
	CResNextPrev=20001;
  CResUpDown=20002;
  CResSaveCurrent=20003;
  CResDeleteItem=20004;
  CResDefault=20005;
  CResSearchSection=20006;

type
	TFC=class(TControl);

	TPanelGroup = class(TPanel)
	private
		FCaptionBox:TPaintTextBox;
		FBevelLine:TBevel;
		FScrollPanel:TScrollBox;

    procedure SetTop(Top:Boolean);

		function GetTitle:string;
		procedure SetTitle(const sCaption:string);

	public
		constructor Create(AOwner:TComponent); override;
		destructor Destroy; override;

		property Title:TPaintTextBox read FCaptionBox;
		property Line:TBevel read FBevelLine;

    property Top:Boolean write SetTop;

	published
		property TitleContext:string read GetTitle write SetTitle;
	end;

	TPanelGroupCascade=class(TPageControl)
	private
    procedure SetTop(Top:Boolean);

  public
    property Top:Boolean write SetTop;
	end;

	TPanelMode=(pmComboList,pmMemo);

	TPanelList=class(TPanel)
	private
		FComboBox:TComboBox;
    FMemo:TMemo;
    FPanelMode:TPanelMode;
		FButtonSave,FButtonDelete,FButtonClear:TBitBtn;
		FSpinButton,FSpinButtonChange:TSpinButton;
    FPopupMenu:TPopupMenu;
		FMISave,FMIDelete,FMIClear,FMISearch:TMenuItem;

    procedure SetItemIndex(ii:Integer);
		procedure SetText(const Text:string);

    function GetText:string;
    function GetItems:TStrings;
    function GetItemIndex:Integer;

		function GetOnListChange:TNotifyEvent;
		procedure SetOnListChange(Event:TNotifyEvent);

		function GetOnNameChange:TNotifyEvent;
		procedure SetOnNameChange(Event:TNotifyEvent);

		function GetOnButtonSaveClick:TNotifyEvent;
		procedure SetOnButtonSaveClick(Event:TNotifyEvent);

		function GetOnButtonDeleteClick:TNotifyEvent;
		procedure SetOnButtonDeleteClick(Event:TNotifyEvent);

		function GetOnButtonClearClick:TNotifyEvent;
		procedure SetOnButtonClearClick(Event:TNotifyEvent);

		procedure OnMIClick(Sender:TObject);
    
	public
		constructor CreatePM(AOwner:TComponent;PM:TPanelMode);
		destructor Destroy; override;

		procedure AdjustControls;

		procedure Resize; override;

    property Text:string read GetText write SetText;
    property Items:TStrings read GetItems;
    property ItemIndex:Integer read GetItemIndex write SetItemIndex;

		property Combo:TComboBox read FComboBox;
		property Memo:TMemo read FMemo;
		property SpinButton:TSpinButton read FSpinButton;
		property SpinButtonChange:TSpinButton read FSpinButtonChange;
		property ButtonSave:TBitBtn read FButtonSave;
		property ButtonDelete:TBitBtn read FButtonDelete;
		property ButtonClear:TBitBtn read FButtonClear;

	published
		property OnListChange:TNotifyEvent read GetOnListChange write SetOnListChange;
		property OnNameChange:TNotifyEvent read GetOnNameChange write SetOnNameChange;
		property OnButtonSaveClick:TNotifyEvent read GetOnButtonSaveClick write SetOnButtonSaveClick;
		property OnButtonDeleteClick:TNotifyEvent read GetOnButtonDeleteClick write SetOnButtonDeleteClick;
		property OnButtonClearClick:TNotifyEvent read GetOnButtonClearClick write SetOnButtonClearClick;
	end;

  TGoToWeb = class(TComponent)
  private
    FUrl:string;

    procedure SetUrl(const Value:string);

  public
    constructor Create(AOwner:TComponent); override;

    function IsAnyBrowser:Boolean;
    function Execute:Boolean;

  published
    property URL:string read FUrl write SetUrl;
  end;

	TEditShortCut=class(TEdit)
	private
		FCode:TShortCut;

		procedure RefreshText;

		procedure SetCode(Code:TShortCut);

	protected
		procedure KeyDown(var Key:word;Shift:TShiftState); override;
	public
		constructor Create(AOwner:TComponent); override;

		procedure SendKeyDown(var Key:word;Shift:TShiftState);

		property Code:TShortCut read FCode write SetCode;
	end;

	TCheckBoxWithSC=class(TCheckBox)
	private
		FBorrowEdit:TEditShortCut;
		FSaveCaption:string;

	protected
		procedure KeyDown(var Key:word;Shift:TShiftState); override;

	public
		constructor Create(AOwner:TComponent); override;

		procedure AddToCaption(const Add:string);

	published
		property BorrowEdit:TEditShortCut read FBorrowEdit write FBorrowEdit;
    property SaveCaption:string write FSaveCaption;
	end;

const CMaxAppColors=37;
			CColors:array[0..CMaxAppColors-1] of TColor=(clBlack,clMaroon,clGreen,clOlive,clNavy,clPurple,clTeal,clGray,
																	clSilver,clRed,clLime,clYellow,clBlue,clFuchsia,clAqua,clWhite,
																	clActiveBorder, clActiveCaption, clAppWorkSpace, clBackground, clBtnFace,
                                  clBtnHighlight, clBtnShadow, clBtnText, clCaptionText, clGrayText, clHighlight,
                                  clHighlightText, clInactiveBorder, clInactiveCaption, clInactiveCaptionText,
                                  clMenu, clMenuText, clScrollBar, clWindow, clWindowFrame, clWindowText);

type
	TComboGetColor=class(TComboBox)
	protected
		procedure MeasureItem(Index: Integer; var Height: Integer); override;
		procedure DrawItem(Index:Integer;r:TRect;ODState:TOwnerDrawState); override;
	public
		constructor Create(AOwner:TComponent); override;
	end;

function VKeyOfShortCut(sCut:TShortCut):word;

function GoToWeb(const URL:string):Boolean;
procedure ArrangeToFit(WControl:TWinControl;Distance:Integer;ArrangeWidth:Boolean);

implementation

uses INIMis,INISec;

const nbr_extension=4;
      Extension:array[0..nbr_extension-1] of string=('.HTM','.HTML','.SHTML','.SHTM');

procedure ArrangeToFit(WControl:TWinControl;Distance:Integer;ArrangeWidth:Boolean);
var i,cc,dfW:Integer;
    WC:TControl;
begin
  cc:=WControl.ControlCount-1;

  for i:=0 to cc do begin
    WC:=WControl.Controls[i];

    if WC.Visible then begin
      if WC.Left+WC.Width>WControl.Width then begin
        dfW:=-WControl.Width+(WC.Left+WC.Width);

      	if ArrangeWidth then
  	      WC.Width:=WC.Width-dfW-Distance
        else
  	      WC.Left:=WC.Left-dfW-Distance;
      end;

      if WC is TWinControl then ArrangeToFit(WC as TWinControl,Distance,ArrangeWidth);
    end;
  end
end;


constructor TPaintTextBox.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FTextFormat:=[dtCenter,dtVCenter,dtWordBreak];
	FXBorder:=0;
	FYBorder:=0;
	FShadow:=0;

  FImageList:=nil;
  FImageIndex:=-1;

  FShadowColor:=clBlack;
end;

procedure TPaintTextBox.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

  if Operation=opRemove then begin
  	if AComponent=FImageList then FImageList:=nil;
  end;
end;

procedure TPaintTextBox.SetMessageText(const Msg:string);
begin
	FMsgText:=Msg;

	Refresh;
end;

procedure TPaintTextBox.SetTextFormat(Format:TTextFormatFlag);
begin
	FTextFormat:=Format;

	Refresh;
end;

procedure TPaintTextBox.SetXBorder(xBor:Integer);
begin
	FXBorder:=xBor;

  Refresh;
end;

procedure TPaintTextBox.SetYBorder(yBor:Integer);
begin
	FYBorder:=yBor;

	Refresh;
end;

procedure TPaintTextBox.SetShadow(sh:Integer);
begin
	FShadow:=sh;

	Refresh;
end;

procedure TPaintTextBox.SetShadowColor(shCol:TColor);
begin
	FShadowColor:=shCol;

	Refresh;
end;

function TPaintTextBox.GetFormat:Word;
var dtw:TDrawTextFlags;
begin
	Result:=0;

	for dtw:=Low(TDrawTextFlags) to High(TDrawTextFlags) do begin
		if dtw in TextFormat then Result:=Result or CDrawTextFlagValues[dtw];
	end;
end;

procedure TPaintTextBox.Paint;
var r,rs:TRect;
		Left,Top,W,mLen:Integer;
		fValue:Word;
		saveColor:TColor;
    Msg:string;
begin
	inherited Paint;

  Msg:=MessageText;

  mLen:=Length(Msg);

	SetBkMode(Canvas.Handle,OPAQUE);
	Canvas.Brush.Color:=Color;
	Canvas.FillRect(Rect(0,0,Width,Height));

	r:=Rect(XBorder,YBorder,Width-XBorder,Height-YBorder);

  if (FImageList<>nil) and (FImageIndex>=0) and (FImageIndex<FImageList.Count) then begin
  	r.Left:=r.Left+FImageList.Width+3;

    FImageList.Draw(Canvas,1,(Height-FImageList.Height) div 2,FImageIndex);
  end;

  W:=r.Right-r.Left+1;

	if ([dtWordBreak]*TextFormat=[]) and (Canvas.TextWidth(Msg)>W) and (W>0) then begin
  	repeat
    	Msg:=INIMis.Left(Msg,Length(Msg)-1);
    until Canvas.TextWidth(Msg+'...')<W;

    Msg:=Msg+'...';

    mLen:=Length(Msg);
  end;

	fValue:=FormatValue;

	DrawText(Canvas.Handle,@(Msg[1]),mLen,r,DT_CALCRECT or fValue);

	if [dtVCenter,dtBottom]*TextFormat<>[] then begin
		Top:=Height-(r.Bottom-r.Top);

		if dtVCenter in TextFormat then Top:=Top div 2;

		r.Bottom:=Top+(r.Bottom-r.Top);
		r.Top:=Top;
	end;

	if [dtCenter,dtRight]*TextFormat<>[] then begin
		Left:=Width-(r.Right-r.Left);

		if dtCenter in TextFormat then Left:=Left div 2;

		r.Right:=Left+(r.Right-r.Left);
		r.Left:=Left;
	end;

	if Shadow=0 then begin
		DrawText(Canvas.Handle,@(Msg[1]),mLen,r,fValue);
	end else begin
		rs:=r;
		OffsetRect(rs,Shadow,Shadow);

		saveColor:=Canvas.Font.Color;
		Canvas.Font.Color:=ShadowColor;
		DrawText(Canvas.Handle,@(Msg[1]),mLen,rs,fValue);

		SetBkMode(Canvas.Handle,TRANSPARENT);
		Canvas.Font.Color:=saveColor;
		DrawText(Canvas.Handle,@(Msg[1]),mLen,r,fValue);
	end;
end;

procedure TShowHintBox.SetFont1(font:TFont);
begin
	FFont1.Assign(font);

  Paint;
end;

procedure TShowHintBox.SetFont2(font:TFont);
begin
	FFont2.Assign(font);

	Paint;
end;

procedure TShowHintBox.SetFont3(font:TFont);
begin
	FFont3.Assign(font);

	Paint;
end;

procedure TShowHintBox.SetHintContext(const hc:string);
begin
	if hc<>FHintContext then begin
		FHintContext:=hc;

		Paint;
	end;
end;

procedure TShowHintBox.Paint;
var i,hcLen:Integer;
		xOfs,yOfs:Integer;
begin
	inherited Paint;

	hcLen:=Length(HintContext);

	SetBkMode(Canvas.Handle,OPAQUE);

	Canvas.Font:=Font;
	Canvas.Brush.Color:=Color;
	Canvas.FillRect(Rect(0,0,Width,Height));

	xOfs:=0;

  i:=1;

  while i<=hcLen do begin
		if HintContext[i]='#' then begin
   		if i+1<=hcLen then begin
				case HintContext[i+1] of
					'1':Canvas.Font:=Font1;
					'2':Canvas.Font:=Font2;
					'3':Canvas.Font:=Font3;
				else
					Canvas.Font:=Font;
				end
			end;

      i:=i+1;
		end else begin
			yOfs:=(Height-Canvas.TextHeight(HintContext[i])) div 2;

			Canvas.TextOut(xOfs,yOfs,HintContext[i]);

      xOfs:=xOfs+Canvas.TextWidth(HintContext[i]);
		end;

    i:=i+1;
  end;
end;

constructor TShowHintBox.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FFont1:=TFont.Create;
	FFont2:=TFont.Create;
	FFont3:=TFont.Create;
end;

destructor TShowHintBox.Destroy;
begin
	FFont1.Free;
	FFont2.Free;
	FFont3.Free;

	inherited Destroy;
end;


constructor TAutoAlignPanel.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

  FBorders[0]:=0;
  FBorders[1]:=0;
  FDistances[0]:=0;
  FDistances[1]:=0;
  FRows:=1;

  FOnlyVisible:=True;
  FUpdating:=False;
end;

function TAutoAlignPanel.GetBorder(Index:Integer):Integer;
begin
	Result:=FBorders[Index];
end;

function TAutoAlignPanel.GetDistance(Index:Integer):Integer;
begin
	Result:=FDistances[Index];
end;

procedure TAutoAlignPanel.SetBorder(Index:Integer;bor:Integer);
begin
	FBorders[Index]:=bor;
	Resize;
end;

procedure TAutoAlignPanel.SetDistance(Index:Integer;Distance:Integer);
begin
	FDistances[Index]:=Distance;
	Resize;
end;

procedure TAutoAlignPanel.SetRows(r:Integer);
begin
	if r>0 then FRows:=r;
	Resize;
end;

procedure TAutoAlignPanel.SetWinCtrl(ir,ic:Integer;wCtrl:TControl);
begin
	if (ir<1) or (ir>CMaxRows) or (ic<1) or (ic>CMaxCols) then raise EPropertyError.Create('Index out of range');

	FWinCtrls[ir,ic]:=wCtrl;
end;

function TAutoAlignPanel.GetWinCtrl(ir,ic:Integer):TControl;
begin
	if (ir<1) or (ir>CMaxRows) or (ic<1) or (ic>CMaxCols) then raise EPropertyError.Create('Index out of range');

	Result:=FWinCtrls[ir,ic];
end;

procedure TAutoAlignPanel.ResetIterator;
begin
	Iterator:=-1;
end;

function TAutoAlignPanel.NextControl:TControl;
begin
	Result:=GetControl(Iterator);
end;

function TAutoAlignPanel.GetControl(var Order:Integer):TControl;
var i:Integer;
begin
	for i:=Order+1 to ControlCount-1 do begin
		if (not OnlyVisible) or (OnlyVisible and Controls[i].Visible) then begin
			Order:=i;
			Result:=Controls[Order];

			Exit;
		end;
	end;

	Order:=ControlCount;

	Result:=nil;
end;

procedure TAutoAlignPanel.Resize;
var ir,ic,i:Integer;
		W,H,cW,cH,cX,cY:Integer;
		Rest,inThisRow:Integer;
		currControl:TControl;
begin
	inherited Resize;

  if not FUpdating then begin
    NOfSB:=0;

    for i:=0 to ControlCount-1 do begin
      if (not OnlyVisible) or (OnlyVisible and Controls[i].Visible) then NOfSB:=NOfSB+1;
    end;

    if NOfSB>0 then begin
      InRow:=NOfSB div Rows;

      if InRow=0 then begin
        FRows:=NOfSB;
        InRow:=1;
      end;

      Rest:=NOfSB-InRow*Rows;

      H:=ClientHeight-(BorderWidth shl 1)-(BorderY shl 1)-(Rows-1)*DistanceVer;
      cH:=H div Rows;

      cY:=BorderWidth+BorderY;

      ResetIterator;

      for ir:=1 to Rows do begin
        inThisRow:=InRow+Integer(Rest>0);

        if ir=Rows then cH:=H-cH*(Rows-1);

        W:=ClientWidth-(BorderWidth shl 1)-(BorderX shl 1)-(inThisRow-1)*DistanceHor;
        cW:=W div inThisRow;

        cX:=BorderWidth+BorderX;

        for ic:=1 to inThisRow do begin
          currControl:=NextControl;

          if ic=inThisRow then cW:=W-cW*(inThisRow-1);

          currControl.SetBounds(cX,cY,cW,cH);

          cX:=cX+(cW+DistanceHor);
        end;

        cY:=cY+(cH+DistanceVer);

        Rest:=Rest-1;
      end;
    end;
  end;
end;

function TAutoAlignPanel.HasControlTag(No:LongInt):Boolean;
var i,contrC:LongInt;
begin
	contrC:=ControlCount-1;

	for i:=0 to contrC do begin
		if Controls[i].Tag=No then begin
			Result:=True;

			Exit;
		end;
	end;

	Result:=False;
end;

procedure TAutoAlignPanel.SetBorders(xBor,yBor:Integer);
begin
	FBorders[0]:=xBor;
	FBorders[1]:=yBor;

	Resize;
end;

procedure TAutoAlignPanel.SetDistances(xDist,yDist:Integer);
begin
	FDistances[0]:=xDist;
	FDistances[1]:=yDist;

	Resize;
end;

procedure TAutoAlignPanel.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if Operation=opInsert then begin
  	if AComponent is TControl then begin
    	if (AComponent as TControl).Parent=Self then Resize;
    end;
  end;
end;

procedure TAutoAlignPanel.BeginUpdate;
begin
	FUpdating:=True;
end;

procedure TAutoAlignPanel.EndUpdate(DoResize:Boolean);
begin
	FUpdating:=False;

  if DoResize then Resize;
end;

procedure TLineForResizing.NotifyFinnished;
begin
	if Assigned(FOnFinnished) then FOnFinnished(Self);
end;

procedure TLineForResizing.NotifyResizing;
begin
	if Assigned(FOnResizing) then FOnResizing(Self);
end;

procedure TLineForResizing.DrawLine(x,y:Integer);
var DC:HDC;
    HP:HPEN;
    F,L:TPoint;
begin
	if FDirection=rdHorizontal then begin
  	F.x:=x;
    F.y:=0;

  	F:=ClientToScreen(F);

    L.x:=F.x;
    L.y:=F.y+Height-1;
  end else begin
  	F.x:=0;
    F.y:=y;

  	F:=ClientToScreen(F);

    L.x:=F.x+Width-1;
    L.y:=F.y;
  end;

  DC:=GetWindowDC(0);
  HP:=CreatePen(PS_SOLID,max(2,min(Width,Height)-1),clWhite);

  try
    SelectObject(DC,HP);
    SetROP2(DC,R2_XORPEN);

    Windows.MoveToEx(DC,F.x,F.y,nil);
    Windows.LineTo(DC,L.x,L.y);

    finally
    	DeleteObject(HP);
      ReleaseDC(0,DC);
  end;
end;

procedure TLineForResizing.MouseMove(shift:TShiftState;x,y:Integer);
begin
	inherited MouseMove(shift,x,y);

	if FInResizing then begin
		if ShowResizing then begin
			ChangeControlSize(x,y,True);
		end else begin
			ChangeControlSize(x,y,False);

      if (FLastX<>x) or (FLastY<>y) then begin
        DrawLine(FLastX,FLastY);
        DrawLine(x,y);

        FLastX:=x;
        FLastY:=y;
      end;
    end;

		NotifyResizing;
	end;
end;

procedure TLineForResizing.MouseDown(mb:TMouseButton;shift:TShiftState;x,y:Integer);
begin
	inherited MouseDown(mb,shift,x,y);

	if mb=Button then begin
		FInResizing:=True;

    FRuntimeMinCW:=MinCW;
    FRuntimeMinCH:=MinCH;

    if Parent is TWinControl then begin
      FWParent:=Parent as TWinControl;

      if ResizeControl<>nil then begin
	     	FRuntimeMinCW:=min(MinCW,FWParent.ClientWidth-ResizeControl.Width-Width-1);
  	   	FRuntimeMinCH:=min(MinCH,FWParent.ClientHeight-ResizeControl.Height-Height-1);
      end;
    end;

		FSaveX:=x;
		FSaveY:=y;
    FLastX:=x;
    FLastY:=y;

    if not FShowResizing then begin
    	DrawLine(x,y);
    end;
	end;
end;

procedure TLineForResizing.MouseUp(mb:TMouseButton;shift:TShiftState;x,y:Integer);
begin
	inherited MouseUp(mb,shift,x,y);

	if FInResizing then begin
		FInResizing:=False;

    if not FShowResizing then begin
			ChangeControlSize(x,y,False);
    	DrawLine(FLastX,FLastY);
    end;

		ChangeControlSize(x,y,True);

		NotifyFinnished;
	end;
end;

procedure TLineForResizing.SetResizeControl(Ctrl:TControl);
begin
	FResizeControl:=Ctrl;
end;

procedure TLineForResizing.ChangeControlSize(var newX,newY:Integer;Apply:Boolean);
var newW,newH:Integer;
begin
	if ResizeControl<>nil then begin
		if Direction=rdHorizontal then begin
      if Align=alLeft then
				newW:=ResizeControl.Width+newX-FSaveX
      else
				newW:=ResizeControl.Width-newX+FSaveX;

			if MaxW=0 then
				newW:=Max(MinW,newW)
			else
				newW:=Max(MinW,Min(MaxW,newW));

      if (FWParent<>nil) and (FWParent.ClientWidth-FRuntimeMinCW-Width-1<=newW) then
      	newW:=FWParent.ClientWidth-FRuntimeMinCW-Width-1;

     	if Apply then ResizeControl.Width:=newW;

      if Align=alLeft then
				newX:=newW-ResizeControl.Width+FSaveX
      else
				newX:=newW+ResizeControl.Width-FSaveX;
		end else begin
      if Align=alTop then
				newH:=ResizeControl.Height+newY-FSaveY
      else
				newH:=ResizeControl.Height-newY+FSaveY;

			if MaxH=0 then
				newH:=Max(MinH,newH)
			else
				newH:=Max(MinH,Min(MaxH,newH));

      if (FWParent<>nil) and (FWParent.ClientHeight-FRuntimeMinCH-Height-1<=newH) then
      	newH:=FWParent.ClientHeight-FRuntimeMinCH-Height-1;

     	if Apply then ResizeControl.Height:=newH;

      if Align=alTop then
				newY:=newH-ResizeControl.Height+FSaveY
      else
				newY:=newH+ResizeControl.Height-FSaveY;
		end;
	end;
end;

constructor TLineForResizing.Create(AOwner:TComponent);
begin
	FMinW:=0;
	FMaxW:=0;
	FMinH:=0;
	FMaxH:=0;

	FMinCW:=0;
	FMaxCW:=0;
	FMinCH:=0;
	FMaxCH:=0;

	FSaveX:=0;
	FSaveY:=0;
	FInResizing:=False;

	FResizeControl:=nil;

	FShowResizing:=False;
	FButton:=mbLeft;

	FOnFinnished:=nil;
	FOnResizing:=nil;
  
  FWParent:=nil;

	FDirection:=rdHorizontal;

	inherited Create(AOwner);

  Caption:=' ';
  Cursor:=crHSplit;
end;

destructor TLineForResizing.Destroy;
begin
	inherited Destroy;
end;

procedure TLineForResizing.SetDirection(RD:TDirection);
begin
  FDirection:=RD;

  if RD=rdHorizontal then begin
    Cursor:=crHSplit;
  end else begin
    Cursor:=crVSplit;
  end;
end;

procedure TLineForResizing.SetSize(W,H:Integer);
begin
	if ResizeControl<>nil then begin
  	ResizeControl.SetBounds(ResizeControl.Left,ResizeControl.Top,W,H);
	end;
end;

procedure TLineForResizing.SetWSize(W:Integer);
begin
	SetSize(W,ResizeControl.Height);
end;

procedure TLineForResizing.SetHSize(H:Integer);
begin
	SetSize(ResizeControl.Width,H);
end;

procedure TLineForResizing.Notification(AComponent:TComponent;Operation:TOperation);
begin
	inherited Notification(AComponent,Operation);

	if Operation=opRemove then begin
  	if AComponent=FResizeControl then FResizeControl:=nil;
  end;
end;

procedure TPanelGroup.SetTitle(const sCaption:string);
begin
	FCaptionBox.MessageText:=sCaption;
end;

function TPanelGroup.GetTitle:string;
begin
	Result:=FCaptionBox.MessageText;
end;

constructor TPanelGroup.Create(AOwner:TComponent);

	function GetOwnerForm(AOwner:TComponent):TForm;
  begin
  	if AOwner is TForm then begin
    	Result:=AOwner as TForm;

      Exit;
    end;

    Result:=GetOwnerForm(AOwner.Owner);
  end;

begin
	inherited Create(AOwner);

	Caption:='';
	BevelOuter:=bvNone;
	BevelInner:=bvLowered;

	FCaptionBox:=TPaintTextBox.Create(Self);

	with FCaptionBox do begin
		Parent:=Self;
		Align:=alTop;
		Shadow:=0;
		ShadowColor:=clWhite;
    ParentFont:=True;
		Font.Color:=clWindowText;
    Font.Style:=[fsBold];
		Height:=GetOwnerForm(AOwner).Canvas.TextHeight('Wg')+6;
    TextFormat:=TextFormat-[dtWordBreak];
		XBorder:=5;
	end;

	FBevelLine:=TBevel.Create(Self);

	with FBevelLine do begin
		Parent:=Self;
		Align:=alTop;
		Height:=3;
		Shape:=bsTopLine;
		Style:=bsLowered;
	end;

	FScrollPanel:=TScrollBox.Create(Self);

	with FScrollPanel do begin
		Parent:=Self;
		Align:=alClient;
		BorderStyle:=bsNone;
		VertScrollBar.Increment:=16;
		HorzScrollBar.Increment:=Width;
		HorzScrollBar.Visible:=False;
	end;
end;

destructor TPanelGroup.Destroy;
begin
	FCaptionBox.Free;
	FBevelLine.Free;
	FScrollPanel.Free;

	inherited Destroy;
end;

procedure TPanelGroup.SetTop(Top:Boolean);
begin
	if Top then begin
  	FCaptionBox.Align:=alTop;
    FBevelLine.Align:=alTop;
  end else begin
  	FCaptionBox.Align:=alBottom;
    FBevelLine.Align:=alBottom;
  end;
end;

procedure TPanelGroupCascade.SetTop(Top:Boolean);
begin
	if Top then begin
  	SetWindowLong(Handle,GWL_STYLE,not $0002 and GetWindowLong(Handle,GWL_STYLE));
  end else begin
  	SetWindowLong(Handle,GWL_STYLE,$0002 or GetWindowLong(Handle,GWL_STYLE));
  end;
end;

constructor TPanelList.CreatePM(AOwner:TComponent;PM:TPanelMode);

	function NewMI(const MICaption:string;MITag:LongInt):TMenuItem;
  begin
  	Result:=TMenuItem.Create(FPopupMenu);

    with Result do begin
    	Caption:=MICaption;

    	if Caption<>'-' then Caption:='&'+Caption;

      Tag:=MITag;

      OnClick:=OnMIClick;
    end;
  end;

begin
	inherited Create(AOwner);

  FPanelMode:=PM;

	BevelInner:=bvNone;
	BevelOuter:=bvNone;

  if PM=pmComboList then begin
    BorderWidth:=2;

    FPopupMenu:=TPopupMenu.Create(Self);
    FMISave:=NewMI(LoadStr(CResSaveCurrent),0);
    FMIDelete:=NewMI(LoadStr(CResDeleteItem),1);
    FMIClear:=NewMI(LoadStr(CResDefault),2);
    FMISearch:=NewMI(LoadStr(CResSearchSection),3);

    FPopupMenu.Items.Add(FMISave);
    FPopupMenu.Items.Add(FMIDelete);
    FPopupMenu.Items.Add(FMIClear);
    FPopupMenu.Items.Add(NewMI('-',-1));
    FPopupMenu.Items.Add(FMISearch);

    FComboBox:=TComboBox.Create(Self);

    with FComboBox do begin
      Parent:=Self;
      Top:=BorderWidth;
      Left:=BorderWidth;
      Color:=clWhite;
      Style:=csDropDown;
      PopupMenu:=FPopupMenu;
    end;

    Height:=FComboBox.Height+6;

    FSpinButtonChange:=TSpinButton.Create(Self);

    with FSpinButtonChange do begin
      Parent:=Self;
      Top:=BorderWidth;
      Left:=FComboBox.Left+FComboBox.Width+BorderWidth;
      Hint:=LoadStr(CResNextPrev);
      ShowHint:=True;
    end;

    FSpinButton:=TSpinButton.Create(Self);

    with FSpinButton do begin
      Parent:=Self;
      Top:=BorderWidth;
      Left:=FSpinButtonChange.Left+FSpinButtonChange.Width+BorderWidth;
      Hint:=LoadStr(CResUpDown);
      ShowHint:=True;
    end;

    FButtonSave:=TBitBtn.Create(Self);

    with FButtonSave do begin
      Parent:=Self;
      Caption:='';
      Style:=bsNew;
      Top:=BorderWidth;
      Left:=FSpinButton.Left+FSpinButton.Width+BorderWidth;
      Height:=FComboBox.Height;
      Visible:=False;
      Hint:=LoadStr(CResSaveCurrent);
      ShowHint:=True;
    end;

    FButtonDelete:=TBitBtn.Create(Self);

    with FButtonDelete do begin
      Parent:=Self;
      Caption:='';
      Style:=bsNew;
      Top:=BorderWidth;
      Left:=FButtonSave.Left+FButtonSave.Width+BorderWidth;
      Height:=FComboBox.Height;
      Visible:=False;
      Hint:=LoadStr(CResDeleteItem);
      ShowHint:=True;
    end;

    FButtonClear:=TBitBtn.Create(Self);

    with FButtonClear do begin
      Parent:=Self;
      Caption:='';
      Style:=bsNew;
      Top:=BorderWidth;
      Left:=FButtonDelete.Left+FButtonDelete.Width+BorderWidth;
      Height:=FComboBox.Height;
      Visible:=False;
      Hint:=LoadStr(CResDefault);
      ShowHint:=True;
    end;
  end else begin
    BorderWidth:=0;
    Height:=32;

    FMemo:=TMemo.Create(Self);

    with FMemo do begin
      Parent:=Self;
      Align:=alClient;
    end;
  end;
end;

destructor TPanelList.Destroy;
begin
	inherited Destroy;
end;

procedure TPanelList.OnMIClick(Sender:TObject);
var Event:TNotifyEvent;
begin
	Event:=nil;

	case (Sender as TMenuItem).Tag of
  	0:Event:=FButtonSave.OnClick;
    1:Event:=FButtonDelete.OnClick;
    2:Event:=FButtonClear.OnClick;
    3:PostMessage(FComboBox.Handle,WM_KEYDOWN,VK_F5,0);
  end;

  if Assigned(Event) then Event(Sender);
end;

procedure TPanelList.Resize;
var editW,rightPos,totalW,halfBorder:Integer;
begin
	inherited Resize;

	if FPanelMode=pmComboList then begin
    Height:=FComboBox.Height+6;
    halfBorder:=Max(1,BorderWidth shr 1);

    totalW:=0;
    rightPos:=Self.Width+halfBorder;

    with FButtonClear do begin
      if Visible then begin
        Top:=BorderWidth-1;
        Width:=Height+5;
        Left:=rightPos-halfBorder-Width;

        totalW:=totalW+Width+BorderWidth;
        rightPos:=Left;
      end else Top:=Self.Height+1;
    end;

    with FButtonDelete do begin
      if Visible then begin
        Top:=BorderWidth-1;
        Width:=Height+5;
        Left:=rightPos-halfBorder-Width;

        totalW:=totalW+Width+halfBorder;
        rightPos:=Left;
      end else Top:=Self.Height+1;
    end;

    with FButtonSave do begin
      if Visible then begin
        Top:=BorderWidth-1;
        Width:=Height+5;
        Left:=rightPos-halfBorder-Width;

        totalW:=totalW+Width+halfBorder;
        rightPos:=Left;
      end else Top:=Self.Height+1;
    end;

    if FButtonSave.Visible or FButtonDelete.Visible or FButtonClear.Visible then begin
	    rightPos:=rightPos-1;
  	  totalW:=totalW+1;
    end;

    editW:=Self.Width-(totalW+HalfBorder);

    with FSpinButton do begin
    	if Visible then begin
        Top:=BorderWidth;
        Width:=15;
        Left:=rightPos-halfBorder-Width;
        editW:=editW-(Width+BorderWidth);
        rightPos:=Left;
      end else Top:=Self.Height+1;
    end;

    with FComboBox do begin
      Top:=BorderWidth-1;
      Width:=editW;
      Left:=rightPos-halfBorder-Width;
    end;

    with FSpinButtonChange do begin
    	if Visible then begin
	     	FComboBox.Width:=FComboBox.Width-halfBorder-15;
  	   	FComboBox.Left:=FComboBox.Left+halfBorder+15;

        rightPos:=FComboBox.Left;

	      Top:=BorderWidth;
  	    Width:=15;
    	  Left:=rightPos-halfBorder-Width;

        FComboBox.Width:=FComboBox.Width+Left-halfBorder;
        FComboBox.Left:=FComboBox.Left-(Left-halfBorder);
        
        Left:=halfBorder;
      end else begin
      	Top:=Self.Height+1;

        FComboBox.Width:=FComboBox.Width+FComboBox.Left-halfBorder;
        FComboBox.Left:=halfBorder;
      end;
    end;

    AdjustControls;
  end;
end;

procedure TPanelList.AdjustControls;
begin
	if FPanelMode=pmComboList then begin
	  FSpinButton.Height:=FComboBox.Height-2;
  	FSpinButtonChange.Height:=FComboBox.Height-2;

	  FButtonSave.Height:=FComboBox.Height;
  	FButtonDelete.Height:=FButtonSave.Height;
	  FButtonClear.Height:=FButtonSave.Height;

	  FButtonSave.Font.Size:=TFC(FButtonSave.Parent).Font.Size-1;
  	FButtonDelete.Font.Size:=TFC(FButtonSave.Parent).Font.Size-1;
	  FButtonClear.Font.Size:=TFC(FButtonSave.Parent).Font.Size-1;

    FMISave.Enabled:=FButtonSave.Visible;
    FMIDelete.Enabled:=FButtonDelete.Visible;
    FMIClear.Enabled:=FButtonClear.Visible;
    FMISearch.Enabled:=INISec.FormSearchMultiSectionClass<>nil;
  end;
end;

function TPanelList.GetOnListChange:TNotifyEvent;
begin
	Result:=nil;

	if FPanelMode=pmComboList then Result:=FComboBox.OnClick;
end;

procedure TPanelList.SetOnListChange(Event:TNotifyEvent);
begin
	if FPanelMode=pmComboList then FComboBox.OnClick:=Event;
end;

function TPanelList.GetOnNameChange:TNotifyEvent;
begin
	Result:=nil;

	if FPanelMode=pmComboList then Result:=FComboBox.OnChange;
end;

procedure TPanelList.SetOnNameChange(Event:TNotifyEvent);
begin
	if FPanelMode=pmComboList then FComboBox.OnChange:=Event;
end;

function TPanelList.GetOnButtonSaveClick:TNotifyEvent;
begin
	Result:=nil;

	if FPanelMode=pmComboList then Result:=FButtonSave.OnClick;
end;

procedure TPanelList.SetOnButtonSaveClick(Event:TNotifyEvent);
begin
	if FPanelMode=pmComboList then FButtonSave.OnClick:=Event;
end;

function TPanelList.GetOnButtonDeleteClick:TNotifyEvent;
begin
	Result:=nil;

	if FPanelMode=pmComboList then Result:=FButtonDelete.OnClick;
end;

procedure TPanelList.SetOnButtonDeleteClick(Event:TNotifyEvent);
begin
	if FPanelMode=pmComboList then FButtonDelete.OnClick:=Event;
end;

function TPanelList.GetOnButtonClearClick:TNotifyEvent;
begin
	Result:=nil;

	if FPanelMode=pmComboList then Result:=FButtonClear.OnClick;
end;

procedure TPanelList.SetOnButtonClearClick(Event:TNotifyEvent);
begin
	if FPanelMode=pmComboList then FButtonClear.OnClick:=Event;
end;

function TPanelList.GetText:string;
begin
	if FPanelMode=pmComboList then
  	Result:=FComboBox.Text
  else
  	Result:=FMemo.Text;
end;

procedure TPanelList.SetText(const Text:string);
begin
	if FPanelMode=pmComboList then
  	FComboBox.Text:=Text
  else
  	FMemo.Text:=Text;
end;

function TPanelList.GetItems:TStrings;
begin
	if FPanelMode=pmComboList then
  	Result:=FComboBox.Items
  else
  	Result:=FMemo.Lines;
end;

function TPanelList.GetItemIndex:Integer;
begin
	if FPanelMode=pmComboList then
  	Result:=FComboBox.ItemIndex
  else
  	Result:=0;
end;

procedure TPanelList.SetItemIndex(ii:Integer);
begin
	if FPanelMode=pmComboList then FComboBox.ItemIndex:=ii;
end;

// Component TGotoWeb borrowed from ??? 

function FindBrowser:string;
  var c:Byte;
      hkey1,hkey2:HKEY;
      typ,taille:Integer;
      tmp:array[0..1024] of Char;
      tmpstr:string;

  procedure Arrange;
  var u:Word;
  begin
    tmpstr:=strpas(tmp);
    u:=1;
    repeat
      if tmpstr[u]<>'%' then inc(u) else delete(tmpstr,u,2);
    until u>=length(tmpstr);
    RegCloseKey(hkey2);
    FindBrowser:=tmpstr;
  end;

begin
  c:=0;
  repeat
    if regopenkeyex(hkey_classes_root,pchar(extension[c]),0,key_all_access,hkey1)=0 then
    begin
      if regqueryvalueex(hkey1,nil,nil,@typ,nil,@taille)=error_success then
      begin
        regqueryvalueex(hkey1,nil,nil,nil,@tmp,@taille);
        if tmp<>'' then
        if regopenkeyex(hkey_classes_root,pchar(tmp+'\shell\open\command'),0,key_all_access,hkey2)=error_success then
        begin
          if regqueryvalueex(hkey2,nil,nil,@typ,nil,@taille)=error_success then
          begin
            regqueryvalueex(hkey2,nil,nil,nil,@tmp,@taille);
            if tmp<>'' then
            begin
              Arrange;
              Exit;
            end;
          end;
          regclosekey(hkey2);
        end;
        if regopenkeyex(hkey1,'shell\open\command',0,key_all_access,hkey2)=error_success then
        begin
          if regqueryvalueex(hkey2,nil,nil,@typ,nil,@taille)=error_success then
          begin
            regqueryvalueex(hkey2,nil,nil,nil,@tmp,@taille);
            if tmp<>'' then
            begin
              Arrange;
              Exit;
            end;
          end;
        end;
      end;
      regclosekey(hkey1);
    end;
    inc(c);
  until c=nbr_extension;
  FindBrowser:='';
end;

constructor TGoToWeb.Create(AOwner:Tcomponent);
begin
  inherited Create(AOwner);

  FUrl:='http://public.srce.hr/~khusak/inilib';
end;

function TGoToWeb.IsAnyBrowser:Boolean;
begin
  IsAnyBrowser:=(FindBrowser<>'');
end;

procedure TGoToWeb.SetUrl(const Value:string);
begin
  FUrl:=Value;
end;

function TGotoWeb.Execute:Boolean;
var tmp,para:string;
  	c:Word;
begin
  tmp:=findbrowser;
  if tmp<>'' then
  begin
    tmp:=uppercase(tmp);
    c:=length(tmp)+1;
    repeat
      dec(c);
      para:=copy(tmp,c-3,4);
    until (para='.EXE') or (para='.BAT') or (para='.COM');

    if c<length(tmp) then
    begin
      para:=trim(copy(tmp,c+1,length(tmp)-(c)));
      while (para<>'') and (para[1]='"') do delete(para,1,1);
      delete(tmp,c+1,length(tmp)-(c));
    end else para:='';
    if tmp[1]='"' then delete(tmp,1,1);
    if shellexecute(application.handle,nil,pchar(tmp),pchar(para+' '+FUrl),nil,sw_restore)<33 then execute:=false else execute:=true;
  end else Execute:=false;
end;

function GoToWeb(const URL:string):Boolean;
var gWeb:TGoToWeb;
begin
	gWeb:=TGoToWeb.Create(Application);

  try
  	gWeb.URL:=URL;
    
    Result:=gWeb.Execute;

  	finally
    	gWeb.Free;
  end;
end;

function VKeyOfShortCut(sCut:TShortCut):word;
var shift:TShiftState;
begin
	ShortCutToKey(sCut,Result,shift);
end;

constructor TEditShortCut.Create;
begin
	inherited Create(AOwner);

	ReadOnly:=True;
	Enabled:=False;
end;

procedure TEditShortCut.KeyDown(var Key:word;Shift:TShiftState);
begin
	inherited KeyDown(Key,Shift);

	FCode:=Menus.ShortCut(Key,Shift);

	RefreshText;

	if (Shift=[]) and ((Key=VK_SPACE) or (Key=VK_RETURN)) then begin
		FCode:=0;
		Text:='';
	end;
end;

procedure TEditShortCut.SetCode(Code:TShortCut);
begin
	FCode:=Code;

  RefreshText;
end;

procedure TEditShortCut.RefreshText;
begin
	Text:=AllTrim(ShortCutToText(FCode));
end;

procedure TEditShortCut.SendKeyDown(var Key:word;Shift:TShiftState);
begin
	KeyDown(Key,Shift);
end;

procedure TCheckBoxWithSC.KeyDown(var Key:word;Shift:TShiftState);
begin
	if FBorrowEdit<>nil then FBorrowEdit.SendKeyDown(Key,Shift);
end;

constructor TCheckBoxWithSC.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FBorrowEdit:=nil;
	FSaveCaption:=Caption;
end;

procedure TCheckBoxWithSC.AddToCaption(const Add:string);
begin
	Caption:=FSaveCaption+Add;
end;

procedure TComboGetColor.MeasureItem(Index: Integer; var Height: Integer);
begin
	Height:=GetParentForm(Self).Canvas.TextHeight('Wg')+2;
end;

procedure TComboGetColor.DrawItem(Index:Integer;r:TRect;ODState:TOwnerDrawState);
var cR:TRect;
		clFill:TColor;
begin
	if (Index>=0) and (Index<CMaxAppColors) then begin
    if ODState*[odSelected,odFocused]<>[] then
      clFill:=clHighLight
    else
      clFill:=clWhite;

    Canvas.Brush.Color:=clFill;
    Canvas.FillRect(r);

    Canvas.Brush.Color:=CColors[Index];

    cR.Left:=r.Left+2;
    cR.Top:=r.Top+2;
    cR.Bottom:=r.Bottom-2;
    cR.Right:=cR.Left+(cR.Bottom-cR.Top)-1;

    Canvas.FillRect(cR);
    Canvas.Brush.Color:=clBlack;
    Canvas.FrameRect(cR);

    Canvas.Brush.Color:=clFill;
    Canvas.TextOut(r.Left+cR.Right-cR.Left+12,r.Top,Items.Strings[Index]);
  end;
end;

constructor TComboGetColor.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	Color:=clWhite;
	Style:=csOwnerDrawVariable;
end;

constructor TListViewExtended.Create(AOwner:TComponent);
begin
	inherited Create(AOwner);

	FSortOnColumn:=0;

  OnCompare:=LVExtendedCompare;
end;

procedure TListViewExtended.ColClick(Column:TListColumn);
begin
	inherited ColClick(Column);

	if abs(FSortOnColumn)=Column.Index+1 then
		FSortOnColumn:=-FSortOnColumn
  else
		FSortOnColumn:=Column.Index+1;

	SortType:=stData;

  AlphaSort;
end;

procedure TListViewExtended.LVExtendedCompare(Sender: TObject; Item1,Item2: TListItem; Data: Integer; var Compare: Integer);
begin
	if abs(FSortOnColumn)=1 then
		Compare:=StrIComp(PChar(Item1.Caption),PChar(Item2.Caption))
  else
		Compare:=StrIComp(PChar(Item1.SubItems[abs(FSortOnColumn)-2]),PChar(Item2.SubItems[abs(FSortOnColumn)-2]));

  if FSortOnColumn<0 then Compare:=-Compare;
end;

end.
